/*	$Id: cpufunc.h,v 1.8 2003/01/07 15:12:51 tsubai Exp $	*/

/*-
 * Copyright (c) 2002,2003 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

inline register_t mfmsr(void);
inline void mtmsr(register_t);
inline u_int mfdec(void);
inline void mtdec(u_int);
inline u_int mfpvr(void);
inline u_int64_t mftb(void);

extern inline register_t
mfmsr()
{
	register_t msr;

	asm volatile ("mfmsr %0" : "=r"(msr));
	return msr;
}

extern inline void
mtmsr(msr)
	register_t msr;
{
	asm volatile ("mtmsr %0" :: "r"(msr) : "memory");
}

extern inline u_int
mfdec()
{
	u_int dec;

	asm volatile ("mfdec %0" : "=r"(dec));
	return dec;
}

extern inline void
mtdec(dec)
	u_int dec;
{
	asm volatile ("mtdec %0" :: "r"(dec));
}

extern inline u_int
mfpvr()
{
	u_int pvr;

	asm ("mfpvr %0" : "=r"(pvr));
	return pvr;
}

extern inline u_int64_t
mftb()
{
	u_int64_t tb;
	int tmp;

	asm volatile ("1: mftbu %0; mftb %0+1; mftbu %1; cmplw %0,%1; bne- 1b"
	    : "=r"(tb), "=r"(tmp));
	return tb;
}

#define mfspr(spr) \
  ({ int _val; asm volatile ("mfspr %0,%1" : "=r"(_val) : "i"(spr)); _val; })
#define mtspr(spr, val) asm volatile ("mtspr %0,%1" :: "i"(spr), "r"(val))

#define mfdbatu(bat) \
  ({ int _val; asm volatile ("mfdbatu %0,%1" : "=r"(_val) : "i"(bat)); _val; })
#define mfdbatl(bat) \
  ({ int _val; asm volatile ("mfdbatl %0,%1" : "=r"(_val) : "i"(bat)); _val; })
#define mfibatu(bat) \
  ({ int _val; asm volatile ("mfibatu %0,%1" : "=r"(_val) : "i"(bat)); _val; })
#define mfibatl(bat) \
  ({ int _val; asm volatile ("mfibatl %0,%1" : "=r"(_val) : "i"(bat)); _val; })
#define mtdbatu(bat, val) asm volatile ("mtdbatu %0,%1" :: "i"(bat), "r"(val))
#define mtdbatl(bat, val) asm volatile ("mtdbatl %0,%1" :: "i"(bat), "r"(val))
#define mtibatu(bat, val) asm volatile ("mtibatu %0,%1" :: "i"(bat), "r"(val))
#define mtibatl(bat, val) asm volatile ("mtibatl %0,%1" :: "i"(bat), "r"(val))
