/*	$NetBSD: param.h,v 1.12 2001/06/10 11:01:27 tsubai Exp $	*/

/*-
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _POWERPC_PARAM_H
#define _POWERPC_PARAM_H

/*
 * Machine dependent constants for PowerPC
 */
#ifdef __LP64__
#define MACHINE_ARCH	"powerpc64"
#else
#define MACHINE_ARCH	"powerpc"
#endif
#define MID_MACHINE	MID_POWERPC

#define ALIGNBYTES		15
#define ALIGN(p)		(((u_long)(p) + ALIGNBYTES) & ~ALIGNBYTES)
#define ALIGNED_POINTER(p,t)	((((u_long)(p)) & (sizeof(t)-1)) == 0)

#define PGSHIFT		12
#define NBPG		(1 << PGSHIFT)	/* Page size */
#define PGOFSET		(NBPG - 1)

#define DEV_BSHIFT	9		/* log2(DEV_BSIZE) */
#define DEV_BSIZE	(1 << DEV_BSHIFT)
#define BLKDEV_IOSIZE	NBPG
#define MAXPHYS		(64 * 1024)	/* max raw I/O transfer size */

#define UPAGES		4
#define USPACE		(UPAGES * NBPG)

/*
 * Segment handling stuff
 */
#define SEGMENT_LENGTH	0x10000000
#define SEGMENT_OFFSET	0x0fffffff

/*
 * Fixed segments
 */
#ifdef _LOCORE
#define USER_SR		13
#define KERNEL_SR	14
#else
#define USER_SR		13L
#define KERNEL_SR	14L
#endif
#define EMPTY_SEGMENT	0xfffffff0
#define KERNEL_SEGBASE	0xfff00000
#define KERNEL_SEGMENT	(KERNEL_SEGBASE + KERNEL_SR)
#define USER_SEGMENT	(KERNEL_SEGBASE + USER_SR)
#define USER_ADDR	(USER_SR << ADDR_SR_SHFT)

/*
 * Some system constants
 */
#ifndef NSEGMENTS
#define NSEGMENTS	32768	/* Number of segments in system */
#endif

#ifndef _LOCORE
/* XXX */
struct knote;
int seltrue_kqfilter(dev_t, struct knote *);
#define tty_alloc ttymalloc
#define tty_free ttyfree
#define PCI_FLAGS_IO_ENABLED PCI_FLAGS_IO_OKAY
#define PCI_FLAGS_MEM_ENABLED PCI_FLAGS_MEM_OKAY
#endif

#endif /* _POWERPC_PARAM_H_ */
