/*	$NetBSD: pte.h,v 1.2 1998/08/31 14:43:40 tsubai Exp $	*/

/*-
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _POWERPC64_PTE_H_
#define _POWERPC64_PTE_H_

/*
 * Page Table Entries
 */
#ifndef _LOCORE
struct pte {
	u_int64_t pte_hi;
	u_int64_t pte_lo;
};
typedef struct pte pte_t;

/* High word: */
#define PTE_AVPN	0xffffffffffffff80LL	/* 0-56 */
#define PTE_SW		0x0000000000000078LL	/* 57-60 */
#define PTE_L		0x0000000000000004LL	/* 61: Large page (Cell) */
#define PTE_HID		0x0000000000000002LL	/* 62 */
#define PTE_VALID	0x0000000000000001LL	/* 63 */
/* Low word: */
#define PTE_RPGN	0xfffffffffffff000LL	/* 0-51 */
#define PTE_REF		0x0000000000000100LL	/* 55 */
#define PTE_CHG		0x0000000000000080LL	/* 56 */
#define PTE_WIMG	0x0000000000000078LL	/* 57-60 */
#define PTE_W		0x0000000000000040LL	/* 57 */
#define PTE_I		0x0000000000000020LL	/* 58 */
#define PTE_M		0x0000000000000010LL	/* 59 */
#define PTE_G		0x0000000000000008LL	/* 60 */
#define PTE_PP		0x0000000000000003LL	/* 62-63 */
#define PTE_RO		0x0000000000000003LL
#define PTE_RW		0x0000000000000002LL

struct ste {
	u_int64_t vsid:52;
	u_int ks:1;
	u_int kp:1;
	u_int n:1;
	u_int l:1;
	u_int c:1;
	u_int pad1:7;

	u_int64_t esid:36;
	u_int v:1;
	u_int pad2:15;
	u_int index:12;
};
#endif /* _LOCORE */

/*
 * Extract bits from address
 */
#define ADDR_SR_SHFT	28
#define ADDR_PIDX	0x0ffff000
#define ADDR_PIDX_SHFT	12

#ifndef _LOCORE
#ifdef _KERNEL
extern pte_t *ptable;
extern int ptab_cnt;
extern struct ste ksegtab[];
#endif /* _KERNEL */
#endif /* _LOCORE */

#endif /* _POWERPC64_PTE_H_ */
