/*	$NetBSD: reloc.h,v 1.1 1996/09/30 16:34:33 ws Exp $	*/

/*-
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _POWERPC_RELOC_H_
#define _POWERPC_RELOC_H_

/*
 * Quite a number of relocation types
 */
enum reloc_type {
	RELOC_NONE		= 0,
	RELOC_ADDR32		= 1,
	RELOC_ADDR24		= 2,
	RELOC_ADDR16		= 3,
	RELOC_ADDR16_LO		= 4,
	RELOC_ADDR16_HI		= 5,
	RELOC_ADDR16_HA		= 6,
	RELOC_ADDR14		= 7,
	RELOC_ADDR14_BRTAKEN	= 8,
	RELOC_ADDR14_BRNTAKEN	= 9,
	RELOC_REL24		= 10,
	RELOC_REL14		= 11,
	RELOC_REL14_BRTAKEN	= 12,
	RELOC_REL14_BRNTAKEN	= 13,
	RELOC_GOT16		= 14,
	RELOC_GOT16_LO		= 15,
	RELOC_GOT16_HI		= 16,
	RELOC_GOT16_HA		= 17,
	/* RELOC_PLT24		= 18, */
	RELOC_COPY		= 19,
	RELOC_GLOB_DAT		= 20,
	RELOC_JMP_SLOT		= 21,
	RELOC_RELATIVE		= 22,
	/* RELOC_LOCAL24PC	= 23, */
	RELOC_UADDR32		= 24,
	RELOC_UADDR16		= 25,
	RELOC_REL32		= 26,
	RELOC_PLT32		= 27,
	RELOC_PLTREL32		= 28,
	RELOC_PLT16_LO		= 29,
	RELOC_PLT16_HI		= 30,
	RELOC_PLT16_HA		= 31,
	/* RELOC_SDAREL		= 32, */

	RELOC_SECTOFF		= 33,
	RELOC_SECTOFF_LO	= 34,
	RELOC_SECTOFF_HI	= 35,
	RELOC_SECTOFF_HA	= 36,
	RELOC_ADDR30		= 37,
	RELOC_ADDR64		= 38,
	RELOC_ADDR16_HIGHER	= 39,
	RELOC_ADDR16_HIGHERA	= 40,
	RELOC_ADDR16_HIGHEST	= 41,
	RELOC_ADDR16_HIGHESTA	= 42,
	RELOC_UADDR64		= 43,
	RELOC_REL64		= 44,
	RELOC_PLT64		= 45,
	RELOC_PLTREL64		= 46,
	RELOC_TOC16		= 47,
	RELOC_TOC16_LO		= 48,
	RELOC_TOC16_HI		= 49,
	RELOC_TOC16_HA		= 50,
	RELOC_TOC		= 51,
	RELOC_PLTGOT16		= 52,
	RELOC_PLTGOT16_LO	= 53,
	RELOC_PLTGOT16_HI	= 54,
	RELOC_PLTGOT16_HA	= 55,
	RELOC_ADDR16_DS		= 56,
	RELOC_ADDR16_LO_DS	= 57,
	RELOC_GOT16_DS		= 58,
	RELOC_GOT16_LO_DS	= 59,
	RELOC_PLT16_LO_DS	= 60,
	RELOC_SECTOFF_DS	= 61,
	RELOC_SECTOFF_LO_DS	= 62,
	RELOC_TOC16_DS		= 63,
	RELOC_TOC16_LO_DS	= 64,
	RELOC_PLTGOT16_DS	= 65,
	RELOC_PLTGOT16_LO_DS	= 66,
	RELOC_TLS		= 67,
	RELOC_DTPMOD64		= 68,
	RELOC_TPREL16		= 69,
	RELOC_TPREL16_LO	= 70,
	RELOC_TPREL16_HI	= 71,
	RELOC_TPREL16_HA	= 72,
	RELOC_TPREL64		= 73,
	RELOC_DTPREL16		= 74,
	RELOC_DTPREL16_LO	= 75,
	RELOC_DTPREL16_HI	= 76,
	RELOC_DTPREL16_HA	= 77,
	RELOC_DTPREL64		= 78,
	RELOC_GOT_TLSGD16	= 79,
	RELOC_GOT_TLSGD16_LO	= 80,
	RELOC_GOT_TLSGD16_HI	= 81,
	RELOC_GOT_TLSGD16_HA	= 82,
	RELOC_GOT_TLSLD16	= 83,
	RELOC_GOT_TLSLD16_LO	= 84,
	RELOC_GOT_TLSLD16_HI	= 85,
	RELOC_GOT_TLSLD16_HA	= 86,
	RELOC_GOT_TPREL16_DS	= 87,
	RELOC_GOT_TPREL16_LO_DS	= 88,
	RELOC_GOT_TPREL16_HI	= 89,
	RELOC_GOT_TPREL16_HA	= 90,
	RELOC_GOT_DTPREL16_DS	= 91,
	RELOC_GOT_DTPREL16_LO_DS = 92,
	RELOC_GOT_DTPREL16_HI	= 93,
	RELOC_GOT_DTPREL16_HA	= 94,
	RELOC_TPREL16_DS	= 95,
	RELOC_TPREL16_LO_DS	= 96,
	RELOC_TPREL16_HIGHER	= 97,
	RELOC_TPREL16_HIGHERA	= 98,
	RELOC_TPREL16_HIGHEST	= 99,
	RELOC_TPREL16_HIGHESTA	= 100,
	RELOC_DTPREL16_DS	= 101,
	RELOC_DTPREL16_LO_DS	= 102,
	RELOC_DTPREL16_HIGHER	= 103,
	RELOC_DTPREL16_HIGHERA	= 104,
	RELOC_DTPREL16_HIGHEST	= 105,
	RELOC_DTPREL16_HIGHESTA	= 106,
	RELOC_MAX
};

#endif /* _POWERPC_RELOC_H_ */
