/*	$Id: cpu_970.c,v 1.6 2009/08/23 12:04:23 tsubai Exp $	*/

/*-
 * Copyright (c) 2004,2006 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_multiprocessor.h"

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/cpufunc.h>
#include <machine/spr.h>

/*
 * 970 specific part.
 */

#define HID0_NAP      0x0040000000000000LL	/* Nap mode */
#define HID0_DPM      0x0010000000000000LL
/* ... */

/*
 * Configure HID0 for 970.
 */
void
ppc970_config_hid()
{
	u_int64_t hid0, hid5;

	hid0 = mfspr(HID0);
	hid0 |= 0x80000000;	/* en_mck */

	asm volatile ("sync");
	mtspr(HID0, hid0);
	asm volatile ("isync");

	hid5 = mfspr(HID5);
	hid5 &= ~0x80;		/* DCBZ_size */

	asm volatile ("sync");
	mtspr(HID5, hid5);
	asm volatile ("isync");
}

/*
 * Print HID[0145] configuration.
 */
void
ppc970_print_hid(name)
	const char *name;
{
	u_int64_t hid0, hid1, hid4, hid5;
	char buf[256];

#define HID0_BITMASK "\177\020" \
  "b\077ONEPPC\0" \
  "b\076DOSINGLE\0" \
  "b\075ISYNCSC\0" \
  "b\074SERGP\0" \
  "b\066NAP\0" \
  "b\064DPM\0" \
  "b\062TG\0" \
  "b\061HANG_DIS\0" \
  "b\060NHR\0" \
  "b\057INORDER\0" \
  "b\055TBCTRL\0" \
  "b\054EXTB\0" \
  "b\051CIABR\0" \
  "b\050HDICE\0" \
  "b\047THERM\0" \
  "b\040ATTN\0" \
  "b\037MCK\0" \
  "\0"

	hid0 = mfspr(HID0);
	snprintb(buf, sizeof buf, HID0_BITMASK, hid0);
	printf("%s: HID0 %s\n", name, buf);

#define HID1_BITMASK "\177\020" \
  "f\075\003BHTPM\0" \
  "b\074LS\0" \
  "b\073CC\0" \
  "b\072IC\0" \
  "f\067\002PFMODE\0" \
  "b\066ICBI\0" \
  "b\065IFCACHE\0" \
  "b\064ICREC\0" \
  "b\063IDREC\0" \
  "b\062ERREC\0" \
  "b\061ICPE\0" \
  "b\060ICD0PE\0" \
  "b\056IERPE\0" \
  "b\055SPTW\0" \
  "\0"

	hid1 = mfspr(HID1);
	snprintb(buf, sizeof buf, HID1_BITMASK, hid1);
	printf("%s: HID1 %s\n", name, buf);

#define HID4_BITMASK "\177\020" \
  "f\051\020RMOR(0:15)\0" \
  "b\050RM_CI\0" \
  "b\047FORCE_AI\0" \
  "b\046DIS_PREF\0" \
  "b\045RES_PREF\0" \
  "b\044EN_SPTW\0" \
  "b\043DCFI\0" \
  "f\041\002DIS_DERPC\0" \
  "b\040DIS_DERPG\0" \
  "f\036\002DIS_DERAT\0" \
  "f\034\002DIS_DCTPC\0" \
  "b\033DIS_DCTPG\0" \
  "f\031\002DIS_DCSET\0" \
  "f\027\002DIS_DCPC\0" \
  "b\026DIS_DCPG\0" \
  "f\024\002DIS_DCRTPC\0" \
  "f\020\004DIS_TLBPC\0" \
  "b\017DIS_TLBPG\0" \
  "f\013\004DIS_TLBSET\0" \
  "b\012DIS_SLBPC\0" \
  "b\011DIS_SLBPG\0" \
  "b\010MCKINJ\0" \
  "b\007DIS_STFWD\0" \
  "f\006\001LPES(0)\0" \
  "f\077\001LPES(1)\0" \
  "f\005\001RMLR(0)\0" \
  "f\075\002RMLR(1:2)\0" \
  "b\003DIS_SPLARX\0" \
  "b\002DIS_LG\0" \
  "f\000\002LPID(0:1)\0" \
  "f\071\004LPID(2:5)\0" \
  "\0"

	hid4 = mfspr(HID4);
	snprintb(buf, sizeof buf, HID4_BITMASK, hid4);
	printf("%s: HID4 %s\n", name, buf);

#define HID4_LPES1	0x8000000000000000LL /* LPAR environment selector [1] */
#define HID4_LPES0	0x0000000000000040LL /* LPAR environment selector [0] */
#define HID4_RMLR12	0x6000000000000000LL /* LPAR real mode limit register */
#define HID4_RMLR0	0x0000000000000020LL /* LPAR real mode limit register */
#define HID4_LPID25	0x1e00000000000000LL /* LPAR partition identity [2-5] */
#define HID4_LPID01	0x0000000000000003LL /* LPAR partition identity [01] */
#define HID4_RMOR	0x01fffe0000000000LL /* LPAR real mode offset register */
#define HID4_RM_CI	0x0000010000000000LL /* Real mode cache inhibit */
#define HID4_FORCE_AI	0x0000008000000000LL /* Force ALI interrupt on unaligned op. */
#define HID4_DIS_PREF	0x0000004000000000LL /* Disable data prefetching */
#define HID4_RES_PREF	0x0000002000000000LL /* Reset data prefetching */
#define HID4_SPTW	0x0000001000000000LL /* Enable speculative load tablewalk */
#define HID4_DCFI	0x0000000800000000LL /* L1 data cache flush invalidate */
#define HID4_DIS_DEAPC	0x0000000600000000LL /* Disable D-ERAT parity checking */
#define HID4_DIS_DEAPG	0x0000000100000000LL /* Disable D-ERAT parity generation */
#define HID4_DIS_DERAT	0x00000000c0000000LL /* Disable D-ERAT */
#define HID4_DIS_DCTPC	0x0000000030000000LL /* Disable data cache tag parity checking */
#define HID4_DIS_DCTPG	0x0000000008000000LL /* Disable data cache tag parity generation */
#define HID4_DIS_DCSET	0x0000000006000000LL /* Disable data cache set */
#define HID4_DIS_DCPC	0x0000000001800000LL /* Disable data cache parity checking */
#define HID4_DIS_DCPG	0x0000000000400000LL /* Disable data cache parity generation */
#define HID4_DIS_DCRTPC	0x0000000000300000LL /* Disable data cache real address tag parity checking */
#define HID4_DIS_TLBPC	0x00000000000f0000LL /* Disable TLB parity checking */
#define HID4_DIS_TLBPG	0x0000000000008000LL /* Disable TLB parity generation */
#define HID4_DIS_TLBSET	0x0000000000007800LL /* Disable TLB set */
#define HID4_DIS_SLBPC	0x0000000000000400LL /* Disable SLB parity checking */
#define HID4_DIS_SLBPG	0x0000000000000200LL /* Disable SLB parity generation */
#define HID4_MCKINJ	0x0000000000000100LL /* Enable Machine check error inject */
#define HID4_DIS_STFWD	0x0000000000000080LL /* Disable store forwarding */
#define HID4_DIS_SPLARX	0x0000000000000008LL /* Disable speculative lwarx, ldarx */
#define HID4_DIS_LP	0x0000000000000004LL /* Disable large page */

#if 0
	u_long lpes, rmlr, lpid, rmor;

lpes = (hid4 & HID4_LPES0) >> 5 | (hid4 & HID4_LPES1) >> 63;
rmlr = (hid4 & HID4_RMLR0) >> 3 | (hid4 & HID4_RMLR12) >> 61;
lpid = (hid4 & HID4_LPID01) << 4 | (hid4 & HID4_LPID25) >> 57;
rmor = (hid4 & HID4_RMOR) >> 41;

printf("LPES = 0x%lx\n", lpes);
printf("RMLR = 0x%lx\n", rmlr);
printf("LPID = 0x%lx\n", lpid);
printf("RMOR = 0x%lx\n", rmor);

#endif

#define HID5_BITMASK "\177\020" \
  "f\020\20HRMOR(0:15)\0" \
  "b\015DCMCK\0" \
  "b\014DIS_PWSAVE\0" \
  "b\013FORCE_G\0" \
  "b\012DCRA\0" \
  "b\011HPS\0" \
  "b\010NOPDST\0" \
  "b\007DCBZ32\0" \
  "b\006DCBZ_ILL\0" \
  "b\005TLBMAP\0" \
  "b\004LMQ_PORT\0" \
  "f\003\001LMQ_SIZE(0)\0" \
  "f\000\001LMQ_SIZE(1)\0" \
  "b\001NOPTI\0" \
  "\0"

	hid5 = mfspr(HID5);
	snprintb(buf, sizeof buf, HID5_BITMASK, hid5);
	printf("%s: HID5 %s\n", name, buf);

#define HID5_HRMOR	0xffff0000 /* LPAR hypervisor real mode offset register */
#define HID5_DCMCK	0x00002000 /* Enable MCHK for D-cache and D-cache tag parity errors */
#define HID5_DIS_PWSAVE	0x00001000 /* Disable L1 D-cache, L1 D-cache tag, D-ERAT power savings */
#define HID5_FORCE_G    0x00000800 /* Force guarded (G=1) load */
#define HID5_DCRA       0x00000400 /* Data cache replacement algorithm */
#define HID5_HPS	0x00000200 /* Number of available hardware prefetch streams */
#define HID5_NOPDST	0x00000100 /* No-op the DST. */
#define HID5_DCBZ32	0x00000080 /* Makes dcbz a 32-byte store */
#define HID5_DCBZ_ILL	0x00000040 /* Makes dcbz an illegal instruction */
#define HID5_TLBMAP	0x00000020 /* TLB mapping */
#define HID5_LMQ_PORT	0x00000010 /* Demand miss (LMQ to STS) */
#define HID5_LMQ_SIZE0	0x00000008 /* # of outstanding requests to STS [0] */
#define HID5_NOPTI	0x00000002 /* No-op the dcbt(st) */
#define HID5_LMQ_SIZE1	0x00000001 /* # of outstanding requests to STS [1] */
}

void
ppc970_enable_nap()
{
	u_int64_t hid0;

	hid0 = mfspr(HID0);

	hid0 |= HID0_NAP;
	hid0 |= HID0_DPM;

	asm volatile ("sync");
	mtspr(HID0, hid0);
	asm volatile ("isync");

	/* sync and isync are needed; cf. UM 2.1.3.1. */
}

void
ppc970_enable_largepage()
{
	u_int64_t hid4;

	hid4 = mfspr(HID4);
	hid4 &= ~4LL;
	mtspr(HID4, hid4);
}

void
ppc970_rm_ci(mode)
	int mode;
{
	u_int64_t hid4;

	hid4 = mfspr(HID4);
	if (mode)
		hid4 |= HID4_RM_CI;
	else
		hid4 &= ~HID4_RM_CI;
	asm volatile ("sync");
	mtspr(HID4, hid4);
	asm volatile ("isync");
}

#if 0

HID0:
0     one_ppc     One PowerPC AS instruction per dispatch group mode
1     do_single   Single group completion mode
2     isync_sc    Disable isync scoreboard optimization
3     ser_gp      Serialize Group Dispatch
4-8               Reserved
9     nap         Nap
10                Reserved
11    dpm         Enable dynamic power management
12                Reserved
13    tg          Performance monitor threshold granularity control
14    dis_hang    Disable processor hang detection mechanism
15    nhr         Not Hard Reset
16    inorder     Serialized group issue mode
17                Reserved
18    tb_ctrl     Enable time base counting when processor is stopped
19    ext_tb      External time base enable
20-21             Reserved
22    ciabr       Enable CIABR
23    hdice       Enable hypervisor decrementer interrupt conditionally
24    therm       Enable external thermal interrupts
25-30             Reserved
31    attn        Enable support processor attention instruction
32    mck         Enable external machine check interrupts
33-63             Reserved

HID1:
0-2   bht_pm      BHT prediction mode
3     ls          Enable link stack
4     cc          Enable count cache
5     ic          Enable instruction cache
6                 Reserved
7-8   pf_mode     Prefetch mode
9     icbi        Enable "forced icbi match" mode
10    if_cache    Enable I-fetch cacheability control
11    ic_rec      Enable I-cache parity error recovery
12    id_rec      Enable I-directory parity error recovery
13    er_rec      Enable I-ERAT parity error recovery
14    ic_pe       Force I-cache parity error
15    icd0_pe     Force I-cache directory 0 parity error
16                Reserved
17    ier_pe      Force I-ERAT parity error (error inject)
18    sp_tw       Enable speculative tablewalks.
19-63             Reserved

#endif
