/*	$Id: exception.S,v 1.43 2009/12/25 03:17:52 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_ddb.h"
#include "opt_ipkdb.h"
#include "assym.h"

#include <sys/syscall.h>

#include <machine/param.h>
#include <machine/asm.h>
#include <machine/psl.h>
#include <machine/trap.h>

#ifdef NO_ASR
	.macro mfasr r
	GET_CPUINFO(\r)
	ld	\r,CI_CURPM(\r)
	ld	\r,PM_SEGREAL(\r)
	.endm
#endif

/*
 * This code gets copied to all the trap vectors
 * (except ISI/DSI, ALI, the interrupts, and possibly the debugging
 * traps when using IPKDB).
 */
	.text
	.globl	_C_LABEL(trapcode)
_C_LABEL(trapcode):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beql	1f			/* branch if supervisor */
	ld	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	bl	2f
1:
	mfsprg	1,1			/* restore SP */
2:
	mflr	31
	rlwinm	31,31,0,0,23		/* r31 &= ~0xff */
	mtlr	31
	ba	s_trap

/*
 * For ALI: has to save DSISR and DAR
 */
	.globl	_C_LABEL(alitrap)
_C_LABEL(alitrap):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mfdar	30
	mfdsisr	31
	std	30,CI_TEMPSAVE+32(1)
	std	31,CI_TEMPSAVE+40(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beq	1f			/* branch if supervisor */
	ld	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	li	31,EXC_ALI
	mtlr	31
	ba	s_trap
1:
	mfsprg	1,1			/* restore SP */
	li	31,EXC_ALI
	mtlr	31
	ba	s_trap

/*
 * Similar to the above for DSI
 * Has to handle standard pagetable spills
 */
	.globl	_C_LABEL(dsitrap)
_C_LABEL(dsitrap):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_DISISAVE(1)	/* free r28-r31 */
	std	29,CI_DISISAVE+8(1)
	std	30,CI_DISISAVE+16(1)
	std	31,CI_DISISAVE+24(1)
	mfcr	29			/* save CR */
	mflr	28			/* save LR */
	li	31,EXC_DSI
	mtlr	31
	ba	s_dsitrap

	.globl	_C_LABEL(dsegtrap)
_C_LABEL(dsegtrap):
	ba	s_dsegtrap
s_dsegtrap:
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	3,CI_SEGSAVE+0(1)
	std	4,CI_SEGSAVE+8(1)
	std	5,CI_SEGSAVE+16(1)
	std	6,CI_SEGSAVE+24(1)
	mfcr	6

	mfasr	3
	mfsrr1	4
	andi.	4,4,PSL_PR
	bne	1f			/* branch if user */
	lis	3,ksegtab@ha
	addi	3,3,ksegtab@l
1:	mfdar	4			/* fault EA */
	srdi	4,4,28			/* ESID */
	cmpdi	4,256
	bge	dsegmiss

	sldi	4,4,4
	add	3,3,4
	ld	4,0(3)			/* XXX we should use hash! XXX */
	ld	5,8(3)
	andis.	3,5,0x0800		/* valid? */
	beq	dsegmiss
	slbmte	4,5

	mtcr	6
	ld	3,CI_SEGSAVE+0(1)
	ld	4,CI_SEGSAVE+8(1)
	ld	5,CI_SEGSAVE+16(1)
	ld	6,CI_SEGSAVE+24(1)
	mfsprg	1,1			/* restore SP */
	rfid

dsegmiss:
	mtcr	6
	ld	3,CI_SEGSAVE+0(1)
	ld	4,CI_SEGSAVE+8(1)
	ld	5,CI_SEGSAVE+16(1)
	ld	6,CI_SEGSAVE+24(1)

	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mfdar	30
	mfdsisr	31
	std	30,CI_TEMPSAVE+32(1)
	std	31,CI_TEMPSAVE+40(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beq	1f			/* branch if supervisor */
	ld	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	li	31,EXC_DSEG
	mtlr	31
	ba	s_trap
1:
	mfsprg	1,1			/* restore SP */
	li	31,EXC_DSEG
	mtlr	31
	ba	s_trap

isegmiss:
	mtcr	6
	ld	3,CI_SEGSAVE+0(1)
	ld	4,CI_SEGSAVE+8(1)
	ld	5,CI_SEGSAVE+16(1)
	ld	6,CI_SEGSAVE+24(1)

	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beq	1f			/* branch if supervisor */
	ld	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	li	31,EXC_ISEG
	mtlr	31
	ba	s_trap
1:
	mfsprg	1,1			/* restore SP */
	li	31,EXC_ISEG
	mtlr	31
	ba	s_trap

/*
 * Similar to the above for ISI
 */
	.globl	_C_LABEL(isitrap)
_C_LABEL(isitrap):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_DISISAVE(1)	/* free r28-r31 */
	std	29,CI_DISISAVE+8(1)
	std	30,CI_DISISAVE+16(1)
	std	31,CI_DISISAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	li	31,EXC_ISI
	mtlr	31
	ba	s_isitrap

	.globl	_C_LABEL(isegtrap)
_C_LABEL(isegtrap):
	ba	s_isegtrap
s_isegtrap:
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	3,CI_SEGSAVE+0(1)
	std	4,CI_SEGSAVE+8(1)
	std	5,CI_SEGSAVE+16(1)
	std	6,CI_SEGSAVE+24(1)
	mfcr	6

	mfasr	3
	mfsrr1	4
	andi.	4,4,PSL_PR
	bne	1f			/* branch if user */
	lis	3,ksegtab@ha
	addi	3,3,ksegtab@l
1:	mfsrr0	4			/* fault EA */
	srdi	4,4,28			/* ESID */
	cmpdi	4,256
	bge	isegmiss

	sldi	4,4,4
	add	3,3,4
	ld	4,0(3)
	ld	5,8(3)
	andis.	3,5,0x0800		/* valid? */
	beq	isegmiss
	slbmte	4,5

	mtcr	6
	ld	3,CI_SEGSAVE+0(1)
	ld	4,CI_SEGSAVE+8(1)
	ld	5,CI_SEGSAVE+16(1)
	ld	6,CI_SEGSAVE+24(1)
	mfsprg	1,1			/* restore SP */
	rfid

/*
 * This one for the external interrupt handler.
 */
	.globl	_C_LABEL(extint)
_C_LABEL(extint):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	mfxer	30			/* save XER */
	lwa	31,CI_INTRDEPTH(1)	/* were we already running on intstk? */
	addic.	31,31,1
	stw	31,CI_INTRDEPTH(1)
	beq+	1f
	mfsprg	1,1			/* yes, get old SP */
	ba	extintr
1:
	ld	1,CI_INTSTK(1)		/* get interrupt stack */
	ba	extintr

/*
 * And this one for the decrementer interrupt handler.
 */
	.globl	_C_LABEL(decrint)
_C_LABEL(decrint):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	mfxer	30			/* save XER */
	lwa	31,CI_INTRDEPTH(1)	/* were we already running on intstk? */
	addic.	31,31,1
	stw	31,CI_INTRDEPTH(1)
	beq+	1f
	mfsprg	1,1			/* yes, get old SP */
	ba	decrintr
1:
	ld	1,CI_INTSTK(1)		/* get interrupt stack */
	ba	decrintr

/*
 * Performance Monitor Exception (0x0f00) and
 * AltiVec Unavailable Exception (0x0f20) handler.
 */
	.globl	_C_LABEL(perftrap)
_C_LABEL(perftrap):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	li	31,EXC_PERF
	b	2f

	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	li	31,EXC_VEC
2:
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	mtlr	31
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beq	1f			/* branch if supervisor */
	ld	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	ba	s_trap
1:
	mfsprg	1,1			/* restore SP */
	ba	s_trap

#ifdef DDB
/*
 * In case of DDB we want a separate trap catcher for it
 */
	.globl	_C_LABEL(ddblow)
_C_LABEL(ddblow):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_DDBSAVE(1)	/* free r28-r31 */
	std	29,CI_DDBSAVE+8(1)
	std	30,CI_DDBSAVE+16(1)
	std	31,CI_DDBSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	ld	1,CI_DDBSTK(1)		/* get new SP */
	li	31,0
	stdu	31,-48(1)		/* end of stack chain */
	bl	1f
1:
	mflr	31
	rlwinm	31,31,0,0,23		/* r31 &= ~0xff */
	mtlr	31
	ba	ddbtrap
#endif	/* DDB */

#ifdef IPKDB
#define CI_IPKDBSAVE CI_DDBSAVE
#define CI_IPKDBSTK  CI_DDBSTK

/*
 * In case of IPKDB we want a separate trap catcher for it
 */

	.globl	_C_LABEL(ipkdblow)
_C_LABEL(ipkdblow):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_IPKDBSAVE(1)	/* free r28-r31 */
	std	29,CI_IPKDBSAVE+8(1)
	std	30,CI_IPKDBSAVE+16(1)
	std	31,CI_IPKDBSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	ld	1,CI_IPKDBSTK(1)	/* get new SP */
	li	31,0
	stdu	31,-48(1)		/* end of stack chain */
	bl	1f
1:
	mflr	31
	rlwinm	31,31,0,0,23		/* r31 &= ~0xff */
	mtlr	31
	ba	ipkdbtrap
#endif	/* IPKDB */

/*
 * FRAME_SETUP assumes:
 *	SPRG1		SP (1)
 *	savearea	r28-r31,DAR,DSISR	(DAR & DSISR only for DSI traps)
 *	28		LR
 *	29		CR
 *	1		kernel stack
 *	LR		trap type
 *	SRR0/1		as at start of trap
 */
#define FRAME_SETUP(savearea)						\
/* Have to enable translation to allow access of kernel stack: */	\
	slbia;								\
	GET_CPUINFO(30);						\
	mfsrr0	31;							\
	std	31,savearea+48(30);					\
	mfsrr1	31;							\
	std	31,savearea+56(30);					\
	mfsprg	30,1;							\
	mfmsr	31;							\
	ori	31,31,(PSL_DR|PSL_IR);					\
	mtmsr	31;							\
	isync;								\
	stdu	30,-FRAMELEN(1);					\
	std	0,FRAME_0+112(1);					\
	std	30,FRAME_1+112(1);					\
	std	2,FRAME_2+112(1);					\
	std	28,FRAME_LR+112(1);					\
	std	29,FRAME_CR+112(1);					\
	GET_CPUINFO(2);							\
	ld	28,savearea(2);						\
	ld	29,savearea+8(2);					\
	ld	30,savearea+16(2);					\
	ld	31,savearea+24(2);					\
	std	3,FRAME_3+112(1);					\
	std	4,FRAME_4+112(1);					\
	std	5,FRAME_5+112(1);					\
	std	6,FRAME_6+112(1);					\
	std	7,FRAME_7+112(1);					\
	std	8,FRAME_8+112(1);					\
	std	9,FRAME_9+112(1);					\
	std	10,FRAME_10+112(1);					\
	std	11,FRAME_11+112(1);					\
	std	12,FRAME_12+112(1);					\
	std	13,FRAME_13+112(1);					\
	std	14,FRAME_14+112(1);					\
	std	15,FRAME_15+112(1);					\
	std	16,FRAME_16+112(1);					\
	std	17,FRAME_17+112(1);					\
	std	18,FRAME_18+112(1);					\
	std	19,FRAME_19+112(1);					\
	std	20,FRAME_20+112(1);					\
	std	21,FRAME_21+112(1);					\
	std	22,FRAME_22+112(1);					\
	std	23,FRAME_23+112(1);					\
	std	24,FRAME_24+112(1);					\
	std	25,FRAME_25+112(1);					\
	std	26,FRAME_26+112(1);					\
	std	27,FRAME_27+112(1);					\
	std	28,FRAME_28+112(1);					\
	std	29,FRAME_29+112(1);					\
	std	30,FRAME_30+112(1);					\
	std	31,FRAME_31+112(1);					\
	ld	28,savearea+32(2);					\
	ld	29,savearea+40(2);					\
	ld	30,savearea+48(2);					\
	ld	31,savearea+56(2);					\
	mfxer	3;							\
	mfctr	4;							\
	mflr	5;							\
	std	3,FRAME_XER+112(1);					\
	std	4,FRAME_CTR+112(1);					\
	stw	5,FRAME_EXC+112(1);	/* 32 bit */			\
	std	28,FRAME_DAR+112(1);					\
	stw	29,FRAME_DSISR+112(1);	/* 32 bit */			\
	std	30,FRAME_SRR0+112(1);					\
	std	31,FRAME_SRR1+112(1)

#define FRAME_LEAVE(savearea)						\
/* Now restore regs: */							\
	ld	3,FRAME_SRR0+112(1);					\
	ld	4,FRAME_SRR1+112(1);					\
	ld	5,FRAME_CTR+112(1);					\
	ld	6,FRAME_XER+112(1);					\
	ld	7,FRAME_LR+112(1);					\
	GET_CPUINFO(2);							\
	std	3,savearea+48(2);					\
	std	4,savearea+56(2);					\
	mtctr	5;							\
	mtxer	6;							\
	mtlr	7;							\
	ld	3,FRAME_3+112(1);					\
	ld	4,FRAME_4+112(1);					\
	ld	5,FRAME_5+112(1);					\
	ld	6,FRAME_6+112(1);					\
	ld	7,FRAME_7+112(1);					\
	ld	8,FRAME_8+112(1);					\
	ld	9,FRAME_9+112(1);					\
	ld	10,FRAME_10+112(1);					\
	ld	11,FRAME_11+112(1);					\
	ld	12,FRAME_12+112(1);					\
	ld	13,FRAME_13+112(1);					\
	ld	14,FRAME_14+112(1);					\
	ld	15,FRAME_15+112(1);					\
	ld	16,FRAME_16+112(1);					\
	ld	17,FRAME_17+112(1);					\
	ld	18,FRAME_18+112(1);					\
	ld	19,FRAME_19+112(1);					\
	ld	20,FRAME_20+112(1);					\
	ld	21,FRAME_21+112(1);					\
	ld	22,FRAME_22+112(1);					\
	ld	23,FRAME_23+112(1);					\
	ld	24,FRAME_24+112(1);					\
	ld	25,FRAME_25+112(1);					\
	ld	26,FRAME_26+112(1);					\
	ld	27,FRAME_27+112(1);					\
	ld	28,FRAME_28+112(1);					\
	ld	29,FRAME_29+112(1);					\
	ld	30,FRAME_30+112(1);					\
	ld	31,FRAME_31+112(1);					\
	std	28,savearea(2);						\
	std	29,savearea+8(2);					\
	std	31,savearea+24(2);					\
	ld	28,FRAME_2+112(1);	/* save r2 in r28 */		\
	ld	29,FRAME_CR+112(1);	/* save cr in r29 */		\
	ld	0,FRAME_0+112(1);					\
	ld	1,FRAME_1+112(1);					\
/* Disable translation, machine check and recoverability: */		\
	mfmsr	31;							\
	andi.	31,31,~(PSL_DR|PSL_IR|PSL_ME|PSL_RI)@l;			\
	mtmsr	31;							\
	isync;								\
	mtsprg	2,28;			/* save r2 */			\
	ld	31,savearea+48(2);					\
	mtsrr0	31;							\
	ld	31,savearea+56(2);					\
	mtsrr1	31;							\
/* Decide whether we return to user mode: */				\
	andi.	31,31,PSL_PR;						\
	beq	1f;			/* branch if supervisor */	\
/* Restore user & kernel access SR: */					\
	slbia;								\
1:	mtcr	29;			/* restore cr */		\
	ld	28,savearea(2);						\
	ld	29,savearea+8(2);					\
	ld	31,savearea+24(2);					\
	mfsprg	2,2			/* restore r2 */

/*
 * Preamble code for DSI/ISI traps
 */
disitrap:
	ld	30,CI_DISISAVE(1)
	ld	31,CI_DISISAVE+8(1)
	std	30,CI_TEMPSAVE(1)
	std	31,CI_TEMPSAVE+8(1)
	ld	30,CI_DISISAVE+16(1)
	ld	31,CI_DISISAVE+24(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mfdar	30
	mfdsisr	31
	std	30,CI_TEMPSAVE+32(1)
	std	31,CI_TEMPSAVE+40(1)
realtrap:
/* Test whether we already had PR set */
	mfsrr1	1
	andi.	1,1,PSL_PR
	mfsprg	1,1			/* restore SP (might have been
					   overwritten) */
	beq	s_trap			/* branch if supervisor */
	GET_CPUINFO(1)
	ld	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */

/*
 * Now the common trap catching code.
 */
s_trap:
	FRAME_SETUP(CI_TEMPSAVE)
/* Now we can recover interrupts again: */
trapagain:
	mfmsr	7
	ori	7,7,(PSL_EE|PSL_ME|PSL_RI)@l
	mtmsr	7
	isync
/* Call C trap code: */
	lis	3,tocbase@ha
	ld	2,tocbase@l(3)
	addi	3,1,112
	bl	_C_LABEL(trap)
	nop
trapexit:
/* Disable interrupts: */
	mfmsr	3
	andi.	3,3,~PSL_EE@l
	mtmsr	3
/* Test AST pending: */
	ld	5,FRAME_SRR1+112(1)
	andi.	5,5,PSL_PR
	beq	1f			/* branch if supervisor */
	GET_CPUINFO(3)
	lwz	4,CI_ASTPENDING(3)
	andi.	4,4,1
	beq	1f
	li	6,EXC_AST
	stw	6,FRAME_EXC+112(1)
	b	trapagain
1:
	FRAME_LEAVE(CI_TEMPSAVE)
	rfid

	.align	3
tocbase:
	.llong	.TOC.@tocbase

/*
 * Child comes here at the end of a fork.
 * Return to userspace via the trap return path.
 * XXX r3 and r4 are passed from cpu_switchto() unchanged.
 */
	.globl	_C_LABEL(fork_trampoline)
_C_LABEL(fork_trampoline):
	lis	5,tocbase@ha
	ld	2,tocbase@l(5)
	bl	_C_LABEL(lwp_startup)
	nop
	ld	0,0(31)			/* load function address from FDESC */
	ld	2,8(31)
	mtlr	0
	mr	3,30
	blrl				/* jump indirect to r31 */
	b	trapexit

	.globl	_C_LABEL(setfunc_trampoline)
_C_LABEL(setfunc_trampoline):
	ld	0,0(31)			/* load function address from FDESC */
	ld	2,8(31)
	mtlr	0
	mr	3,30
	blrl				/* jump indirect to r31 */
	b	trapexit

/*
 * DSI second stage fault handler
 */
s_dsitrap:
	mfdsisr	31			/* test whether this may be a
					   spill fault */
	andis.	31,31,0x40000000@h
	beq	disitrap		/* branch if table miss is false */
	ld	1,CI_SPILLSTK(1)	/* get spill stack */
	stdu	1,-256(1)
	std	0,112(1)		/* save volatile registers */
	std	2,120(1)
	std	3,128(1)
	std	4,136(1)
	std	5,144(1)
	std	6,152(1)
	std	7,160(1)
	std	8,168(1)
	std	9,176(1)
	std	10,184(1)
	std	11,192(1)
	std	12,200(1)
	mfxer	30			/* save XER */
	mfctr	31			/* & CTR */
	mfdar	3
	lis	4,tocbase@ha
	ld	2,tocbase@l(4)
	bl	_C_LABEL(pte_spill)	/* try a spill */
	nop
	li	4,EXC_DSI
s_pte_spill:
	cmpwi	3,0
	mtlr	4			/* trap type */
	mtxer	30			/* restore XER */
	mtctr	31			/* restore CTR */
	ld	0,112(1)		/* restore volatile registers */
	ld	2,120(1)
	ld	3,128(1)
	ld	4,136(1)
	ld	5,144(1)
	ld	6,152(1)
	ld	7,160(1)
	ld	8,168(1)
	ld	9,176(1)
	ld	10,184(1)
	ld	11,192(1)
	ld	12,200(1)
	GET_CPUINFO(1)
	beq	disitrap
	mtcr	29			/* restore CR */
	mtlr	28			/* restore LR */
	ld	28,CI_DISISAVE(1)	/* restore r28-r31 */
	ld	29,CI_DISISAVE+8(1)
	ld	30,CI_DISISAVE+16(1)
	ld	31,CI_DISISAVE+24(1)
	mfsprg	1,1			/* restore SP */
	rfid				/* return to trapped code */

/*
 * ISI second stage fault handler
 */
s_isitrap:
	mfsrr1	31			/* test whether this may be a
					   spill fault */
	andis.	31,31,0x40000000@h
	beq	disitrap		/* branch if table miss is false */
	ld	1,CI_SPILLSTK(1)	/* get spill stack */
	stdu	1,-256(1)
	std	0,112(1)		/* save volatile registers */
	std	2,120(1)
	std	3,128(1)
	std	4,136(1)
	std	5,144(1)
	std	6,152(1)
	std	7,160(1)
	std	8,168(1)
	std	9,176(1)
	std	10,184(1)
	std	11,192(1)
	std	12,200(1)
	mfxer	30			/* save XER */
	mfctr	31			/* & ctr */
	mfsrr0	3
	lis	4,tocbase@ha
	ld	2,tocbase@l(4)
	bl	_C_LABEL(pte_spill)	/* try a spill */
	nop
	li	4,EXC_ISI
	b	s_pte_spill		/* above */

#define INTRFRAME_srr0	(112+8*0)
#define INTRFRAME_srr1	(112+8*1)
#define INTRFRAME_pri	(112+8*2)
#define INTRFRAME_depth	(112+8*2+4)	/* ޤ clockframe */
#define INTRFRAME_r0	(112+8*3)
#define INTRFRAME_r1	0		/* original SP */
#define INTRFRAME_r2	(112+8*4)
#define INTRFRAME_r3	(112+8*5)
#define INTRFRAME_r4	(112+8*6)
#define INTRFRAME_r5	(112+8*7)
#define INTRFRAME_r6	(112+8*8)
#define INTRFRAME_r7	(112+8*9)
#define INTRFRAME_r8	(112+8*10)
#define INTRFRAME_r9	(112+8*11)
#define INTRFRAME_r10	(112+8*12)
#define INTRFRAME_r11	(112+8*13)
#define INTRFRAME_r12	(112+8*14)
#define INTRFRAME_lr	(112+8*15)
#define INTRFRAME_ctr	(112+8*16)
#define INTRFRAME_cr	(112+8*17)
#define INTRFRAME_xer	(112+8*18)

/*
 * External interrupt second level handler
 */
#define INTRENTER							\
/* Save volatile registers: */						\
	stdu	1,-480(1);		/* temporarily */		\
	std	0,INTRFRAME_r0(1);					\
	mfsprg	0,1;			/* get original SP */		\
	std	0,INTRFRAME_r1(1);	/* and store it */		\
	std	2,INTRFRAME_r2(1);					\
	std	3,INTRFRAME_r3(1);					\
	std	4,INTRFRAME_r4(1);					\
	std	5,INTRFRAME_r5(1);					\
	std	6,INTRFRAME_r6(1);					\
	std	7,INTRFRAME_r7(1);					\
	std	8,INTRFRAME_r8(1);					\
	std	9,INTRFRAME_r9(1);					\
	std	10,INTRFRAME_r10(1);					\
	std	11,INTRFRAME_r11(1);					\
	std	12,INTRFRAME_r12(1);					\
	std	28,INTRFRAME_lr(1);	/* saved LR */			\
	std	29,INTRFRAME_cr(1);	/* saved CR */			\
	std	30,INTRFRAME_xer(1);	/* saved XER */			\
	GET_CPUINFO(5);							\
	ld	28,CI_TEMPSAVE(5);	/* restore r28-r31 */		\
	ld	29,CI_TEMPSAVE+8(5);					\
	ld	30,CI_TEMPSAVE+16(5);					\
	ld	31,CI_TEMPSAVE+24(5);					\
	mfctr	6;							\
	lwa	5,CI_INTRDEPTH(5);					\
	mfsrr0	4;							\
	mfsrr1	3;							\
	std	6,INTRFRAME_ctr(1);					\
	stw	5,INTRFRAME_depth(1);					\
	std	4,INTRFRAME_srr0(1);					\
	std	3,INTRFRAME_srr1(1);					\
/* interrupts are recoverable here, and enable translation */		\
	andi.	0,3,PSL_PR;						\
	beq	1f;			/* branch if supervisor */	\
	slbia;								\
1:	mfmsr	5;							\
	ori	5,5,(PSL_IR|PSL_DR|PSL_RI);				\
	mtmsr	5;							\
	isync

	.globl	_C_LABEL(extint_call)
extintr:
	INTRENTER
_C_LABEL(extint_call):
	lis	2,0
	ori	2,2,0
	bl	_C_LABEL(extint_call)	/* to be filled in later */

intr_exit:
/* Disable interrupts (should already be disabled) and MMU here: */
	mfmsr	3
	andi.	3,3,~(PSL_EE|PSL_ME|PSL_RI|PSL_DR|PSL_IR)@l
	sync
	mtmsr	3
	isync
/* restore possibly overwritten registers: */
	ld	12,INTRFRAME_r12(1)
	ld	11,INTRFRAME_r11(1)
	ld	10,INTRFRAME_r10(1)
	ld	9,INTRFRAME_r9(1)
	ld	8,INTRFRAME_r8(1)
	ld	7,INTRFRAME_r7(1)
	ld	6,INTRFRAME_srr1(1)
	ld	5,INTRFRAME_srr0(1)
	ld	4,INTRFRAME_ctr(1)
	ld	3,INTRFRAME_xer(1)
	mtsrr1	6
	mtsrr0	5
	mtctr	4
	mtxer	3
/* Returning to user mode? */
	andi.	6,6,PSL_PR		/* saved SRR1 */
	beq	1f			/* branch if supervisor */
	slbia
	GET_CPUINFO(4)
	lwz	6,CI_ASTPENDING(4)	/* Test AST pending */
	andi.	6,6,1
	beq	1f
/* Setup for entry to realtrap: */
	ld	3,INTRFRAME_r1(1)	/* get saved SP */
	mtsprg	1,3
	li	6,EXC_AST
	std	28,CI_TEMPSAVE(4)		/* establish tempsave again */
	std	29,CI_TEMPSAVE+8(4)
	std	30,CI_TEMPSAVE+16(4)
	std	31,CI_TEMPSAVE+24(4)
	mtlr	6
	ld	28,INTRFRAME_lr(1)	/* saved LR */
	ld	29,INTRFRAME_cr(1)	/* saved CR */
	ld	6,INTRFRAME_r6(1)
	ld	5,INTRFRAME_r5(1)
	ld	4,INTRFRAME_r4(1)
	ld	3,INTRFRAME_r3(1)
	ld	2,INTRFRAME_r2(1)
	ld	0,INTRFRAME_r0(1)
	GET_CPUINFO(30)
	lwa	31,CI_INTRDEPTH(30)	/* adjust reentrancy count */
	addi	31,31,-1
	stw	31,CI_INTRDEPTH(30)
	b	realtrap
1:
/* Here is the normal exit of extintr: */
	ld	5,INTRFRAME_cr(1)
	ld	6,INTRFRAME_lr(1)
	mtcr	5
	mtlr	6
	ld	6,INTRFRAME_r6(1)
	ld	5,INTRFRAME_r5(1)
	GET_CPUINFO(3)
	lwa	4,CI_INTRDEPTH(3)	/* adjust reentrancy count */
	addi	4,4,-1
	stw	4,CI_INTRDEPTH(3)
	ld	4,INTRFRAME_r4(1)
	ld	3,INTRFRAME_r3(1)
	ld	2,INTRFRAME_r2(1)
	ld	0,INTRFRAME_r0(1)
	ld	1,INTRFRAME_r1(1)
	rfid

/*
 * Decrementer interrupt second level handler
 */
decrintr:
	INTRENTER
	lis	3,tocbase@ha
	ld	2,tocbase@l(3)
	addi	3,1,INTRFRAME_srr0	/* clockframe */
	bl	_C_LABEL(decr_intr)
	nop
	b	intr_exit

#ifdef DDB
/*
 * Deliberate entry to ddbtrap
 */
FDESC(ddb_trap)

ENTRY_NOPROFILE(ddb_trap)
	mtsprg	1,1
	mfmsr	3
	mtsrr1	3
	andi.	3,3,~(PSL_EE|PSL_ME)@l
	mtmsr	3			/* disable interrupts */
	isync
	GET_CPUINFO(3)
	std	28,CI_DDBSAVE(3)
	std	29,CI_DDBSAVE+8(3)
	std	30,CI_DDBSAVE+16(3)
	std	31,CI_DDBSAVE+24(3)
	mflr	28
	li	29,EXC_BPT
	mtlr	29
	mfcr	29
	mtsrr0	28

/*
 * Now the ddb trap catching code.
 */
ddbtrap:
	FRAME_SETUP(CI_DDBSAVE)
/* Call C trap code: */
	lis	3,tocbase@ha
	ld	2,tocbase@l(3)
	addi	3,1,112
	bl	_C_LABEL(ddb_trap_glue)
	nop
	cmpwi	3,0
	bne	ddbleave
/* This wasn't for DDB, so switch to real trap: */
	GET_CPUINFO(4)
	lwz	3,FRAME_EXC+112(1)	/* save exception */
	std	3,CI_DDBSAVE+32(4)
	FRAME_LEAVE(CI_DDBSAVE)
	mtsprg	1,1			/* prepare for entrance to realtrap */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28
	mfcr	29
	ld	31,CI_DDBSAVE+32(1)
	mtlr	31
	b	realtrap
ddbleave:
	FRAME_LEAVE(CI_DDBSAVE)
	rfid
#endif /* DDB */

#ifdef IPKDB
/*
 * Deliberate entry to ipkdbtrap
 */
FDESC(ipkdb_trap)

ENTRY_NOPROFILE(ipkdb_trap)
	mtsprg	1,1
	mfmsr	3
	mtsrr1	3
	andi.	3,3,~(PSL_EE|PSL_ME)@l
	mtmsr	3			/* disable interrupts */
	isync
	GET_CPUINFO(3)
	std	28,CI_IPKDBSAVE(3)
	std	29,CI_IPKDBSAVE+8(3)
	std	30,CI_IPKDBSAVE+16(3)
	std	31,CI_IPKDBSAVE+24(3)
	mflr	28
	li	29,EXC_BPT
	mtlr	29
	mfcr	29
	mtsrr0	28

/*
 * Now the ipkdb trap catching code.
 */
ipkdbtrap:
	FRAME_SETUP(CI_IPKDBSAVE)
/* Call C trap code: */
	lis	3,tocbase@ha
	ld	2,tocbase@l(3)
	addi	3,1,112
	bl	_C_LABEL(ipkdb_trap_glue)
	nop
	cmpwi	3,0
	bne	ipkdbleave
/* This wasn't for IPKDB, so switch to real trap: */
	GET_CPUINFO(4)
	lwz	3,FRAME_EXC+112(1)	/* save exception */
	std	3,CI_IPKDBSAVE+32(4)
	FRAME_LEAVE(CI_IPKDBSAVE)
	mtsprg	1,1			/* prepare for entrance to realtrap */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28
	mfcr	29
	ld	31,CI_IPKDBSAVE+32(1)
	mtlr	31
	b	realtrap
ipkdbleave:
	FRAME_LEAVE(CI_IPKDBSAVE)
	rfi

ipkdbfault:
	ba	_ipkdbfault
_ipkdbfault:
	mfsrr0	3
	addi	3,3,4
	mtsrr0	3
	li	3,-1
	rfi

/*
 * int ipkdbfbyte(unsigned char *p)
 */
	.globl	_C_LABEL(ipkdbfbyte)
_C_LABEL(ipkdbfbyte):
	li	9,EXC_DSI		/* establish new fault routine */
	lwz	5,0(9)
	lis	6,ipkdbfault@ha
	lwz	6,ipkdbfault@l(6)
	stw	6,0(9)
#ifdef IPKDBUSERHACK
	/* XXX see powerpc/trap_subr.S */
	li	8,USER_SR
	sldi	8,8,ADDR_SR_SHFT
	slbie	8
	isync
#endif
	dcbst	0,9			/* flush data... */
	sync
	icbi	0,9			/* and instruction caches */
	lbz	3,0(3)			/* fetch data */
	stw	5,0(9)			/* restore previous fault handler */
	dcbst	0,9			/* and flush data... */
	sync
	icbi	0,9			/* and instruction caches */
	blr

/*
 * int ipkdbsbyte(unsigned char *p, int c)
 */
	.globl	_C_LABEL(ipkdbsbyte)
_C_LABEL(ipkdbsbyte):
	li	9,EXC_DSI		/* establish new fault routine */
	lwz	5,0(9)
	lis	6,ipkdbfault@ha
	lwz	6,ipkdbfault@l(6)
	stw	6,0(9)
#ifdef IPKDBUSERHACK
	/* XXX see powerpc/trap_subr.S */
	li	8,USER_SR
	sldi	8,8,ADDR_SR_SHFT
	slbie	8
	isync
#endif
	dcbst	0,9			/* flush data... */
	sync
	icbi	0,9			/* and instruction caches */
	mr	6,3
	li	3,0
	stb	4,0(6)
	dcbst	0,6			/* Now do appropriate flushes
					   to data... */
	sync
	icbi	0,6			/* and instruction caches */
	stw	5,0(9)			/* restore previous fault handler */
	dcbst	0,9			/* and flush data... */
	sync
	icbi	0,9			/* and instruction caches */
	blr
#endif	/* IPKDB */

	.globl	_C_LABEL(syscalltrap)
_C_LABEL(syscalltrap):
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	ld	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	li	31,EXC_SC
	mtlr	31
	ba	1f
1:
	FRAME_SETUP(CI_TEMPSAVE)
/* Now we can recover interrupts again: */
	mfmsr	7
	ori	7,7,(PSL_EE|PSL_ME|PSL_RI)@l
	mtmsrd	7
	isync
	GET_CPUINFO(3)
	ld	3,CI_CURLWP(3)
	ld	3,L_PROC(3)
	ld	4,P_MD_SYSCALL(3)	/* FDESC */
	ld	0,0(4)			/* function address */
	ld	2,8(4)
	mtlr	0
	addi	3,1,112
	blrl
	b	trapexit
