/*	$Id: exception32.S,v 1.23 2009/12/25 03:17:52 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_ddb.h"
#include "opt_ipkdb.h"
#include "assym.h"

#include <sys/syscall.h>

#include <machine/param.h>
#include <machine/asm.h>
#include <machine/psl.h>
#include <machine/trap.h>

#ifdef NO_ASR
	.macro mfasr r
	GET_CPUINFO(\r)
	ld	\r,CI_CURPM(\r)
	ld	\r,PM_SEGREAL(\r)
	.endm
#endif

#define NENTRY(x)	.globl _C_LABEL(x); _C_LABEL(x):

/*
 * This code gets copied to all the trap vectors
 * (except ISI/DSI, ALI, the interrupts, and possibly the debugging
 * traps when using IPKDB).
 */
	.text
NENTRY(trapcode)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beql	1f			/* branch if supervisor */
	lwz	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	bl	2f
1:
	mfsprg	1,1			/* restore SP */
2:
	mflr	31
	rlwinm	31,31,0,0,23		/* r31 &= ~0xff */
	mtlr	31
	ba	s_trap

/*
 * For ALI: has to save DSISR and DAR
 */
NENTRY(alitrap)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mfdar	30
	mfdsisr	31
	std	30,CI_TEMPSAVE+32(1)
	std	31,CI_TEMPSAVE+40(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beq	1f			/* branch if supervisor */
	lwz	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	li	31,EXC_ALI
	mtlr	31
	ba	s_trap
1:
	mfsprg	1,1			/* restore SP */
	li	31,EXC_ALI
	mtlr	31
	ba	s_trap

/*
 * Similar to the above for DSI
 * Has to handle standard pagetable spills
 */
NENTRY(dsitrap)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_DISISAVE(1)	/* free r28-r31 */
	std	29,CI_DISISAVE+8(1)
	std	30,CI_DISISAVE+16(1)
	std	31,CI_DISISAVE+24(1)
	mfcr	29			/* save CR */
	mflr	28			/* save LR */
	li	31,EXC_DSI
	mtlr	31
	ba	s_dsitrap

NENTRY(dsegtrap)
	ba	s_dsegtrap
s_dsegtrap:
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	3,CI_SEGSAVE+0(1)
	std	4,CI_SEGSAVE+8(1)
	std	5,CI_SEGSAVE+16(1)
	std	6,CI_SEGSAVE+24(1)
	mfcr	6

	mfasr	3
	mfsrr1	4
	andi.	4,4,PSL_PR
	bne	1f			/* branch if user */
	lis	3,ksegtab@ha
	addi	3,3,ksegtab@l
1:	mfdar	4			/* fault EA */
	srdi	4,4,28			/* ESID */

	sldi	4,4,4
	add	3,3,4
	ld	4,0(3)			/* XXX we should use hash! XXX */
	ld	5,8(3)
	andis.	3,5,0x0800		/* valid? */
	beq	dsegmiss
	slbmte	4,5

	mtcr	6
	ld	3,CI_SEGSAVE+0(1)
	ld	4,CI_SEGSAVE+8(1)
	ld	5,CI_SEGSAVE+16(1)
	ld	6,CI_SEGSAVE+24(1)
	mfsprg	1,1			/* restore SP */
	rfid

dsegmiss:
	mtcr	6
	ld	3,CI_SEGSAVE+0(1)
	ld	4,CI_SEGSAVE+8(1)
	ld	5,CI_SEGSAVE+16(1)
	ld	6,CI_SEGSAVE+24(1)

	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mfdar	30
	mfdsisr	31
	std	30,CI_TEMPSAVE+32(1)
	std	31,CI_TEMPSAVE+40(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beq	1f			/* branch if supervisor */
	lwz	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	li	31,EXC_DSEG
	mtlr	31
	ba	s_trap
1:
	mfsprg	1,1			/* restore SP */
	li	31,EXC_DSEG
	mtlr	31
	ba	s_trap

isegmiss:
	mtcr	6
	ld	3,CI_SEGSAVE+0(1)
	ld	4,CI_SEGSAVE+8(1)
	ld	5,CI_SEGSAVE+16(1)
	ld	6,CI_SEGSAVE+24(1)

	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beq	1f			/* branch if supervisor */
	lwz	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	li	31,EXC_ISEG
	mtlr	31
	ba	s_trap
1:
	mfsprg	1,1			/* restore SP */
	li	31,EXC_ISEG
	mtlr	31
	ba	s_trap

/*
 * Similar to the above for ISI
 */
NENTRY(isitrap)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_DISISAVE(1)	/* free r28-r31 */
	std	29,CI_DISISAVE+8(1)
	std	30,CI_DISISAVE+16(1)
	std	31,CI_DISISAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	li	31,EXC_ISI
	mtlr	31
	ba	s_isitrap

NENTRY(isegtrap)
	ba	s_isegtrap
s_isegtrap:
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	3,CI_SEGSAVE+0(1)
	std	4,CI_SEGSAVE+8(1)
	std	5,CI_SEGSAVE+16(1)
	std	6,CI_SEGSAVE+24(1)
	mfcr	6

	mfasr	3
	mfsrr1	4
	andi.	4,4,PSL_PR
	bne	1f			/* branch if user */
	lis	3,ksegtab@ha
	addi	3,3,ksegtab@l
1:	mfsrr0	4			/* fault EA */
	srdi	4,4,28			/* ESID */

	sldi	4,4,4
	add	3,3,4
	ld	4,0(3)
	ld	5,8(3)
	andis.	3,5,0x0800		/* valid? */
	beq	isegmiss
	slbmte	4,5

	mtcr	6
	ld	3,CI_SEGSAVE+0(1)
	ld	4,CI_SEGSAVE+8(1)
	ld	5,CI_SEGSAVE+16(1)
	ld	6,CI_SEGSAVE+24(1)
	mfsprg	1,1			/* restore SP */
	rfid

/*
 * This one for the external interrupt handler.
 */
NENTRY(extint)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	mfxer	30			/* save XER */
	lwa	31,CI_INTRDEPTH(1)	/* were we already running on intstk? */
	addic.	31,31,1
	stw	31,CI_INTRDEPTH(1)
	beq+	1f
	mfsprg	1,1			/* yes, get old SP */
	ba	extintr
1:
	lwz	1,CI_INTSTK(1)		/* get interrupt stack */
	ba	extintr

/*
 * And this one for the decrementer interrupt handler.
 */
NENTRY(decrint)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	mfxer	30			/* save XER */
	lwa	31,CI_INTRDEPTH(1)	/* were we already running on intstk? */
	addic.	31,31,1
	stw	31,CI_INTRDEPTH(1)
	beq+	1f
	mfsprg	1,1			/* yes, get old SP */
	ba	decrintr
1:
	lwz	1,CI_INTSTK(1)		/* get interrupt stack */
	ba	decrintr

/*
 * Performance Monitor Exception (0x0f00) and
 * AltiVec Unavailable Exception (0x0f20) handler.
 */
NENTRY(perftrap)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	li	31,EXC_PERF
	b	2f

	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	li	31,EXC_VEC
2:
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	mtlr	31
/* Test whether we already had PR set */
	mfsrr1	31
	andi.	31,31,PSL_PR
	beq	1f			/* branch if supervisor */
	lwz	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	ba	s_trap
1:
	mfsprg	1,1			/* restore SP */
	ba	s_trap

#ifdef DDB
/*
 * In case of DDB we want a separate trap catcher for it
 */
NENTRY(ddblow)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_DDBSAVE(1)	/* free r28-r31 */
	std	29,CI_DDBSAVE+8(1)
	std	30,CI_DDBSAVE+16(1)
	std	31,CI_DDBSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	lwz	1,CI_DDBSTK(1)		/* get new SP */
	li	31,0
	stwu	31,-16(1)		/* end of stack chain */
	bl	1f
1:
	mflr	31
	rlwinm	31,31,0,0,23		/* r31 &= ~0xff */
	mtlr	31
	ba	ddbtrap
#endif	/* DDB */

#ifdef IPKDB
#define CI_IPKDBSAVE CI_DDBSAVE
#define CI_IPKDBSTK  CI_DDBSTK

/*
 * In case of IPKDB we want a separate trap catcher for it
 */
NENTRY(ipkdblow)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_IPKDBSAVE(1)	/* free r28-r31 */
	std	29,CI_IPKDBSAVE+8(1)
	std	30,CI_IPKDBSAVE+16(1)
	std	31,CI_IPKDBSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	lwz	1,CI_IPKDBSTK(1)	/* get new SP */
	li	31,0
	stwu	31,-16(1)		/* end of stack chain */
	bl	1f
1:
	mflr	31
	rlwinm	31,31,0,0,23		/* r31 &= ~0xff */
	mtlr	31
	ba	ipkdbtrap
#endif	/* IPKDB */

/*
 * FRAME_SETUP assumes:
 *	SPRG1		SP (1)
 *	savearea	r28-r31,DAR,DSISR	(DAR & DSISR only for DSI traps)
 *	28		LR
 *	29		CR
 *	1		kernel stack
 *	LR		trap type
 *	SRR0/1		as at start of trap
 */
#define FRAME_SETUP(savearea)						\
/* Have to enable translation to allow access of kernel stack: */	\
	slbia;								\
	GET_CPUINFO(30);						\
	mfsrr0	31;							\
	std	31,savearea+48(30);					\
	mfsrr1	31;							\
	std	31,savearea+56(30);					\
	mfsprg	30,1;							\
	mfmsr	31;							\
	ori	31,31,(PSL_DR|PSL_IR);					\
	mtmsr	31;							\
	isync;								\
	stwu	30,-FRAMELEN(1);					\
	stw	0,FRAME_0+8(1);						\
	stw	30,FRAME_1+8(1);					\
	stw	2,FRAME_2+8(1);						\
	stw	28,FRAME_LR+8(1);					\
	stw	29,FRAME_CR+8(1);					\
	GET_CPUINFO(2);							\
	ld	28,savearea(2);						\
	ld	29,savearea+8(2);					\
	ld	30,savearea+16(2);					\
	ld	31,savearea+24(2);					\
	stmw	3,FRAME_3+8(1);						\
	ld	28,savearea+32(2);					\
	ld	29,savearea+40(2);					\
	ld	30,savearea+48(2);					\
	ld	31,savearea+56(2);					\
	mfxer	3;							\
	mfctr	4;							\
	mflr	5;							\
	stw	3,FRAME_XER+8(1);					\
	stw	4,FRAME_CTR+8(1);					\
	stw	5,FRAME_EXC+8(1);					\
	stw	28,FRAME_DAR+8(1);					\
	stw	29,FRAME_DSISR+8(1);					\
	stw	30,FRAME_SRR0+8(1);					\
	stw	31,FRAME_SRR1+8(1)

#define FRAME_LEAVE(savearea)						\
/* Now restore regs: */							\
	lwz	3,FRAME_SRR0+8(1);					\
	lwz	4,FRAME_SRR1+8(1);					\
	lwz	5,FRAME_CTR+8(1);					\
	lwz	6,FRAME_XER+8(1);					\
	lwz	7,FRAME_LR+8(1);					\
	GET_CPUINFO(2);							\
	std	3,savearea+48(2);					\
	std	4,savearea+56(2);					\
	mtctr	5;							\
	mtxer	6;							\
	mtlr	7;							\
	lmw	3,FRAME_3+8(1);						\
	std	28,savearea(2);						\
	std	29,savearea+8(2);					\
	std	31,savearea+24(2);					\
	lwz	28,FRAME_2+8(1);	/* save r2 in r28 */		\
	lwz	29,FRAME_CR+8(1);	/* save cr in r29 */		\
	lwz	0,FRAME_0+8(1);						\
	lwz	1,FRAME_1+8(1);						\
/* Disable translation, machine check and recoverability: */		\
	mfmsr	31;							\
	andi.	31,31,~(PSL_DR|PSL_IR|PSL_ME|PSL_RI)@l;			\
	mtmsr	31;							\
	isync;								\
	mtsprg	2,28;			/* save r2 */			\
	ld	31,savearea+48(2);					\
	mtsrr0	31;							\
	ld	31,savearea+56(2);					\
	mtsrr1	31;							\
/* Decide whether we return to user mode: */				\
	andi.	31,31,PSL_PR;						\
	beq	1f;			/* branch if supervisor */	\
/* Restore user & kernel access SR: */					\
	slbia;								\
1:	mtcr	29;			/* restore cr */		\
	ld	28,savearea(2);						\
	ld	29,savearea+8(2);					\
	ld	31,savearea+24(2);					\
	mfsprg	2,2			/* restore r2 */

/*
 * Preamble code for DSI/ISI traps
 */
disitrap:
	ld	30,CI_DISISAVE(1)
	ld	31,CI_DISISAVE+8(1)
	std	30,CI_TEMPSAVE(1)
	std	31,CI_TEMPSAVE+8(1)
	ld	30,CI_DISISAVE+16(1)
	ld	31,CI_DISISAVE+24(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mfdar	30
	mfdsisr	31
	std	30,CI_TEMPSAVE+32(1)
	std	31,CI_TEMPSAVE+40(1)
realtrap:
/* Test whether we already had PR set */
	mfsrr1	1
	andi.	1,1,PSL_PR
	mfsprg	1,1			/* restore SP (might have been
					   overwritten) */
	beq	s_trap			/* branch if supervisor */
	GET_CPUINFO(1)
	lwz	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */

/*
 * Now the common trap catching code.
 */
s_trap:
	FRAME_SETUP(CI_TEMPSAVE)
	mfmsr	3
	clrldi	3,3,1			/* clear MSR[SF] */
	mtmsrd	3
/* Now we can recover interrupts again: */
trapagain:
	mfmsr	7
	ori	7,7,(PSL_EE|PSL_ME|PSL_RI)@l
	mtmsr	7
	isync
/* Call C trap code: */
	addi	3,1,8
	bl	_C_LABEL(trap)
trapexit:
/* Disable interrupts: */
	mfmsr	3
	andi.	3,3,~PSL_EE@l
	mtmsr	3
/* Test AST pending: */
	lwz	5,FRAME_SRR1+8(1)
	andi.	5,5,PSL_PR
	beq	1f			/* branch if supervisor */
	GET_CPUINFO(3)
	lwz	4,CI_ASTPENDING(3)
	andi.	4,4,1
	beq	1f
	li	6,EXC_AST
	stw	6,FRAME_EXC+8(1)
	b	trapagain
1:
	FRAME_LEAVE(CI_TEMPSAVE)
	rfid

/*
 * Child comes here at the end of a fork.
 * Return to userspace via the trap return path.
 * XXX r3 and r4 are passed from cpu_switchto() unchanged.
 */
ENTRY_NOPROFILE(fork_trampoline)
	bl	_C_LABEL(lwp_startup)
	mtlr	31
	mr	3,30
	blrl				/* jump indirect to r31 */
	b	trapexit

ENTRY_NOPROFILE(setfunc_trampoline)
	mtlr	31
	mr	3,30
	blrl				/* jump indirect to r31 */
	b	trapexit

/*
 * DSI second stage fault handler
 */
s_dsitrap:
	mfmsr	31
	clrldi	31,31,1			/* clear MSR[SF] */
	mtmsrd	31

	mfdsisr	31			/* test whether this may be a
					   spill fault */
	andis.	31,31,0x40000000@h
	beq	disitrap		/* branch if table miss is false */
	lwz	1,CI_SPILLSTK(1)	/* get spill stack */
	stwu	1,-128(1)
	std	0,24(1)			/* save volatile registers */
	std	3,32(1)
	std	4,40(1)
	std	5,48(1)
	std	6,56(1)
	std	7,64(1)
	std	8,72(1)
	std	9,80(1)
	std	10,88(1)
	std	11,96(1)
	std	12,104(1)
	mfxer	30			/* save XER */
	mfctr	31			/* & CTR */
	mfdar	3
	bl	_C_LABEL(pte_spill)	/* try a spill */
	li	4,EXC_DSI
s_pte_spill:
	cmpwi	3,0
	mtlr	4			/* trap type */
	mtxer	30			/* restore XER */
	mtctr	31			/* restore CTR */
	ld	0,24(1)			/* restore volatile registers */
	ld	3,32(1)
	ld	4,40(1)
	ld	5,48(1)
	ld	6,56(1)
	ld	7,64(1)
	ld	8,72(1)
	ld	9,80(1)
	ld	10,88(1)
	ld	11,96(1)
	ld	12,104(1)
	GET_CPUINFO(1)
	beq	disitrap
	mtcr	29			/* restore CR */
	mtlr	28			/* restore LR */
	ld	28,CI_DISISAVE(1)	/* restore r28-r31 */
	ld	29,CI_DISISAVE+8(1)
	ld	30,CI_DISISAVE+16(1)
	ld	31,CI_DISISAVE+24(1)
	mfsprg	1,1			/* restore SP */
	rfid				/* return to trapped code */

/*
 * ISI second stage fault handler
 */
s_isitrap:
	mfmsr	31
	clrldi	31,31,1			/* clear MSR[SF] */
	mtmsrd	31

	mfsrr1	31			/* test whether this may be a
					   spill fault */
	andis.	31,31,0x40000000@h
	beq	disitrap		/* branch if table miss is false */
	lwz	1,CI_SPILLSTK(1)	/* get spill stack */
	stwu	1,-128(1)
	std	0,24(1)			/* save volatile registers */
	std	3,32(1)
	std	4,40(1)
	std	5,48(1)
	std	6,56(1)
	std	7,64(1)
	std	8,72(1)
	std	9,80(1)
	std	10,88(1)
	std	11,96(1)
	std	12,104(1)
	mfxer	30			/* save XER */
	mfctr	31			/* & ctr */
	mfsrr0	3
	bl	_C_LABEL(pte_spill)	/* try a spill */
	li	4,EXC_ISI
	b	s_pte_spill		/* above */

#define INTRFRAME_srr0	8
#define INTRFRAME_srr1	16
#define INTRFRAME_depth	28
#define INTRFRAME_r0	48
#define INTRFRAME_r1	0		/* original SP */
#define INTRFRAME_r3	56
#define INTRFRAME_r4	64
#define INTRFRAME_r5	72
#define INTRFRAME_r6	80
#define INTRFRAME_r7	88
#define INTRFRAME_r8	96
#define INTRFRAME_r9	104
#define INTRFRAME_r10	112
#define INTRFRAME_r11	120
#define INTRFRAME_r12	128
#define INTRFRAME_lr	136
#define INTRFRAME_cr	144
#define INTRFRAME_xer	152
#define INTRFRAME_ctr	160

/*
 * External interrupt second level handler
 */
#define INTRENTER							\
/* Save volatile registers: */						\
	stwu	1,-192(1);		/* temporarily */		\
	std	0,INTRFRAME_r0(1);					\
	mfsprg	0,1;			/* get original SP */		\
	stw	0,INTRFRAME_r1(1);	/* and store it */		\
	std	3,INTRFRAME_r3(1);					\
	std	4,INTRFRAME_r4(1);					\
	std	5,INTRFRAME_r5(1);					\
	std	6,INTRFRAME_r6(1);					\
	std	7,INTRFRAME_r7(1);					\
	std	8,INTRFRAME_r8(1);					\
	std	9,INTRFRAME_r9(1);					\
	std	10,INTRFRAME_r10(1);					\
	std	11,INTRFRAME_r11(1);					\
	std	12,INTRFRAME_r12(1);					\
	std	28,INTRFRAME_lr(1);	/* saved LR */			\
	std	29,INTRFRAME_cr(1);	/* saved CR */			\
	std	30,INTRFRAME_xer(1);	/* saved XER */			\
	GET_CPUINFO(5);							\
	ld	28,CI_TEMPSAVE(5);	/* restore r28-r31 */		\
	ld	29,CI_TEMPSAVE+8(5);					\
	ld	30,CI_TEMPSAVE+16(5);					\
	ld	31,CI_TEMPSAVE+24(5);					\
	mfctr	6;							\
	lwz	5,CI_INTRDEPTH(5);					\
	mfsrr0	4;							\
	mfsrr1	3;							\
	std	6,INTRFRAME_ctr(1);					\
	stw	5,INTRFRAME_depth(1);					\
	std	4,INTRFRAME_srr0(1);					\
	std	3,INTRFRAME_srr1(1);					\
/* interrupts are recoverable here, and enable translation */		\
	andi.	0,3,PSL_PR;						\
	beq	1f;			/* branch if supervisor */	\
	slbia;								\
1:	mfmsr	5;							\
	ori	5,5,(PSL_IR|PSL_DR|PSL_RI);				\
	mtmsr	5;							\
	isync

extintr:
	INTRENTER
	mfmsr	3
	clrldi	3,3,1			/* clear MSR[SF] */
	mtmsrd	3
NENTRY(extint_call)
	bl	_C_LABEL(extint_call)	/* to be filled in later */

intr_exit:
/* Disable interrupts (should already be disabled) and MMU here: */
	mfmsr	3
	andi.	3,3,~(PSL_EE|PSL_ME|PSL_RI|PSL_DR|PSL_IR)@l
	sync
	mtmsr	3
	isync
/* restore possibly overwritten registers: */
	ld	12,INTRFRAME_r12(1)
	ld	11,INTRFRAME_r11(1)
	ld	10,INTRFRAME_r10(1)
	ld	9,INTRFRAME_r9(1)
	ld	8,INTRFRAME_r8(1)
	ld	7,INTRFRAME_r7(1)
	ld	6,INTRFRAME_srr1(1)
	ld	5,INTRFRAME_srr0(1)
	ld	4,INTRFRAME_ctr(1)
	ld	3,INTRFRAME_xer(1)
	mtsrr1	6
	mtsrr0	5
	mtctr	4
	mtxer	3
/* Returning to user mode? */
	andi.	6,6,PSL_PR		/* saved SRR1 */
	beq	1f			/* branch if supervisor */
	slbia
	GET_CPUINFO(4)
	lwz	6,CI_ASTPENDING(4)	/* Test AST pending */
	andi.	6,6,1
	beq	1f
/* Setup for entry to realtrap: */
	lwz	3,INTRFRAME_r1(1)	/* get saved SP */
	mtsprg	1,3
	li	6,EXC_AST
	std	28,CI_TEMPSAVE(4)	/* establish tempsave again */
	std	29,CI_TEMPSAVE+8(4)
	std	30,CI_TEMPSAVE+16(4)
	std	31,CI_TEMPSAVE+24(4)
	mtlr	6
	ld	28,INTRFRAME_lr(1)	/* saved LR */
	ld	29,INTRFRAME_cr(1)	/* saved CR */
	ld	6,INTRFRAME_r6(1)
	ld	5,INTRFRAME_r5(1)
	ld	4,INTRFRAME_r4(1)
	ld	3,INTRFRAME_r3(1)
	ld	0,INTRFRAME_r0(1)
	GET_CPUINFO(30)
	lwz	31,CI_INTRDEPTH(30)	/* adjust reentrancy count */
	addi	31,31,-1
	stw	31,CI_INTRDEPTH(30)
	b	realtrap
1:
/* Here is the normal exit of extintr: */
	ld	5,INTRFRAME_cr(1)
	ld	6,INTRFRAME_lr(1)
	mtcr	5
	mtlr	6
	ld	6,INTRFRAME_r6(1)
	ld	5,INTRFRAME_r5(1)
	GET_CPUINFO(3)
	lwz	4,CI_INTRDEPTH(3)	/* adjust reentrancy count */
	addi	4,4,-1
	stw	4,CI_INTRDEPTH(3)
	ld	4,INTRFRAME_r4(1)
	ld	3,INTRFRAME_r3(1)
	ld	0,INTRFRAME_r0(1)
	lwz	1,INTRFRAME_r1(1)
	rfid

/*
 * Decrementer interrupt second level handler
 */
decrintr:
	INTRENTER
	mfmsr	3
	clrldi	3,3,1			/* clear MSR[SF] */
	mtmsrd	3
	addi	3,1,INTRFRAME_srr0	/* intr frame */
	bl	_C_LABEL(decr_intr)
	b	intr_exit

#ifdef DDB
/*
 * Deliberate entry to ddbtrap
 */
NENTRY(ddb_trap)
	mtsprg	1,1
	mfmsr	3
	mtsrr1	3
	andi.	3,3,~(PSL_EE|PSL_ME)@l
	mtmsr	3			/* disable interrupts */
	isync
	GET_CPUINFO(3)
	std	28,CI_DDBSAVE(3)
	std	29,CI_DDBSAVE+8(3)
	std	30,CI_DDBSAVE+16(3)
	std	31,CI_DDBSAVE+24(3)
	mflr	28
	li	29,EXC_BPT
	mtlr	29
	mfcr	29
	mtsrr0	28

/*
 * Now the ddb trap catching code.
 */
ddbtrap:
	FRAME_SETUP(CI_DDBSAVE)
	mfmsr	3
	clrldi	3,3,1			/* clear MSR[SF] */
	mtmsrd	3
/* Call C trap code: */
	addi	3,1,8
	bl	_C_LABEL(ddb_trap_glue)
	cmpwi	3,0
	bne	ddbleave
/* This wasn't for DDB, so switch to real trap: */
	GET_CPUINFO(4)
	lwz	3,FRAME_EXC+8(1)	/* save exception */
	std	3,CI_DDBSAVE+32(4)
	FRAME_LEAVE(CI_DDBSAVE)
	mtsprg	1,1			/* prepare for entrance to realtrap */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28
	mfcr	29
	ld	31,CI_DDBSAVE+32(1)
	mtlr	31
	b	realtrap
ddbleave:
	FRAME_LEAVE(CI_DDBSAVE)
	rfid
#endif /* DDB */

#ifdef IPKDB
/*
 * Deliberate entry to ipkdbtrap
 */
NENTRY(ipkdb_trap)
	mtsprg	1,1
	mfmsr	3
	mtsrr1	3
	andi.	3,3,~(PSL_EE|PSL_ME)@l
	mtmsr	3			/* disable interrupts */
	isync
	GET_CPUINFO(3)
	std	28,CI_IPKDBSAVE(3)
	std	29,CI_IPKDBSAVE+8(3)
	std	30,CI_IPKDBSAVE+16(3)
	std	31,CI_IPKDBSAVE+24(3)
	mflr	28
	li	29,EXC_BPT
	mtlr	29
	mfcr	29
	mtsrr0	28

/*
 * Now the ipkdb trap catching code.
 */
ipkdbtrap:
	FRAME_SETUP(CI_IPKDBSAVE)
	mfmsr	3
	clrldi	3,3,1			/* clear MSR[SF] */
	mtmsrd	3
/* Call C trap code: */
	addi	3,1,8
	bl	_C_LABEL(ipkdb_trap_glue)
	cmpwi	3,0
	bne	ipkdbleave
/* This wasn't for IPKDB, so switch to real trap: */
	GET_CPUINFO(4)
	lwz	3,FRAME_EXC+8(1)	/* save exception */
	std	3,CI_IPKDBSAVE+32(4)
	FRAME_LEAVE(CI_IPKDBSAVE)
	mtsprg	1,1			/* prepare for entrance to realtrap */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28
	mfcr	29
	ld	31,CI_IPKDBSAVE+32(1)
	mtlr	31
	b	realtrap
ipkdbleave:
	FRAME_LEAVE(CI_IPKDBSAVE)
	rfi

ipkdbfault:
	ba	_ipkdbfault
_ipkdbfault:
	mfsrr0	3
	addi	3,3,4
	mtsrr0	3
	li	3,-1
	rfi

/*
 * int ipkdbfbyte(unsigned char *p)
 */
NENTRY(ipkdbfbyte)
	li	9,EXC_DSI		/* establish new fault routine */
	lwz	5,0(9)
	lis	6,ipkdbfault@ha
	lwz	6,ipkdbfault@l(6)
	stw	6,0(9)
#ifdef IPKDBUSERHACK
	/* XXX see powerpc/trap_subr.S */
	lis	8,USER_ADDR@h
	slbie	8
	isync
#endif
	dcbst	0,9			/* flush data... */
	sync
	icbi	0,9			/* and instruction caches */
	lbz	3,0(3)			/* fetch data */
	stw	5,0(9)			/* restore previous fault handler */
	dcbst	0,9			/* and flush data... */
	sync
	icbi	0,9			/* and instruction caches */
	blr

/*
 * int ipkdbsbyte(unsigned char *p, int c)
 */
NENTRY(ipkdbsbyte)
	li	9,EXC_DSI		/* establish new fault routine */
	lwz	5,0(9)
	lis	6,ipkdbfault@ha
	lwz	6,ipkdbfault@l(6)
	stw	6,0(9)
#ifdef IPKDBUSERHACK
	/* XXX see powerpc/trap_subr.S */
	lis	8,USER_ADDR@h
	slbie	8
	isync
#endif
	dcbst	0,9			/* flush data... */
	sync
	icbi	0,9			/* and instruction caches */
	mr	6,3
	li	3,0
	stb	4,0(6)
	dcbst	0,6			/* Now do appropriate flushes
					   to data... */
	sync
	icbi	0,6			/* and instruction caches */
	stw	5,0(9)			/* restore previous fault handler */
	dcbst	0,9			/* and flush data... */
	sync
	icbi	0,9			/* and instruction caches */
	blr
#endif	/* IPKDB */

NENTRY(syscalltrap)
	mtsprg	1,1			/* save SP */
	GET_CPUINFO(1)
	std	28,CI_TEMPSAVE(1)	/* free r28-r31 */
	std	29,CI_TEMPSAVE+8(1)
	std	30,CI_TEMPSAVE+16(1)
	std	31,CI_TEMPSAVE+24(1)
	mflr	28			/* save LR */
	mfcr	29			/* save CR */
	lwz	1,CI_CURPCB(1)
	addi	1,1,USPACE		/* stack is top of user struct */
	li	31,EXC_SC
	mtlr	31
	ba	1f
1:
	FRAME_SETUP(CI_TEMPSAVE)
	mfmsr	7
	clrldi	7,7,1			/* clear MSR[SF] */
/* Now we can recover interrupts again: */
	ori	7,7,(PSL_EE|PSL_ME|PSL_RI)@l
	mtmsrd	7
	isync
	GET_CPUINFO(3)
	lwz	3,CI_CURLWP(3)
	lwz	3,L_PROC(3)
	lwz	4,P_MD_SYSCALL(3)
	mtlr	4
	addi	3,1,8
	blrl
	b	trapexit
