/*	$Id: locore_subr.S,v 1.45 2009/12/25 03:17:52 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_ddb.h"
#include "opt_ipkdb.h"
#include "opt_lockdebug.h"
#include "opt_multiprocessor.h"
#include "assym.h"

#include <sys/syscall.h>

#include <machine/param.h>
#include <machine/pmap.h>
#include <machine/psl.h>
#include <machine/spr.h>
#include <machine/trap.h>
#include <machine/asm.h>

	.comm	ksegtab,NBPG,NBPG

	.text
/*
 * struct lwp *
 * cpu_switchto(struct lwp *current, struct lwp *new)
 * Switch to the indicated new LWP.
 * 	r3 - current LWP
 * 	r4 - LWP to switch to
 */
FDESC(cpu_switchto)

ENTRY(cpu_switchto)
	mflr	0			/* save lr */
	std	0,16(1)
	stdu	1,-64(1)
	std	30,24(1)
	std	31,32(1)

	mr	30,3			/* r30 = curlwp */
	mr	31,4			/* r31 = newlwp */

	cmpdi	30,0			/* old lwp was exiting? */
	beq	switch_exited

	mfcr	11			/* save cr */
	stdu	1,-SFRAMELEN(1)		/* still running on old stack */
	std	11,32(1)
	std	2,40(1)
	std	13,48(1)
	std	14,56(1)
	std	15,64(1)
	std	16,72(1)
	std	17,80(1)
	std	18,88(1)
	std	19,96(1)
	std	20,104(1)
	std	21,112(1)
	std	22,120(1)
	std	23,128(1)
	std	24,136(1)
	std	25,144(1)
	std	26,152(1)
	std	27,160(1)
	std	28,168(1)
	std	29,176(1)
	std	30,184(1)
	std	31,192(1)
	ld	5,L_ADDR(30)
	std	1,PCB_SP(5)		/* save SP */

switch_exited:
	GET_CPUINFO(5)
	std	31,CI_CURLWP(5)		/* record new process */
	ld	6,L_ADDR(31)
	stdu	6,CI_CURPCB(5)		/* indicate new pcb */
	stdcx.	6,0,5			/* clear possible reservation */

	ld	1,PCB_SP(6)		/* get new procs SP */
	ld	11,32(1)		/* get other regs */
	ld	2,40(1)
	ld	13,48(1)
	ld	14,56(1)
	ld	15,64(1)
	ld	16,72(1)
	ld	17,80(1)
	ld	18,88(1)
	ld	19,96(1)
	ld	20,104(1)
	ld	21,112(1)
	ld	22,120(1)
	ld	23,128(1)
	ld	24,136(1)
	ld	25,144(1)
	ld	26,152(1)
	ld	27,160(1)
	ld	28,168(1)
	ld	29,176(1)
	ld	30,184(1)
	ld	31,192(1)
	ld	1,0(1)			/* get saved SP */
	mtcr	11			/* get saved cr */

	ld	30,24(1)
	ld	31,32(1)
	addi	1,1,64
	ld	0,16(1)
	mtlr	0
	blr

/*
 * int setfault()
 *
 * Similar to setjmp to setup for handling faults on accesses to user memory.
 * Any routine using this may only call bcopy, either the form below,
 * or the (currently used) C code optimized, so it doesn't use any non-volatile
 * registers.
 */
FDESC(setfault)

ENTRY(setfault)
	mflr	0
	mfcr	12
	GET_CPUINFO(4)
	ld	4,CI_CURPCB(4)
	std	3,PCB_FAULT(4)
	std	0,0(3)
	std	1,8(3)
	std	2,16(3)
	std	12,24(3)
	std	13,32(3)
	std	14,40(3)
	std	15,48(3)
	std	16,56(3)
	std	17,64(3)
	std	18,72(3)
	std	19,80(3)
	std	20,88(3)
	std	21,96(3)
	std	22,104(3)
	std	23,112(3)
	std	24,120(3)
	std	25,128(3)
	std	26,136(3)
	std	27,144(3)
	std	28,152(3)
	std	29,160(3)
	std	30,168(3)
	std	31,176(3)
	sync
	isync
	li	3,0
	blr

/*
 * int _lock_cas(uintptr_t *ptr, uintptr_t old, uintptr_t new);
 * XXX Assume r3 != r1.
 */
FDESC(_lock_cas)

ENTRY_NOPROFILE(_lock_cas)
1:	ld	6,0(3)
	cmpd	6,4
	bne	3f

	ldarx	6,0,3
	cmpd	6,4
	bne	2f

	stdcx.	5,0,3
	bne-	1b

	li	3,1
	blr

2:	addi	3,1,-8
	stdcx.	3,0,3			/* Clear reservation. */
3:	li	3,0
	blr

#ifndef LOCKDEBUG
/*
 * void mutex_enter(kmutex_t *);
 */
FDESC(mutex_enter)

ENTRY_NOPROFILE(mutex_enter)
	GET_CPUINFO(4)
	ld	5,CI_CURLWP(4)
1:
	ldarx	0,0,3
	cmpdi	0,0
	bne	2f

	stdcx.	5,0,3
	bne-	1b
	isync
	blr
2:
	mflr	0
	std	0,16(1)
	stdu	1,-128(1)
	bl	_C_LABEL(mutex_vector_enter)
	nop

	addi	3,1,64
	stdcx.	3,0,3			/* Clear reservation. */

	addi	1,1,128
	ld	0,16(1)
	mtlr	0
	blr

/*
 * void mutex_exit(kmutex_t *);
 */
FDESC(mutex_exit)

ENTRY_NOPROFILE(mutex_exit)
	GET_CPUINFO(4)
	ld	5,CI_CURLWP(4)

	ld	0,0(3)
	cmpd	0,5
	bne	2f

	sync
	li	0,0
	std	0,0(3)
	blr
2:
	mflr	0
	std	0,16(1)
	stdu	1,-128(1)
	bl	_C_LABEL(mutex_vector_exit)
	nop
	addi	1,1,128
	ld	0,16(1)
	mtlr	0
	blr

#endif /* !LOCKDEBUG */

/*
 * The following code gets copied to the top of the user stack on process
 * execution.  It does signal trampolining on signal delivery.
 *
 * On entry r1 points to a struct sigcontext at bottom of current stack.
 * All other registers are unchanged.
 */
	.globl	_C_LABEL(sigcode),_C_LABEL(esigcode)
_C_LABEL(sigcode):
	addi	1,1,-32			/* reserved space for callee */
	mr	31,5			/* save scp */
	blrl
	mr	3,31			/* restore scp */
	li	0,SYS_compat_16___sigreturn14
	sc				/* sigreturn(scp) */
	li	0,SYS_exit
	sc				/* exit(errno) */
_C_LABEL(esigcode):
