/*	$NetBSD: ofw_machdep.c,v 1.8 2001/06/10 11:36:03 tsubai Exp $	*/

/*
 * Copyright (C) 1996 Wolfgang Solfrank.
 * Copyright (C) 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <dev/ofw/openfirm.h>
#include <machine/powerpc.h>

#define OFMEM_REGIONS	32
static struct mem_region OFmem[OFMEM_REGIONS + 1], OFavail[OFMEM_REGIONS + 3];

/*
 * This is called during initppc, before the system is really initialized.
 * It shall provide the total and the available regions of RAM.
 * Both lists must have a zero-size entry as terminator.
 * The available regions need not take the kernel into account, but needs
 * to provide space for two additional entry beyond the terminating one.
 */
void
mem_regions(memp, availp)
	struct mem_region **memp, **availp;
{
	int phandle, i, cnt, n;
	struct {
		u_int64_t start;
		u_int32_t size;
	} __packed reg[OFMEM_REGIONS];
	struct {
		u_int32_t start;
		u_int32_t size;
	} avail[OFMEM_REGIONS];

	/*
	 * Get memory.
	 */
	if ((phandle = OF_finddevice("/memory")) == -1)
		goto error;
	bzero(reg, sizeof reg);
	cnt = OF_getprop(phandle, "reg", reg, sizeof reg);
	if (cnt <= 0)
		goto error;
	cnt /= sizeof reg[0];

	bzero(OFmem, sizeof OFmem);
	for (n = 0, i = 0; i < cnt; i++) {
#ifndef __LP64__
		if (reg[i].start >= 0x100000000ULL)
			continue;
#endif
		if (reg[i].size == 0)
			continue;
		OFmem[n].start = reg[i].start;
		OFmem[n].size  = reg[i].size;
		n++;
	}

	bzero(avail, sizeof avail);
	cnt = OF_getprop(phandle, "available", avail, sizeof avail);
	if (cnt <= 0)
		goto error;
	cnt /= sizeof avail[0];

	bzero(OFavail, sizeof OFavail);
	for (n = 0, i = 0; i < cnt; i++) {
		if (avail[i].size == 0)
			continue;
		OFavail[n].start = avail[i].start;
		OFavail[n].size  = avail[i].size;
		n++;
	}
#ifdef __LP64__notyet
	for (i = 0; i < OFMEM_REGIONS; i++) {
		if (OFmem[i].start >= 0x100000000) {
			OFavail[n].start = OFmem[i].start;
			OFavail[n].size = OFmem[i].size;
			n++;
		}
	}
#endif

	*memp = OFmem;
	*availp = OFavail;
	return;

error:
	panic("no memory?");
}

#include <dev/cons.h>
#include <machine/cpufunc.h>
extern char ofvector[];

void
ppc_exit()
{
	mtdec(0x7fffffff);

	/* Restore trap vector. */
	bcopy(ofvector, (void *)0, 0x1000);
	cnpollc(1);

	OF_exit();
}

void
ppc_boot(str)
	char *str;
{
	OF_boot(str);
}
