/*	$NetBSD: openfirm.c,v 1.10 2001/06/19 12:02:58 simonb Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/param.h>
#include <sys/systm.h>

#include <machine/psl.h>
#include <machine/stdarg.h>

#include <dev/ofw/openfirm.h>

void ofw_stack(void);

static int OF_buf_space[NBPG/4];
char *OF_buf = (void *)OF_buf_space;

int
OF_peer(phandle)
	int phandle;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int phandle;
		int sibling;
	} args = {
		0, /* "peer" */
		1,
		1,
	};

	ofw_stack();
	args.name = (long)"peer";
	args.phandle = phandle;
	if (openfirmware(&args) == -1)
		return 0;
	return args.sibling;
}

int
OF_child(phandle)
	int phandle;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int phandle;
		int child;
	} args = {
		0, /* "child" */
		1,
		1,
	};

	ofw_stack();
	args.name = (long)"child";
	args.phandle = phandle;
	if (openfirmware(&args) == -1)
		return 0;
	return args.child;
}

int
OF_parent(phandle)
	int phandle;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int phandle;
		int parent;
	} args = {
		0, /* "parent" */
		1,
		1,
	};

	ofw_stack();
	args.name = (long)"parent";
	args.phandle = phandle;
	if (openfirmware(&args) == -1)
		return 0;
	return args.parent;
}

int
OF_instance_to_package(ihandle)
	int ihandle;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int ihandle;
		int phandle;
	} args = {
		0, /* "instance-to-package" */
		1,
		1,
	};

	ofw_stack();
	args.name = (long)"instance-to-package";
	args.ihandle = ihandle;
	if (openfirmware(&args) == -1)
		return -1;
	return args.phandle;
}

int
OF_getproplen(handle, prop)
	int handle;
	const char *prop;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int phandle;
		int prop;
		int proplen;
	} args = {
		0, /* "getproplen" */
		2,
		1,
	};

	ofw_stack();
	args.name = (long)"getproplen";
	args.phandle = handle;
	args.prop = (long)prop;
	if (openfirmware(&args) == -1)
		return -1;
	return args.proplen;
}

int
OF_getprop(handle, prop, buf, buflen)
	int handle;
	const char *prop;
	void *buf;
	int buflen;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int phandle;
		int prop;
		int buf;
		int buflen;
		int size;
	} args = {
		0, /* "getprop" */
		4,
		1,
	};

	ofw_stack();
	if (buflen > NBPG)
		return -1;
	args.name = (long)"getprop";
	args.phandle = handle;
	args.prop = (long)prop;
	args.buf = (long)OF_buf;
	args.buflen = buflen;
	if (openfirmware(&args) == -1)
		return -1;
	if (args.size > buflen)
		args.size = buflen;
	if (args.size > 0)
		ofbcopy(OF_buf, buf, args.size);
	return args.size;
}

int
OF_nextprop(handle, prop, nextprop)
	int handle;
	const char *prop;
	void *nextprop;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int phandle;
		int prop;
		int buf;
		int flag;
	} args = {
		0, /* "nextprop" */
		3,
		1,
	};

	ofw_stack();
	args.name = (long)"nextprop";
	args.phandle = handle;
	args.prop = (long)prop;
	args.buf = (long)OF_buf;
	if (openfirmware(&args) == -1)
		return -1;
	strncpy(nextprop, OF_buf, 32);
	return args.flag;
}

int
OF_finddevice(name)
	const char *name;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int device;
		int phandle;
	} args = {
		0, /* "finddevice" */
		1,
		1,
	};

	ofw_stack();
	args.name = (long)"finddevice";
	args.device = (long)name;
	if (openfirmware(&args) == -1)
		return -1;
	return args.phandle;
}

int
OF_instance_to_path(ihandle, buf, buflen)
	int ihandle;
	char *buf;
	int buflen;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int ihandle;
		int buf;
		int buflen;
		int length;
	} args = {
		0, /* "instance-to-path" */
		3,
		1,
	};

	ofw_stack();
	if (buflen > NBPG)
		return -1;
	args.name = (long)"instance-to-path";
	args.ihandle = ihandle;
	args.buf = (long)OF_buf;
	args.buflen = buflen;
	if (openfirmware(&args) < 0)
		return -1;
	if (args.length > buflen)
		args.length = buflen;
	if (args.length > 0)
		ofbcopy(OF_buf, buf, args.length);
	return args.length;
}

int
OF_package_to_path(phandle, buf, buflen)
	int phandle;
	char *buf;
	int buflen;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int phandle;
		int buf;
		int buflen;
		int length;
	} args = {
		0, /* "package-to-path" */
		3,
		1,
	};

	ofw_stack();
	if (buflen > NBPG)
		return -1;
	args.name = (long)"package-to-path";
	args.phandle = phandle;
	args.buf = (long)OF_buf;
	args.buflen = buflen;
	if (openfirmware(&args) < 0)
		return -1;
	if (args.length > buflen)
		args.length = buflen;
	if (args.length > 0)
		ofbcopy(OF_buf, buf, args.length);
	return args.length;
}

int
#ifdef __STDC__
OF_call_method(const char *method, int ihandle, int nargs, int nreturns, ...)
#else
OF_call_method(method, ihandle, nargs, nreturns, va_alist)
	const char *method;
	int ihandle;
	int nargs;
	int nreturns;
	va_dcl
#endif
{
	va_list ap;
	static struct {
		int name;
		int nargs;
		int nreturns;
		int method;
		int ihandle;
		int args_n_results[12];
	} args = {
		0, /* "call-method" */
		2,
		1,
	};
	int *ip, n;

	if (nargs > 6)
		return -1;
	args.name = (long)"call-method";
	args.nargs = nargs + 2;
	args.nreturns = nreturns + 1;
	args.method = (long)method;
	args.ihandle = ihandle;
	va_start(ap, nreturns);
	for (ip = args.args_n_results + (n = nargs); --n >= 0;)
		*--ip = va_arg(ap, int);
	ofw_stack();
	if (openfirmware(&args) == -1)
		return -1;
	if (args.args_n_results[nargs])
		return args.args_n_results[nargs];
	for (ip = args.args_n_results + nargs + (n = args.nreturns); --n > 0;)
		*va_arg(ap, int *) = *--ip;
	va_end(ap);
	return 0;
}

int
#ifdef __STDC__
OF_call_method_1(const char *method, int ihandle, int nargs, ...)
#else
OF_call_method_1(method, ihandle, nargs, va_alist)
	const char *method;
	int ihandle;
	int nargs;
	va_dcl
#endif
{
	va_list ap;
	static struct {
		int name;
		int nargs;
		int nreturns;
		int method;
		int ihandle;
		int args_n_results[8];
	} args = {
		0, /* "call-method" */
		2,
		2,
	};
	int *ip, n;

	if (nargs > 6)
		return -1;
	args.name = (long)"call-method";
	args.nargs = nargs + 2;
	args.method = (long)method;
	args.ihandle = ihandle;
	va_start(ap, nargs);
	for (ip = args.args_n_results + (n = nargs); --n >= 0;)
		*--ip = va_arg(ap, int);
	va_end(ap);
	ofw_stack();
	if (openfirmware(&args) == -1)
		return -1;
	if (args.args_n_results[nargs])
		return -1;
	return args.args_n_results[nargs + 1];
}

int
OF_open(dname)
	const char *dname;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int dname;
		int handle;
	} args = {
		0, /* "open" */
		1,
		1,
	};
	int l;

	ofw_stack();
	if ((l = strlen(dname)) >= NBPG)
		return -1;
	ofbcopy(dname, OF_buf, l + 1);
	args.name = (long)"open";
	args.dname = (long)OF_buf;
	if (openfirmware(&args) == -1)
		return -1;
	return args.handle;
}

void
OF_close(handle)
	int handle;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int handle;
	} args = {
		0, /* "close" */
		1,
		0,
	};

	ofw_stack();
	args.name = (long)"close";
	args.handle = handle;
	openfirmware(&args);
}

/*
 * This assumes that character devices don't read in multiples of NBPG.
 */
int
OF_read(handle, addr, len)
	int handle;
	void *addr;
	int len;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int ihandle;
		int addr;
		int len;
		int actual;
	} args = {
		0, /* "read" */
		3,
		1,
	};
	int l, act = 0;
	char *p = addr;

	ofw_stack();
	args.name = (long)"read";
	args.ihandle = handle;
	args.addr = (long)OF_buf;
	for (; len > 0; len -= l, p += l) {
		l = min(NBPG, len);
		args.len = l;
		if (openfirmware(&args) == -1)
			return -1;
		if (args.actual > 0) {
			ofbcopy(OF_buf, p, args.actual);
			act += args.actual;
		}
		if (args.actual < l) {
			if (act)
				return act;
			else
				return args.actual;
		}
	}
	return act;
}

int
OF_write(handle, addr, len)
	int handle;
	const void *addr;
	int len;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int ihandle;
		int addr;
		int len;
		int actual;
	} args = {
		0, /* "write" */
		3,
		1,
	};
	int l, act = 0;
	const char *p = addr;

	ofw_stack();
	args.name = (long)"write";
	args.ihandle = handle;
	args.addr = (long)OF_buf;
	for (; len > 0; len -= l, p += l) {
		l = min(NBPG, len);
		ofbcopy(p, OF_buf, l);
		args.len = l;
		if (openfirmware(&args) == -1)
			return -1;
		l = args.actual;
		act += l;
	}
	return act;
}

int
OF_seek(handle, pos)
	int handle;
	u_quad_t pos;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int handle;
		int poshi;
		int poslo;
		int status;
	} args = {
		0, /* "seek" */
		3,
		1,
	};

	ofw_stack();
	args.name = (long)"seek";
	args.handle = handle;
	args.poshi = (int)(pos >> 32);
	args.poslo = (int)pos;
	if (openfirmware(&args) == -1)
		return -1;
	return args.status;
}

void
OF_boot(bootspec)
	const char *bootspec;
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int bootspec;
	} args = {
		0, /* "boot" */
		1,
		0,
	};
	int l;

	if ((l = strlen(bootspec)) >= NBPG)
		panic("OF_boot");
	ofw_stack();
	ofbcopy(bootspec, OF_buf, l + 1);
	args.name = (long)"boot";
	args.bootspec = (long)OF_buf;
	openfirmware(&args);
	for (;;);			/* just in case */
}

void
OF_enter()
{
	static struct {
		int name;
		int nargs;
		int nreturns;
	} args = {
		0, /* "enter" */
		0,
		0,
	};

	ofw_stack();
	args.name = (long)"enter";
	openfirmware(&args);
}

void
OF_exit()
{
	struct {
		int name;
		int nargs;
		int nreturns;
	} args;

	args.name = (long)"exit";
	args.nargs = 0;
	args.nreturns = 0;
	openfirmware(&args);
	for (;;);			/* just in case */
}

void
(*OF_set_callback (void (*newfunc)(void *))) (void *)
{
	static struct {
		int name;
		int nargs;
		int nreturns;
		int newfunc;
		int oldfunc;
	} args = {
		0, /* "set-callback" */
		1,
		1,
	};

	ofw_stack();
	args.name = (long)"set-callback";
	args.newfunc = (long)newfunc;
	if (openfirmware(&args) == -1)
		return 0;
	return (void *)(long)args.oldfunc;
}

/*
 * This version of bcopy doesn't work for overlapping regions!
 */
void
ofbcopy(src, dst, len)
	const void *src;
	void *dst;
	size_t len;
{
	const char *sp = src;
	char *dp = dst;

	if (src == dst)
		return;

	/*
	 * Do some optimization?						XXX
	 */
	while (len-- > 0)
		*dp++ = *sp++;
}
