/*	$Id: pmap.c,v 1.40 2011/01/18 03:17:55 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * TODO
 * wired Ƴ L ʥڡɤФʤ褦ˤ
 */

#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/pool.h>
#include <sys/proc.h>
#include <sys/queue.h>
#include <sys/systm.h>

#include <uvm/uvm.h>

#include <machine/pcb.h>
#include <machine/powerpc.h>

pte_t *ptable;
int ptab_cnt;
u_int ptab_mask;
#define HTABSIZE (ptab_cnt * sizeof(struct pte) * 8)

struct pte_ovfl {
	LIST_ENTRY(pte_ovfl) po_list;	/* Linked list of overflow entries */
	struct pte po_pte;		/* PTE for this mapping */
};

LIST_HEAD(pte_ovtab, pte_ovfl) *potable; /* Overflow entries for ptable */

struct pmap kernel_pmap_;

int physmem;
static int npgs;
#ifndef MSGBUFADDR
extern paddr_t msgbuf_paddr;
#endif

static struct mem_region *mem, *avail;

/*
 * This is a cache of referenced/modified bits.
 * Bits herein are shifted by ATTRSHFT.
 */
#define ATTRSHFT 4

struct pv_entry {
	struct pv_entry *pv_next;	/* Linked list of mappings */
	u_int pv_exec : 1;		/* Page has EXEC mapping */
	u_int pv_idx : 31;		/* Index into ptable */
	vaddr_t pv_va;			/* Virtual address of mapping */
	struct pmap *pv_pmap;
};

struct po_page;
struct po_page_info {
	LIST_ENTRY(po_page) pgi_list;
	struct vm_page *pgi_page;
	LIST_HEAD(po_freelist, pte_ovfl) pgi_freelist;
	int pgi_nfree;
};
#define NPOPPG	((NBPG - sizeof(struct po_page_info)) / sizeof(struct pte_ovfl))
struct po_page {
	struct po_page_info pop_pgi;
	struct pte_ovfl pop_po[NPOPPG];
};
LIST_HEAD(po_page_list, po_page) po_page_freelist;
int po_nfree;
int po_pcnt;
static struct pte_ovfl *poalloc __P((void));
static void pofree __P((struct pte_ovfl *, int));

static u_int usedsr[NSEGMENTS / sizeof(u_int) / 8];

static int pmap_initialized;

void pmap_release __P((struct pmap *));

static inline void tlbie __P((vaddr_t));
static inline void tlbsync __P((void));
static inline void tlbia __P((void));
static inline sr_t ptesr __P((struct ste *, vaddr_t));
static inline int pteidx __P((sr_t, vaddr_t));
static inline int ptematch __P((pte_t *, sr_t, vaddr_t, int));
static inline struct pv_entry *pa_to_pv __P((paddr_t));
static inline char *pa_to_attr __P((paddr_t));
static int pte_insert __P((int, pte_t *));
int pte_spill __P((vaddr_t));
static inline int pmap_enter_pv(struct pmap *, int, vaddr_t, paddr_t,
    vm_prot_t);
void pmap_remove_pv(struct pmap *, int, vaddr_t, paddr_t, struct pte *);
static void pmap_kenter_lpage(vaddr_t, paddr_t);
static pte_t *pte_find __P((struct pmap *, vaddr_t));
void *pmap_pool_palloc(struct pool *, int);
void pmap_pool_prelease(struct pool *, void *);

#define pmap_alloc_pv()		pool_get(&pmap_pv_pool, PR_NOWAIT)
#define pmap_free_pv(pv)	pool_put(&pmap_pv_pool, pv)

#ifndef PMAP_PV_LOWAT
#define PMAP_PV_LOWAT 16
#endif

int pmap_pv_lowat = PMAP_PV_LOWAT;
struct pool pmap_pmap_pool;
struct pool pmap_pv_pool;
struct pool pmap_stp_pool;

struct pool_allocator pmap_palloc = {
	pmap_pool_palloc, pmap_pool_prelease, 0
};

#define PTE_VSID(sr) ((sr) << 12)
#define PTE_API(va)  (((va) & 0x0f800000) >> 16)

/*
 * These small routines may have to be replaced,
 * if/when we support processors other that the 604.
 */
static inline void
tlbie(ea)
	vaddr_t ea;
{
	asm volatile ("tlbie %0" :: "r"(ea));
}

static inline void
tlbsync()
{
	asm volatile ("sync; tlbsync; sync");
}

static void
tlbia()
{
	vaddr_t i;

	asm volatile ("sync");
	for (i = 0; i < (vaddr_t)0x00100000; i += 0x00001000)
		tlbie(i);
	tlbsync();
}

static inline sr_t
ptesr(segtab, addr)
	struct ste *segtab;
	vaddr_t addr;
{
	return segtab[addr >> ADDR_SR_SHFT].vsid;
}

static inline int
pteidx(sr, addr)
	sr_t sr;
	vaddr_t addr;
{
	int hash;

	hash = sr ^ ((addr & ADDR_PIDX) >> ADDR_PIDX_SHFT);
	return hash & ptab_mask;
}

static inline int
ptematch(ptp, sr, va, which)
	pte_t *ptp;
	sr_t sr;
	vaddr_t va;
	int which;
{
	return ptp->pte_hi == (PTE_VSID(sr) | PTE_API(va) | which);
}

static inline struct pv_entry *
pa_to_pv(pa)
	paddr_t pa;
{
	int bank, pg;

	bank = vm_physseg_find(atop(pa), &pg);
	if (bank == -1)
		return NULL;
	return &vm_physmem[bank].pmseg.pvent[pg];
}

static inline char *
pa_to_attr(pa)
	paddr_t pa;
{
	int bank, pg;

	bank = vm_physseg_find(atop(pa), &pg);
	if (bank == -1)
		return NULL;
	return &vm_physmem[bank].pmseg.attrs[pg];
}

/*
 * Try to insert page table entry *pt into the ptable at idx.
 *
 * Note: *pt mustn't have PTE_VALID set.
 * This is done here as required by Book III, 4.12.
 */
static int
pte_insert(idx, pt)
	int idx;
	pte_t *pt;
{
	pte_t *ptp;
	int i;

	/*
	 * First try primary hash.
	 */
	for (ptp = ptable + idx * 8, i = 8; --i >= 0; ptp++)
		if (!(ptp->pte_hi & PTE_VALID)) {
			*ptp = *pt;
			ptp->pte_hi &= ~PTE_HID;
			asm volatile ("sync");
			ptp->pte_hi |= PTE_VALID;
			return 1;
		}
	idx ^= ptab_mask;
	for (ptp = ptable + idx * 8, i = 8; --i >= 0; ptp++)
		if (!(ptp->pte_hi & PTE_VALID)) {
			*ptp = *pt;
			ptp->pte_hi |= PTE_HID;
			asm volatile ("sync");
			ptp->pte_hi |= PTE_VALID;
			return 1;
		}
	return 0;
}

/*
 * Spill handler.
 *
 * Tries to spill a page table entry from the overflow area.
 * Note that this routine runs in real mode on a separate stack,
 * with interrupts disabled.
 */
int
pte_spill(addr)
	vaddr_t addr;
{
	int idx, i;
	sr_t sr;
	struct pte_ovfl *po;
	pte_t ps;
	pte_t *pt;
	struct ste *segtab;
	u_int srr1;

	asm ("mfsrr1 %0" : "=r"(srr1));
	if (srr1 & PSL_PR)
		asm ("mfasr %0" : "=r"(segtab));
	else
		segtab = ksegtab;
	sr = segtab[addr >> 28].vsid;
	idx = pteidx(sr, addr);
	for (po = LIST_FIRST(&potable[idx]); po; po = LIST_NEXT(po, po_list))
		if (ptematch(&po->po_pte, sr, addr, 0))
			goto found;
	return 0;

found:
	/*
	 * Now found an entry to be spilled into the real ptable.
	 */
	if (pte_insert(idx, &po->po_pte)) {
		LIST_REMOVE(po, po_list);
		pofree(po, 0);
		return 1;
	}
printf("pte_spill: replace\n");
	/*
	 * Have to substitute some entry. Use the primary hash for this.
	 * (XXX Should use atomic operation)
	 *
	 * Use low bits of timebase as random generator
	 */
	asm ("mftb %0" : "=r"(i));
	pt = ptable + idx * 8 + (i & 7);
	pt->pte_hi &= ~PTE_VALID;
	ps = *pt;
	asm volatile ("sync");
	tlbie(addr);
	tlbsync();
	*pt = po->po_pte;
	asm volatile ("sync");
	pt->pte_hi |= PTE_VALID;
	po->po_pte = ps;
	if (ps.pte_hi & PTE_HID) {
		/*
		 * We took an entry that was on the alternate hash
		 * chain, so move it to it's original chain.
		 */
		po->po_pte.pte_hi &= ~PTE_HID;
		LIST_REMOVE(po, po_list);
		LIST_INSERT_HEAD(potable + (idx ^ ptab_mask), po, po_list);
	}
	return 1;
}

/*
 * This is called during initppc, before the system is really initialized.
 */
void
pmap_bootstrap(kernelstart, kernelend)
	u_int kernelstart, kernelend;
{
	struct mem_region *mp, *mp1;
	int cnt, i;
	paddr_t s, e;
	psize_t sz;

	/*
	 * Get memory.
	 */
	mem_regions(&mem, &avail);
#if 1
	for (mp = mem; mp->size; mp++)
		printf("mem 0x%08lx 0x%08lx\n", mp->start, mp->size);
	for (mp = avail; mp->size; mp++)
		printf("avail 0x%08lx 0x%08lx (0x%08lx)\n",
		    mp->start, mp->start + mp->size, mp->size);
#endif

	for (mp = mem; mp->size; mp++)
		physmem += btoc(mp->size);

	/*
	 * Count the number of available entries.
	 */
	for (cnt = 0, mp = avail; mp->size; mp++)
		cnt++;

	/*
	 * Page align all regions.
	 * Non-page aligned memory isn't very interesting to us.
	 * Also, sort the entries for ascending addresses.
	 */
	kernelstart &= ~PGOFSET;
	kernelend = (kernelend + PGOFSET) & ~PGOFSET;
	for (mp = avail; mp->size; mp++) {
		s = mp->start;
		e = mp->start + mp->size;

		/*
		 * Don't use before the kernel.  (XXX)
		 */
		if (e <= kernelend)
			goto empty;

		/*
		 * Look whether this regions starts within the kernel.
		 */
		if (s < kernelend)
			s = kernelend;

		/*
		 * Now page align the start and size of the region.
		 */
		s = round_page(s);
		e = trunc_page(e);
		if (e < s)
			e = s;
		sz = e - s;

		/*
		 * Check whether some memory is left here.
		 */
		if (sz == 0) {
		empty:
			bcopy(mp + 1, mp, (cnt - (mp - avail)) * sizeof *mp);
			cnt--;
			mp--;
			continue;
		}

		/*
		 * Do an insertion sort.
		 */
		npgs += btoc(sz);
		for (mp1 = avail; mp1 < mp; mp1++)
			if (s < mp1->start)
				break;
		if (mp1 < mp) {
			bcopy(mp1, mp1 + 1, (char *)mp - (char *)mp1);
			mp1->start = s;
			mp1->size = sz;
		} else {
			mp->start = s;
			mp->size = sz;
		}
	}

	/*
	 * The PEM recommends that the total number of PTEGs should be
	 * at least 1/2 of the number of physical pages.
	 */
#ifdef HTABENTS
	ptab_cnt = HTABENTS;
#else
	ptab_cnt = (physmem + 1) / 2;

	/* The minimum is 2048 PTEGs. */
	if (ptab_cnt < 2048)
		ptab_cnt = 2048;

	/* Round up to power of 2. */
	asm ("cntlzw %0,%1" : "=r"(i) : "r"(ptab_cnt - 1));
	ptab_cnt = 1 << (32 - i);
#endif

	/*
	 * Find suitably aligned memory for HTAB.
	 */
	for (mp = avail; mp->size; mp++) {
		s = roundup(mp->start, HTABSIZE) - mp->start;
		if (mp->size < s + HTABSIZE)
			continue;
		ptable = (pte_t *)(long)(mp->start + s);
		if (mp->size == s + HTABSIZE) {
			if (s)
				mp->size = s;
			else {
				bcopy(mp + 1, mp,
				      (cnt - (mp - avail)) * sizeof *mp);
				mp = avail;
			}
			break;
		}
		if (s != 0) {
			bcopy(mp, mp + 1, (cnt - (mp - avail)) * sizeof *mp);
			mp++->size = s;
			cnt++;
		}
		mp->start += s + HTABSIZE;
		mp->size -= s + HTABSIZE;
		break;
	}
	if (!mp->size)
		panic("not enough memory?");

	npgs -= btoc(HTABSIZE);
	bzero(ptable, HTABSIZE);
	ptab_mask = ptab_cnt - 1;

	/*
	 * We cannot do pmap_steal_memory here,
	 * since we don't run with translation enabled yet.
	 */
	s = sizeof(struct pte_ovtab) * ptab_cnt;
	sz = round_page(s);
	for (mp = avail; mp->size; mp++)
		if (mp->size >= sz)
			break;
	if (!mp->size)
		panic("not enough memory?");

	npgs -= btoc(sz);
	potable = (struct pte_ovtab *)(long)mp->start;
	mp->size -= sz;
	mp->start += sz;
	if (mp->size <= 0)
		bcopy(mp + 1, mp, (cnt - (mp - avail)) * sizeof *mp);
	for (i = 0; i < ptab_cnt; i++)
		LIST_INIT(potable + i);

#ifndef MSGBUFADDR
	/*
	 * allow for msgbuf
	 */
	sz = round_page(MSGBUFSIZE);
	mp = NULL;
	for (mp1 = avail; mp1->size; mp1++)
		if (mp1->size >= sz)
			mp = mp1;
	if (mp == NULL)
		panic("not enough memory?");

	npgs -= btoc(sz);
	msgbuf_paddr = mp->start + mp->size - sz;
	mp->size -= sz;
	if (mp->size <= 0)
		bcopy(mp + 1, mp, (cnt - (mp - avail)) * sizeof *mp);
#endif

	for (mp = avail; mp->size; mp++)
		uvm_page_physload(atop(mp->start), atop(mp->start + mp->size),
			atop(mp->start), atop(mp->start + mp->size),
			VM_FREELIST_DEFAULT);

	/*
	 * Initialize the pmap pools.
	 */
	pool_init(&pmap_pmap_pool, sizeof(struct pmap), 0, 0, 0, "pmappool",
	    &pool_allocator_nointr, IPL_NONE);

	pool_init(&pmap_pv_pool, sizeof(struct pv_entry), 0, 0, 0, "pvpool",
	    &pmap_palloc, IPL_NONE);

	pool_init(&pmap_stp_pool, NBPG, NBPG, 0, 0, "stppool",
	    &pool_allocator_nointr, IPL_NONE);

	/*
	 * Initialize kernel pmap and hardware.
	 */
#if NSEGMENTS >= KERNEL_SEGMENT
#error XXX Too large NSEGMENTS!
	usedsr[KERNEL_SEGMENT / (sizeof usedsr[0] * 8)]
		|= 1 << (KERNEL_SEGMENT % (sizeof usedsr[0] * 8));
#endif
	pmap_kernel()->pm_segtab = ksegtab;
	pmap_kernel()->pm_segreal = ksegtab;
	asm volatile ("mtasr %0" :: "r"(ksegtab));

	/* XXX hash table */
	for (i = 0; i < 63; i++) {
		ksegtab[i].esid = i;
		ksegtab[i].index = i;
		ksegtab[i].vsid = KERNEL_SEGBASE + i * 29;	/* XXX */
		ksegtab[i].v = 1;
	}
	ksegtab[0].index = 63;

	asm volatile ("sync; slbia; slbie %0; isync" :: "r"(0));

	int ptab_size;
	ptab_size = ffs(ptab_cnt) - 1;
	asm volatile ("sync; mtsdr1 %0; isync"
		      :: "r"((u_long)ptable | (ptab_size - 11)));
	tlbia();

	ppc970_enable_largepage();

	/* Emulate BAT. */
	for (mp = mem; mp->size; mp++) {
		s = mp->start;
		e = s + mp->size;

		while (s < e) {
			pmap_kenter_lpage(s, s);
			ksegtab[s >> ADDR_SR_SHFT].l = 1;
			s += 16*1024*1024;
		}
	}
}

/*
 * Restrict given range to physical memory
 */
void
pmap_real_memory(start, size)
	paddr_t *start;
	psize_t *size;
{
	struct mem_region *mp;

	for (mp = mem; mp->size; mp++) {
		if (*start + *size > mp->start
		    && *start < mp->start + mp->size) {
			if (*start < mp->start) {
				*size -= mp->start - *start;
				*start = mp->start;
			}
			if (*start + *size > mp->start + mp->size)
				*size = mp->start + mp->size - *start;
			return;
		}
	}
	*size = 0;
}

/*
 * Initialize anything else for pmap handling.
 * Called during vm_init().
 */
void
pmap_init()
{
	struct pv_entry *pv;
	char *attr;
	vsize_t sz;
	vaddr_t addr;
	int s, bank;

	sz = (sizeof(struct pv_entry) + 1) * npgs;
	sz = round_page(sz);
	addr = uvm_km_alloc(kernel_map, sz, 0, UVM_KMF_WIRED | UVM_KMF_ZERO);
	if (addr == 0)
		panic("pmap_init: unable to allocate pv_table");

	s = splvm();
	pv = (struct pv_entry *)addr;
	attr = (char *)(pv + npgs);
	for (bank = 0; bank < vm_nphysseg; bank++) {
		sz = vm_physmem[bank].end - vm_physmem[bank].start;
		vm_physmem[bank].pmseg.pvent = pv;
		vm_physmem[bank].pmseg.attrs = attr;
		pv += sz;
		attr += sz;
	}

	/*
	 * Set a low water mark on the pv_entry pool, so that we are
	 * more likely to have these around even in extreme memory
	 * starvation.
	 */
	pool_setlowat(&pmap_pv_pool, pmap_pv_lowat);

	pmap_initialized = 1;
	splx(s);
}

/*
 * How much virtual space is available to the kernel?
 */
void
pmap_virtual_space(start, end)
	vaddr_t *start, *end;
{
	/*
	 * Reserve one segment for kernel virtual memory
	 */
	*start = (vaddr_t)(KERNEL_SR << ADDR_SR_SHFT);
	*end = *start + SEGMENT_LENGTH;
}

/*
 * Create and return a physical map.
 */
struct pmap *
pmap_create()
{
	struct pmap *pm;

	pm = pool_get(&pmap_pmap_pool, PR_WAITOK);
	bzero(pm, sizeof *pm);
	pm->pm_segtab = pool_get(&pmap_stp_pool, PR_WAITOK);
	pm->pm_segreal = (void *)vtophys((vaddr_t)pm->pm_segtab);
	bzero(pm->pm_segtab, NBPG);
	pmap_pinit(pm);

	return pm;
}

#if 1
#define mapsr(x)   bitrev(x)
#define unmapsr(x) bitrev(x)
#else
#define mapsr(x)   (x)
#define unmapsr(x) (x)
#endif
static inline u_int bitrev(u_int);

static inline u_int
bitrev(x)
	u_int x;
{
	u_int mask = ptab_mask;
	u_int y = x & ~mask;
	int shift;

	asm ("cntlzw %0,%1" : "=r"(shift) : "r"(mask));
	x <<= shift;

	while (x) {
		asm ("cntlzw %0,%1" : "=r"(shift) : "r"(x));
		x &= ~(0x80000000 >> shift);
		y |= 1 << shift;
	}

	return y;
}

static inline int allocsr(void);
static inline void freesr(int);

int
allocsr()
{
	int i, j, s;
	int sr;

	s = splvm();
	for (i = 0; i < sizeof usedsr / sizeof usedsr[0]; i++)
		if (usedsr[i] != 0xffffffff) {
			j = ffs(~usedsr[i]) - 1;
			usedsr[i] |= 1 << j;
			splx(s);
			sr = i * sizeof usedsr[0] * 8 + j;
			sr = mapsr(sr);
			return sr;
		}
	panic("out of segments");
}

void
freesr(sr)
	int sr;
{
	int i, j;

	sr = unmapsr(sr);
	j = sr % (sizeof usedsr[0] * 8);
	i = sr / (sizeof usedsr[0] * 8);
	usedsr[i] &= ~(1 << j);
}

int
pmap_allocseg(pm, i)
	struct pmap *pm;
	u_long i;
{
	if (i >= 63) {
		printf("pmap_allocseg: invalid segment (0x%lx)\n", i);
		return 1;
	}

	pm->pm_segtab[i].esid = i;
	pm->pm_segtab[i].index = i;
	pm->pm_segtab[i].vsid = allocsr();
	if (i == 0)
		pm->pm_segtab[i].index = 63;
	asm volatile ("sync" ::: "memory");
	pm->pm_segtab[i].v = 1;
	return 0;
}

/*
 * Initialize a preallocated and zeroed pmap structure.
 */
void
pmap_pinit(pm)
	struct pmap *pm;
{
	pm->pm_refs = 1;
}

/*
 * Add a reference to the given pmap.
 */
void
pmap_reference(pm)
	struct pmap *pm;
{
	pm->pm_refs++;
}

/*
 * Retire the given pmap from service.
 * Should only be called if the map contains no valid mappings.
 */
void
pmap_destroy(pm)
	struct pmap *pm;
{
	if (--pm->pm_refs == 0) {
		pmap_release(pm);
		pool_put(&pmap_stp_pool, pm->pm_segtab);
		pool_put(&pmap_pmap_pool, pm);
	}
}

/*
 * Release any resources held by the given physical map.
 * Called when a pmap initialized by pmap_pinit is being released.
 */
void
pmap_release(pm)
	struct pmap *pm;
{
	int i;

	for (i = 0; i < 256; i++)
		if (pm->pm_segtab[i].v)
			freesr(pm->pm_segtab[i].vsid);
}

/*
 * Copy the range specified by src_addr/len
 * from the source map to the range dst_addr/len
 * in the destination map.
 *
 * This routine is only advisory and need not do anything.
 */
void
pmap_copy(dst_pmap, src_pmap, dst_addr, len, src_addr)
	struct pmap *dst_pmap, *src_pmap;
	vaddr_t dst_addr, src_addr;
	vsize_t len;
{
}

/*
 * Fill the given physical page with zeroes.
 */
void
pmap_zero_page(pa)
	paddr_t pa;
{
#if 0
	bzero((caddr_t)pa, NBPG);
#else
	int i;

	for (i = NBPG/CACHELINESIZE; i > 0; i--) {
		asm volatile ("dcbz 0,%0" :: "r"(pa));
		pa += CACHELINESIZE;
	}
#endif
}

/*
 * Copy the given physical source page to its destination.
 */
void
pmap_copy_page(src, dst)
	paddr_t src, dst;
{
#if 0
	bcopy((caddr_t)src, (caddr_t)dst, NBPG);
#else
	long *s = (long *)src;
	long *d = (long *)dst;
	int i, j;

	for (i = NBPG/CACHELINESIZE; i > 0; i--) {
		asm volatile ("dcbz 0,%0" :: "r"(d) : "memory");
		for (j = CACHELINESIZE/sizeof *s; j > 0; j--)
			*d++ = *s++;
	}
#endif
}

/*
 * We really hope that we don't need overflow entries
 * before the VM system is initialized!					XXX
 */
static struct pte_ovfl *
poalloc()
{
	struct po_page *pop;
	struct pte_ovfl *po;
	struct vm_page *mem;
	int i;

	if (!pmap_initialized)
		panic("poalloc");

printf("poalloc\n");
	if (po_nfree == 0) {
		/*
		 * Since we cannot use maps for potable allocation,
		 * we have to steal some memory from the VM system.	XXX
		 */
		mem = uvm_pagealloc(NULL, 0, NULL, UVM_PGA_USERESERVE);
		po_pcnt++;
		pop = (struct po_page *)VM_PAGE_TO_PHYS(mem);
		pop->pop_pgi.pgi_page = mem;
		LIST_INIT(&pop->pop_pgi.pgi_freelist);
		for (i = NPOPPG - 1, po = pop->pop_po + 1; --i >= 0; po++)
			LIST_INSERT_HEAD(&pop->pop_pgi.pgi_freelist, po, po_list);
		po_nfree += pop->pop_pgi.pgi_nfree = NPOPPG - 1;
		LIST_INSERT_HEAD(&po_page_freelist, pop, pop_pgi.pgi_list);
		po = pop->pop_po;
	} else {
		po_nfree--;
		pop = LIST_FIRST(&po_page_freelist);
		if (--pop->pop_pgi.pgi_nfree <= 0)
			LIST_REMOVE(pop, pop_pgi.pgi_list);
		po = LIST_FIRST(&pop->pop_pgi.pgi_freelist);
		LIST_REMOVE(po, po_list);
	}
	return po;
}

static void
pofree(po, freepage)
	struct pte_ovfl *po;
	int freepage;
{
	struct po_page *pop;

	pop = (struct po_page *)trunc_page((vaddr_t)po);
	switch (++pop->pop_pgi.pgi_nfree) {
	case NPOPPG:
		if (!freepage)
			break;
		po_nfree -= NPOPPG - 1;
		po_pcnt--;
		LIST_REMOVE(pop, pop_pgi.pgi_list);
		uvm_pagefree(pop->pop_pgi.pgi_page);
		return;
	case 1:
		LIST_INSERT_HEAD(&po_page_freelist, pop, pop_pgi.pgi_list);
	default:
		break;
	}
	LIST_INSERT_HEAD(&pop->pop_pgi.pgi_freelist, po, po_list);
	po_nfree++;
}

/*
 * This returns whether this is the first (EXEC) mapping of a page.
 */
static inline int
pmap_enter_pv(pm, pteidx, va, pa, prot)
	struct pmap *pm;
	int pteidx;
	vaddr_t va;
	paddr_t pa;
	vm_prot_t prot;
{
	struct pv_entry *pv, *npv;
	int s, exec;

	if (!pmap_initialized)
		return 0;

	s = splvm();

	pv = pa_to_pv(pa);
	if (pv->pv_pmap == NULL) {
		/*
		 * No entries yet, use header as the first entry.
		 */
		exec = 0;	/* No previous exec mapping. */
		pv->pv_va = va;
		pv->pv_idx = pteidx;
		pv->pv_pmap = pm;
		pv->pv_exec = (prot & VM_PROT_EXECUTE) ? 1 : 0;
		pv->pv_next = NULL;
	} else {
		/*
		 * There is at least one other VA mapping this page.
		 * Place this entry after the header.
		 */
		/* Is there at least one exec mapping of this page? */
		for (npv = pv; npv; npv = npv->pv_next) {
			exec = npv->pv_exec;
			if (exec)
				break;
		}

		npv = pmap_alloc_pv();
		if (npv == NULL)
			return ENOMEM;
		npv->pv_va = va;
		npv->pv_idx = pteidx;
		npv->pv_pmap = pm;
		npv->pv_exec = (prot & VM_PROT_EXECUTE) ? 1 : 0;
		npv->pv_next = pv->pv_next;
		pv->pv_next = npv;
	}
	splx(s);
	return exec;
}

void
pmap_remove_pv(pm, pteidx, va, pa, pte)
	struct pmap *pm;
	int pteidx;
	vaddr_t va;
	paddr_t pa;
	struct pte *pte;
{
	struct pv_entry *pv, *npv;
	char *attr;

	/*
	 * First transfer reference/change bits to cache.
	 */
	attr = pa_to_attr(pa);
	if (attr == NULL)
		return;
	*attr |= (pte->pte_lo & (PTE_REF | PTE_CHG)) >> ATTRSHFT;

	/*
	 * Remove from the PV table.
	 */
	pv = pa_to_pv(pa);

	/*
	 * If it is the first entry on the list, it is actually
	 * in the header and we must copy the following entry up
	 * to the header.  Otherwise we must search the list for
	 * the entry.  In either case we free the now unused entry.
	 */
	if (pm == pv->pv_pmap && va == pv->pv_va) {
		if (pteidx != pv->pv_idx)
			panic("pmap_remove_pv");
		npv = pv->pv_next;
		if (npv) {
			*pv = *npv;
			pmap_free_pv(npv);
		} else
			pv->pv_pmap = NULL;
	} else {
		for (; (npv = pv->pv_next) != NULL; pv = npv)
			if (pm == npv->pv_pmap && va == npv->pv_va)
				break;
		if (npv) {
			if (pteidx != npv->pv_idx)
				panic("pmap_remove_pv");
			pv->pv_next = npv->pv_next;
			pmap_free_pv(npv);
		}
#ifdef DIAGNOSTIC
		else
			panic("pmap_remove_pv: not on list");
#endif
	}
}

/*
 * Insert physical page at pa into the given pmap at virtual address va.
 */
int
pmap_enter(pm, va, pa, prot, flags)
	struct pmap *pm;
	vaddr_t va;
	paddr_t pa;
	vm_prot_t prot;
	u_int flags;
{
	int s, idx, bank, pg, managed;
	sr_t sr;
	pte_t pte, *ptp;
	struct pte_ovfl *po;
	struct mem_region *mp;
	u_int wimg = pa & PTE_WIMG;

#ifdef DIAGNOSTIC
	if ((wimg & (PTE_I|PTE_M)) == (PTE_I|PTE_M))
		panic("pmap_enter (wimg)");
	if (pm->pm_segtab[va >> ADDR_SR_SHFT].l)
		panic("pmap_enter (L=1)");
#endif
	pa &= PTE_RPGN;

	/*
	 * Allocate a segment if needed.
	 */
	if (pm->pm_segtab[va >> ADDR_SR_SHFT].v == 0)
		pmap_allocseg(pm, va >> ADDR_SR_SHFT);

	/*
	 * Compute the HTAB index.
	 */
	idx = pteidx(sr = ptesr(pm->pm_segtab, va), va);

	/*
	 * Construct the PTE.
	 *
	 * Note: Don't set the valid bit for correct operation of tlb update.
	 */
	pte.pte_hi = PTE_VSID(sr) | PTE_API(va);
	pte.pte_lo = pa | PTE_M | PTE_I | PTE_G;

	managed = 0;
	if ((bank = vm_physseg_find(atop(pa), &pg)) != -1) {
		managed = 1;
		wimg = PTE_M;
	}
	if (wimg)
		pte.pte_lo = pa | wimg;
	else {
		for (mp = mem; mp->size; mp++) {
			if (pa >= mp->start && pa < mp->start + mp->size) {
				pte.pte_lo &= ~(PTE_I | PTE_G);
				break;
			}
		}
	}
	if (prot & VM_PROT_WRITE)
		pte.pte_lo |= PTE_RW;
	else
		pte.pte_lo |= PTE_RO;

	s = splvm();
	ptp = pte_find(pm, va);
	if (ptp == NULL)
		goto enter;

	/* XXX poentry */
	if (ptp < ptable || ptp >= ptable + ptab_cnt * 8) {
		pmap_remove(pm, va, va + NBPG);
		goto enter;
	}
	if ((ptp->pte_lo & PTE_RPGN) == pa) {
		/* Protection change. */
		ptp->pte_hi &= ~PTE_VALID;
		asm volatile ("sync");
		tlbie(va);
		tlbsync();

		/* Sync R/M bit */
		if (managed) {
			vm_physmem[bank].pmseg.attrs[pg] |=
			    (ptp->pte_lo & (PTE_REF|PTE_CHG)) >> ATTRSHFT;
		}
		ptp->pte_lo = pte.pte_lo;
		asm volatile ("sync");
		ptp->pte_hi |= PTE_VALID;
		/* asm volatile ("sync"); */
		splx(s);
		return 0;
	}

	ptp->pte_hi &= ~PTE_VALID;
	asm volatile ("sync");
	tlbie(va);
	tlbsync();
	pmap_remove_pv(pm, idx, va, ptp->pte_lo, ptp);
	pm->pm_stats.resident_count--;

enter:
	/*
	 * Now record mapping for later back-translation.
	 */
	if (pmap_initialized && managed) {
		int rv;

		rv = pmap_enter_pv(pm, idx, va, pa, prot);
		if (rv == ENOMEM) {
			splx(s);
			if (flags & PMAP_CANFAIL)
				return rv;
			panic("pmap_enter: out of memory");
		}
		if (rv == 0) {
			/*
			 * Flush the real memory from the cache.
			 */
			if (prot & VM_PROT_EXECUTE)
				__syncicache((void *)pa, NBPG);
		}
	}

	pm->pm_stats.resident_count++;
	/*
	 * Try to insert directly into HTAB.
	 */
	if (pte_insert(idx, &pte)) {
		splx(s);
		return 0;
	}

	/*
	 * Have to allocate overflow entry.
	 *
	 * Note, that we must use real addresses for these.
	 */
	po = poalloc();
	po->po_pte = pte;
	LIST_INSERT_HEAD(potable + idx, po, po_list);
	splx(s);

	return 0;
}

void
pmap_kenter_pa(va, pa, prot, flags)
	vaddr_t va;
	paddr_t pa;
	vm_prot_t prot;
	u_int flags;
{
	struct pmap *pm = pmap_kernel();
	int s, idx;
	sr_t sr;
	pte_t pte;
	struct pte_ovfl *po;
	struct mem_region *mp;
	u_int wimg = pa & PTE_WIMG;

#ifdef DIAGNOSTIC
	if ((wimg & (PTE_I|PTE_M)) == (PTE_I|PTE_M))
		panic("pmap_kenter_pa (wimg)");
	if (pm->pm_segtab[va >> ADDR_SR_SHFT].l)
		panic("pmap_kenter_pa (L=1)");
#endif
	pa &= PTE_RPGN;

	/*
	 * Compute the HTAB index.
	 */
	idx = pteidx(sr = ptesr(pm->pm_segtab, va), va);

	/*
	 * Construct the PTE.
	 *
	 * Note: Don't set the valid bit for correct operation of tlb update.
	 */
	pte.pte_hi = PTE_VSID(sr) | PTE_API(va);
	pte.pte_lo = pa | PTE_M | PTE_I | PTE_G;

	if (wimg)
		pte.pte_lo = pa | wimg;
	else {
		for (mp = mem; mp->size; mp++) {
			if (pa >= mp->start && pa < mp->start + mp->size) {
				pte.pte_lo &= ~(PTE_I | PTE_G);
				break;
			}
		}
	}
	if (prot & VM_PROT_WRITE)
		pte.pte_lo |= PTE_RW;
	else
		pte.pte_lo |= PTE_RO;

#if 0
	/*
	 * Flush the real memory from the cache.
	 * (XXX This should not happen)
	 */
	if (prot & VM_PROT_EXECUTE)
		__syncicache((void *)pa, NBPG);
#endif

	s = splvm();
	pm->pm_stats.resident_count++;
	/*
	 * Try to insert directly into HTAB.
	 */
	if (pte_insert(idx, &pte)) {
		splx(s);
		return;
	}

	/*
	 * Have to allocate overflow entry.
	 *
	 * Note, that we must use real addresses for these.
	 */
	po = poalloc();
	po->po_pte = pte;
	LIST_INSERT_HEAD(potable + idx, po, po_list);
	splx(s);
}

/*
 * Like pmap_kenter_pa(), but uses large page (16MB).
 */
void
pmap_kenter_lpage(va, pa)
	vaddr_t va;
	paddr_t pa;
{
	struct pmap *pm = pmap_kernel();
	int idx, hash;
	sr_t sr;
	pte_t pte;

	pa &= PTE_RPGN;

	/*
	 * Compute the HTAB index.
	 */
	sr = ptesr(pm->pm_segtab, va);
	hash = sr ^ ((va & 0x0f000000) >> 24);
	idx = hash & ptab_mask;

	/*
	 * Construct the PTE.
	 *
	 * Note: Don't set the valid bit for correct operation of tlb update.
	 */
	pte.pte_hi = PTE_VSID(sr) | PTE_API(va);
	pte.pte_lo = pa | PTE_M | PTE_RW;

	/*
	 * Try to insert directly into HTAB.
	 */
	if (pte_insert(idx, &pte))
		return;

	panic("pmap_kenter_lpage(): Increase HTABSIZE!");
}

/*
 * Remove the given range of mapping entries.
 */
void
pmap_remove(pm, va, endva)
	struct pmap *pm;
	vaddr_t va, endva;
{
	int idx, i, s;
	sr_t sr;
	pte_t *ptp;
	struct pte_ovfl *po, *npo;

	s = splvm();
	while (va < endva) {
		if (pm->pm_segtab[va >> ADDR_SR_SHFT].v == 0)
			goto next;
		idx = pteidx(sr = ptesr(pm->pm_segtab, va), va);
		for (ptp = ptable + idx * 8, i = 8; --i >= 0; ptp++)
			if (ptematch(ptp, sr, va, PTE_VALID)) {
				ptp->pte_hi &= ~PTE_VALID;
				asm volatile ("sync");
				tlbie(va);
				tlbsync();
				pmap_remove_pv(pm, idx, va, ptp->pte_lo, ptp);
				pm->pm_stats.resident_count--;
				goto next;
			}
		for (ptp = ptable + (idx ^ ptab_mask) * 8, i = 8; --i >= 0; ptp++)
			if (ptematch(ptp, sr, va, PTE_VALID | PTE_HID)) {
				ptp->pte_hi &= ~PTE_VALID;
				asm volatile ("sync");
				tlbie(va);
				tlbsync();
				pmap_remove_pv(pm, idx, va, ptp->pte_lo, ptp);
				pm->pm_stats.resident_count--;
				goto next;
			}
		for (po = LIST_FIRST(&potable[idx]); po; po = npo) {
			npo = LIST_NEXT(po, po_list);
			if (ptematch(&po->po_pte, sr, va, 0)) {
				LIST_REMOVE(po, po_list);
				tlbie(va);
				tlbsync();
				pmap_remove_pv(pm, idx, va, po->po_pte.pte_lo,
					       &po->po_pte);
				pofree(po, 1);
				pm->pm_stats.resident_count--;
				goto next;
			}
		}
next:
		va += NBPG;
	}
	splx(s);
}

void
pmap_kremove(va, len)
	vaddr_t va;
	vsize_t len;
{
	struct pmap *pm = pmap_kernel();
	vaddr_t endva = va + len;
	int idx, i, s;
	sr_t sr;
	pte_t *ptp;
	struct pte_ovfl *po, *npo;

	s = splvm();
	while (va < endva) {
		idx = pteidx(sr = ptesr(pm->pm_segtab, va), va);
		for (ptp = ptable + idx * 8, i = 8; --i >= 0; ptp++)
			if (ptematch(ptp, sr, va, PTE_VALID)) {
				ptp->pte_hi &= ~PTE_VALID;
				asm volatile ("sync");
				tlbie(va);
				tlbsync();
				pm->pm_stats.resident_count--;
				goto next;
			}
		for (ptp = ptable + (idx ^ ptab_mask) * 8, i = 8; --i >= 0; ptp++)
			if (ptematch(ptp, sr, va, PTE_VALID | PTE_HID)) {
				ptp->pte_hi &= ~PTE_VALID;
				asm volatile ("sync");
				tlbie(va);
				tlbsync();
				pm->pm_stats.resident_count--;
				goto next;
			}
		for (po = LIST_FIRST(&potable[idx]); po; po = npo) {
			npo = LIST_NEXT(po, po_list);
			if (ptematch(&po->po_pte, sr, va, 0)) {
				LIST_REMOVE(po, po_list);
				pofree(po, 1);
				tlbie(va);
				tlbsync();
				pm->pm_stats.resident_count--;
				goto next;
			}
		}
next:
		va += NBPG;
	}
	splx(s);
}

static pte_t *
pte_find(pm, va)
	struct pmap *pm;
	vaddr_t va;
{
	int idx, i;
	sr_t sr;
	pte_t *ptp;
	struct pte_ovfl *po;

	if (pm->pm_segtab[va >> ADDR_SR_SHFT].v == 0)
		return 0;
	idx = pteidx(sr = ptesr(pm->pm_segtab, va), va);
	for (ptp = ptable + idx * 8, i = 8; --i >= 0; ptp++)
		if (ptematch(ptp, sr, va, PTE_VALID))
			return ptp;
	for (ptp = ptable + (idx ^ ptab_mask) * 8, i = 8; --i >= 0; ptp++)
		if (ptematch(ptp, sr, va, PTE_VALID | PTE_HID))
			return ptp;
	for (po = LIST_FIRST(&potable[idx]); po; po = LIST_NEXT(po, po_list))
		if (ptematch(&po->po_pte, sr, va, 0))
			return &po->po_pte;
	return 0;
}

/*
 * Get the physical page address for the given pmap/virtual address.
 */
bool
pmap_extract(pm, va, pap)
	struct pmap *pm;
	vaddr_t va;
	paddr_t *pap;
{
	pte_t *ptp;
	int s = splvm();

	if (!(ptp = pte_find(pm, va))) {
		splx(s);
		return FALSE;
	}
	if (pap)
		*pap = (ptp->pte_lo & PTE_RPGN) | (va & PGOFSET);
	splx(s);
	return TRUE;
}

/*
 * Lower the protection on the specified range of this pmap.
 *
 * There are only two cases: either the protection is going to 0,
 * or it is going to read-only.
 */
void
pmap_protect(pm, sva, eva, prot)
	struct pmap *pm;
	vaddr_t sva, eva;
	vm_prot_t prot;
{
	pte_t *ptp;
	int valid, s;

	if (prot & VM_PROT_READ) {
		s = splvm();
		while (sva < eva) {
			if ((ptp = pte_find(pm, sva)) != NULL) {
				valid = ptp->pte_hi & PTE_VALID;
				ptp->pte_hi &= ~PTE_VALID;
				asm volatile ("sync");
				tlbie(sva);
				tlbsync();
				ptp->pte_lo &= ~PTE_PP;
				ptp->pte_lo |= PTE_RO;
				asm volatile ("sync");
				ptp->pte_hi |= valid;
			}
			sva += NBPG;
		}
		splx(s);
		return;
	}
	pmap_remove(pm, sva, eva);
}

boolean_t
ptemodify(pg, mask, val)
	struct vm_page *pg;
	u_int mask;
	u_int val;
{
	paddr_t pa = VM_PAGE_TO_PHYS(pg);
	struct pv_entry *pv;
	pte_t *ptp;
	struct pte_ovfl *po;
	int i, s;
	char *attr;
	int rv;

	/*
	 * First modify bits in cache.
	 */
	attr = pa_to_attr(pa);
	if (attr == NULL)
		return FALSE;

	rv = (*attr << ATTRSHFT) & mask;
	*attr &= ~(mask >> ATTRSHFT);
	*attr |= val >> ATTRSHFT;

	pv = pa_to_pv(pa);
	if (pv->pv_pmap == NULL)
		return rv != 0;

	s = splvm();
	for (; pv; pv = pv->pv_next) {
		ptp = ptable + pv->pv_idx * 8;
		for (i = 8; --i >= 0; ptp++)
			if ((ptp->pte_lo & PTE_RPGN) == pa &&
			    (ptp->pte_hi & PTE_VALID)) {
				ptp->pte_hi &= ~PTE_VALID;
				asm volatile ("sync");
				tlbie(pv->pv_va);
				tlbsync();
				rv |= ptp->pte_lo & mask;
				ptp->pte_lo &= ~mask;
				ptp->pte_lo |= val;
				asm volatile ("sync");
				ptp->pte_hi |= PTE_VALID;
				goto next;
			}
		ptp = ptable + (pv->pv_idx ^ ptab_mask) * 8;
		for (i = 8; --i >= 0; ptp++)
			if ((ptp->pte_lo & PTE_RPGN) == pa &&
			    (ptp->pte_hi & PTE_VALID)) {
				ptp->pte_hi &= ~PTE_VALID;
				asm volatile ("sync");
				tlbie(pv->pv_va);
				tlbsync();
				rv |= ptp->pte_lo & mask;
				ptp->pte_lo &= ~mask;
				ptp->pte_lo |= val;
				asm volatile ("sync");
				ptp->pte_hi |= PTE_VALID;
				goto next;
			}
		for (po = LIST_FIRST(&potable[pv->pv_idx]); po; po = LIST_NEXT(po, po_list))
			if ((po->po_pte.pte_lo & PTE_RPGN) == pa) {
				rv |= po->po_pte.pte_lo & mask;
				po->po_pte.pte_lo &= ~mask;
				po->po_pte.pte_lo |= val;
				tlbie(pv->pv_va);
				tlbsync();
				goto next;
			}
#ifdef DIAGNOSTIC
		panic("ptemodify: not on ptable");
#endif

next:
		continue;
	}
	splx(s);
	return rv != 0;
}

int
ptebits(pg, bit)
	struct vm_page *pg;
	int bit;
{
	struct pv_entry *pv;
	pte_t *ptp;
	struct pte_ovfl *po;
	int i, s, bits = 0;
	char *attr;
	paddr_t pa = VM_PAGE_TO_PHYS(pg);

	/*
	 * First try the cache.
	 */
	attr = pa_to_attr(pa);
	if (attr == NULL)
		return 0;
	bits |= (*attr << ATTRSHFT) & bit;
	if (bits == bit)
		return bits;

	pv = pa_to_pv(pa);
	if (pv->pv_pmap == NULL)
		return 0;

	s = splvm();
	for (; pv; pv = pv->pv_next) {
		for (ptp = ptable + pv->pv_idx * 8, i = 8; --i >= 0; ptp++)
			if ((ptp->pte_hi & PTE_VALID)
			    && (ptp->pte_lo & PTE_RPGN) == pa) {
				bits |= ptp->pte_lo & bit;
				if (bits == bit) {
					splx(s);
					return bits;
				}
			}
		for (ptp = ptable + (pv->pv_idx ^ ptab_mask) * 8, i = 8; --i >= 0; ptp++)
			if ((ptp->pte_hi & PTE_VALID)
			    && (ptp->pte_lo & PTE_RPGN) == pa) {
				bits |= ptp->pte_lo & bit;
				if (bits == bit) {
					splx(s);
					return bits;
				}
			}
		for (po = LIST_FIRST(&potable[pv->pv_idx]); po; po = LIST_NEXT(po, po_list))
			if ((po->po_pte.pte_lo & PTE_RPGN) == pa) {
				bits |= po->po_pte.pte_lo & bit;
				if (bits == bit) {
					splx(s);
					return bits;
				}
			}
	}
	splx(s);
	return bits;
}

/*
 * Lower the protection on the specified physical page.
 *
 * There are only two cases: either the protection is going to 0,
 * or it is going to read-only.
 */
void
pmap_page_protect(pg, prot)
	struct vm_page *pg;
	vm_prot_t prot;
{
	paddr_t pa = VM_PAGE_TO_PHYS(pg);
	vaddr_t va;
	pte_t *ptp;
	struct pte_ovfl *po, *npo;
	int i, s, idx;
	struct pv_entry *pv;
	struct pmap *pm;

	pa &= ~PGOFSET;
	if (prot & VM_PROT_READ) {
		ptemodify(pg, PTE_PP, PTE_RO);
		return;
	}

	pv = pa_to_pv(pa);
	if (pv == NULL)
		return;

	s = splvm();
	while (pv->pv_pmap) {
		idx = pv->pv_idx;
		va = pv->pv_va;
		pm = pv->pv_pmap;
		for (ptp = ptable + idx * 8, i = 8; --i >= 0; ptp++)
			if ((ptp->pte_hi & PTE_VALID)
			    && (ptp->pte_lo & PTE_RPGN) == pa) {
				ptp->pte_hi &= ~PTE_VALID;
				asm volatile ("sync");
				tlbie(va);
				tlbsync();
				pmap_remove_pv(pm, idx, va, pa, ptp);
				pm->pm_stats.resident_count--;
				goto next;
			}
		for (ptp = ptable + (idx ^ ptab_mask) * 8, i = 8; --i >= 0; ptp++)
			if ((ptp->pte_hi & PTE_VALID)
			    && (ptp->pte_lo & PTE_RPGN) == pa) {
				ptp->pte_hi &= ~PTE_VALID;
				asm volatile ("sync");
				tlbie(va);
				tlbsync();
				pmap_remove_pv(pm, idx, va, pa, ptp);
				pm->pm_stats.resident_count--;
				goto next;
			}
		for (po = LIST_FIRST(&potable[idx]); po; po = npo) {
			npo = LIST_NEXT(po, po_list);
			if ((po->po_pte.pte_lo & PTE_RPGN) == pa) {
				LIST_REMOVE(po, po_list);
				tlbie(va);
				tlbsync();
				pmap_remove_pv(pm, idx, va, pa, &po->po_pte);
				pofree(po, 1);
				pm->pm_stats.resident_count--;
				goto next;
			}
		}
		panic("pmap_page_protect: not found");
next:
		continue;
	}
	splx(s);
}

/*
 * Activate the address space for the specified process.  If the process
 * is the current process, load the new MMU context.
 */
void
pmap_activate(l)
	struct lwp *l;
{
	struct pcb *pcb = l->l_addr;
	struct pmap *pmap = l->l_proc->p_vmspace->vm_map.pmap;
	int psl;

	/*
	 * XXX Normally performed in cpu_fork().
	 */
	if (pcb->pcb_pm != pmap) {
		pcb->pcb_pm = pmap;
		pcb->pcb_pmreal = (void *)vtophys((vaddr_t)pmap);
	}

	if (l == curlwp) {
		/* Disable interrupts while switching. */
		asm volatile ("mfmsr %0" : "=r"(psl));
		psl &= ~PSL_EE;
		asm volatile ("mtmsr %0" :: "r"(psl));

		/* Store pointer to new current pmap. */
		curpm = pcb->pcb_pmreal;

		/* Set new segment table. */
		asm volatile ("mtasr %0" :: "r"(pmap->pm_segreal));

		/* Restore USER_SR in case of copyin/copyout. */
		ksegtab[USER_SR].vsid = pcb->pcb_usersr;
		asm volatile ("sync; slbie %0; isync" :: "r"(USER_ADDR));

		/* Interrupts are OK again. */
		psl |= PSL_EE;
		asm volatile ("mtmsr %0" :: "r"(psl));
	}
}

/*
 * Deactivate the specified process's address space.
 */
void
pmap_deactivate(l)
	struct lwp *l;
{
	struct pcb *pcb = l->l_addr;

	if (l == curlwp)
		pcb->pcb_usersr = ksegtab[USER_SR].vsid;
}

/*
 * Synchronize caches corresponding to [addr, addr+len) in p.
 */
void
pmap_procwr(p, va, len)
	struct proc *p;
	vaddr_t va;
	size_t len;
{
	paddr_t pa;
	int rv;

	rv = pmap_extract(p->p_vmspace->vm_map.pmap, va, &pa);
	if (rv == 0) {
#ifdef DIAGNOSTIC
		printf("pmap_procwr: bogus va 0x%lx\n", va);
#endif
		return;
	}
	__syncicache((void *)pa, len);
}

void *
pmap_pool_palloc(pp, flags)
	struct pool *pp;
	int flags;
{
	struct vm_page *pg;
	void *addr;

	pg = uvm_pagealloc(NULL, 0, NULL, UVM_PGA_USERESERVE|UVM_PGA_ZERO);
	if (pg == NULL)
		return NULL;

	addr = (void *)VM_PAGE_TO_PHYS(pg);

#ifdef DIAGNOSTIC
	/* simple_lock(&pg->mdpage.pvh_slock); */
	if (pg->wire_count != 0) {
		printf("pmap_pool_palloc: page %p has %d references\n",
		    addr, pg->wire_count);
		panic("pmap_physpage_alloc");
	}
	/* simple_unlock(&pg->mdpage.pvh_slock); */
#endif
	return addr;
}

void
pmap_pool_prelease(pp, v)
	struct pool *pp;
	void *v;
{
	struct vm_page *pg;

	if ((pg = PHYS_TO_VM_PAGE((paddr_t)v)) == NULL)
		panic("pmap_pool_prelease: bogus physical page address");

#ifdef DIAGNOSTIC
	/* simple_lock(&pg->mdpage.pvh_slock); */
	if (pg->wire_count != 0)
		panic("pmap_pool_prelease: page still has references");
	/* simple_unlock(&pg->mdpage.pvh_slock); */
#endif

	uvm_pagefree(pg);
}

paddr_t
vtophys(va)
	vaddr_t va;
{
	paddr_t pa;

	if (va < 0x7fffffff)
		return va;

	if (pmap_extract(pmap_kernel(), va, &pa))
		return pa;

	panic("vtophys va = 0x%lx", va);
}
