/*	$Id: obio.c,v 1.3 2007/10/11 09:07:37 tsubai Exp $	*/

/*-
 * Copyright (c) 2006 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/ps3.h>

int obio_match(struct device *, struct cfdata *, void *);
void obio_attach(struct device *, struct device *, void *);

CFATTACH_DECL0(obio, sizeof(struct device), obio_match, obio_attach);

int
obio_match(parent, cf, aux)
	struct device *parent;
	struct cfdata *cf;
	void *aux;
{
	struct confargs *ca = aux;

	if (strcmp(ca->ca_name, "obio") == 0)
		return 1;

	return 0;
}

void
obio_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct confargs ca;
	long bus, n;
	int i, j, k;

	printf("\n");

	for (i = 1; i <= 4; i++) {
		if (ps3_read_bus_id(i, &bus))
			continue;
		if (ps3_read_ndev(i, &n))
			continue;

		for (j = 0; j < n; j++) {
			bzero(&ca, sizeof ca);
			ca.ca_name = "";
			ca.ca_dev = ca.ca_type = -1;
			ca.ca_bus = bus;

			ps3_read_device_id  (i, j, &ca.ca_dev);
			ps3_read_device_type(i, j, &ca.ca_type);
#if 0
			ps3_read_pci_bus    (i, j, &ca.ca_pcibus);
			ps3_read_pci_device (i, j, &ca.ca_pcidev);
			ps3_read_pci_func   (i, j, &ca.ca_pcifunc);
#endif

			for (k = 0; k < 4; k++) {
				ca.ca_intr = ca.ca_intrtype = -1;
				ca.ca_func = k;
				if (ps3_read_interrupt(i, j, k,
				    &ca.ca_intrtype, &ca.ca_intr))
					break;

				config_found(self, &ca, NULL);
			}
		}
	}
}

void obio_devdump(void);
static int dump_devices(int);

void
obio_devdump()
{
	int i;

	return;

	for (i = 1; i <= 4; i++)
		dump_devices(i);
}

int
dump_devices(bus_index)
	int bus_index;
{
	int i, dev_index;
	u_long bus_id, ndev;
	u_long device_id, device_type;
	u_long irq_type, irq_number;

	bus_id = -1;
	ndev = 0;

	printf("bus_index = %d\n", bus_index);
	if (ps3_read_bus_id(bus_index, &bus_id))
		return -1;

	if (ps3_read_ndev(bus_index, &ndev))
		return -1;

	printf("bus_id = %ld, ndev = %ld\n", bus_id, ndev);

	for (dev_index = 0; dev_index < ndev; dev_index++) {
		device_id = device_type = -1;
		ps3_read_device_id(bus_index, dev_index, &device_id);
		ps3_read_device_type(bus_index, dev_index, &device_type);

		printf("%ld %3ld   ", device_id, device_type);

		for (i = 0; i < 4; i++) {
			irq_number = -1;
			ps3_read_interrupt(bus_index, dev_index, 
			    i, &irq_type, &irq_number);
			printf(" %ld", irq_number);
		}
		printf("\n");
	}
	return 0;
}

#if 0

* bus_index, dev_index ɥ쥹롣
* bus  dev  2ˤʤäƤ롣function 3?
* bus_id, dev_id ϥϥɥʤ

 bus_index  1 - 4
 1 internal PCI
 2 external PCI
 3 external PCIe
 4 disk

bus_id = 1, n = 4  type/num
0 3 -1 -1 -1	2/0		 gelic
1 4 -1 -1 -1	3/16 4/10	 usb	ohci  ehci ()
2 4 -1 -1 -1	3/17 4/11	 usb	ohci  ehci (ɥ꡼)
3 6 -1 -1 -1	5/47

bus_id = 2, n = 0

bus_id = 3, n = 0

bus_id = 4, n = 4
0 255 -1 -1 -1	2/0
1 14  -1 -1 -1	2/0
2 0   -1 -1 -1	2/0
3 5   -1 -1 -1	2/0

#endif
