/*	$Id: cpu.h,v 1.8 2009/12/20 13:02:39 tsubai Exp $	*/

/*
 * Copyright (C) 1995-1997 Wolfgang Solfrank.
 * Copyright (C) 1995-1997 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef _MACHINE_CPU_H_
#define _MACHINE_CPU_H_

#if defined(_KERNEL_OPT)
#include "opt_multiprocessor.h"
#endif

#include <machine/frame.h>
#include <machine/psl.h>

#ifdef _KERNEL
#include <machine/intr.h>
#include <sys/cpu_data.h>

struct cpu_info {
	struct cpu_data ci_data;	/* MI per-cpu data */
	int ci_mtx_oldspl;		/* XXX ci_mtx_* are MI */
	int ci_mtx_count;

	struct lwp *ci_curlwp;		/* current owner of the processor */
	struct pcb *ci_curpcb;
	struct pmap *ci_curpm;
	struct lwp *ci_fpulwp;
	struct lwp *ci_veclwp;
	int ci_cpuid;

	volatile int ci_astpending;
	volatile int ci_want_resched;
	u_int ci_lasttb;
	int ci_tickspending;
	volatile int ci_cpl;
	volatile int ci_ipending;
	volatile int ci_idepth;
	int ci_intrdepth;
	char *ci_intstk;
	char *ci_spillstk;
	char *ci_ddbstk;
	int64_t ci_tempsave[8] __aligned(128);
	int64_t ci_ddbsave[8];
	int64_t ci_disisave[4];
	int64_t ci_segsave[4];
} __aligned(128);

static struct cpu_info *curcpu(void);

static __inline struct cpu_info *
curcpu()
{
	struct cpu_info *ci;

	asm ("mfsprg %0,0" : "=r"(ci));
	return ci;
}

extern struct cpu_info cpu_info[];

#ifdef MULTIPROCESSOR
void cpu_boot_secondary_processors(void);
static int cpu_number(void);

static __inline int
cpu_number()
{
	int pir;

	asm ("mfspr %0,1023" : "=r"(pir));
	return pir;
}

#define CPU_IS_PRIMARY(ci)	((ci)->ci_cpuid == 0)

#define CPU_INFO_ITERATOR int
#define CPU_INFO_FOREACH(cii, ci) \
	cii = 0, ci = &cpu_info[0]; \
	({ extern int ncpus; cii < ncpus; }); cii++, ci++

#else
#define cpu_number()		0

#endif /* MULTIPROCESSOR */

bool cpu_intr_p(void);

extern __inline bool
cpu_intr_p()
{
	return curcpu()->ci_idepth != 0;
}

void delay __P((unsigned));
#define DELAY(n)		delay(n)

extern char bootpath[];

#endif /* _KERNEL */

#if defined(_KERNEL) || defined(_STANDALONE)
#define CACHELINESIZE	128
#endif

#include <powerpc64/cpu.h>

#endif /* _MACHINE_CPU_H_ */
