/*	$Id: intr.h,v 1.4 2008/05/01 13:49:26 tsubai Exp $	*/

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Charles M. Hannum.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _PS3_INTR_H_
#define _PS3_INTR_H_

/* Interrupt priority `levels'. */
#define IPL_NONE	0	/* nothing */
#define IPL_SOFTCLOCK	1	/* timeouts */
#define IPL_SOFTBIO	2
#define IPL_SOFTNET	3	/* protocol stacks */
#define IPL_SOFTSERIAL	4	/* serial */
#define IPL_VM		5	/* memory allocation */
#define IPL_SCHED	6
#define IPL_HIGH	7	/* everything */
#define NIPL		8

/* Interrupt sharing types. */
#define IST_NONE	0	/* none */
#define IST_EDGE	1	/* edge-triggered */
#define IST_LEVEL	2	/* level-triggered */
#define IST_PULSE	3	/* pulsed */

#ifndef _LOCORE

/*
 * Interrupt handler chains.  intr_establish() inserts a handler into
 * the list.  The handler is called with its (single) argument.
 */
struct intrhand {
	int	(*ih_fun) __P((void *));
	void	*ih_arg;
	u_long	ih_count;
	struct	intrhand *ih_next;
	int	ih_level;
	int	ih_irq;
};

void clearsoftclock __P((void));
void clearsoftnet __P((void));
void softnet __P((void *));
void softserial __P((void *));

int splraise __P((int));
int spllower __P((int));
void splx __P((int));
void softintr __P((int));
void *softintr_establish(int, void (*)(void *), void *);
void softintr_disestablish(void *);
void softintr_schedule(void *);

extern volatile int astpending, tickspending;
extern long intrcnt[];

#define ICU_LEN		64
#define CNT_CLOCK	59

/* Soft interrupt masks. */
#define SIR_CLOCK	60
#define SIR_BIO		61
#define SIR_NET		62
#define SIR_SERIAL	63

/*
 * Software interrupt masks
 */
#define splsoftclock()	splraise(IPL_SOFTCLOCK)
#define splsoftbio()	splraise(IPL_SOFTBIO)
#define splsoftnet()	splraise(IPL_SOFTNET)
#define splsoftserial()	splraise(IPL_SOFTSERIAL)

/*
 * Miscellaneous
 */
#define splvm()		splraise(IPL_VM)
#define splhigh()	splraise(IPL_HIGH)
#define splsched()	splhigh()
#define spl0()		spllower(IPL_NONE)
#define splraiseipl(x)	splraise(x)

#define setsoftclock()	softintr(SIR_CLOCK)
#define setsoftbio()	softintr(SIR_BIO)
#define setsoftnet()	softintr(SIR_NET)
#define setsoftserial()	softintr(SIR_SERIAL)

#define makeiplcookie(ipl) (ipl)
typedef int ipl_cookie_t;

#endif /* _LOCORE */

#endif /* _PS3_INTR_H_ */
