/*	$Id: extintr.c,v 1.14 2010/12/30 22:34:17 tsubai Exp $	*/

/*-
 * Copyright (c) 2000-2004 Tsubai Masanari.
 * Copyright (c) 1995 Per Fogelstrom
 * Copyright (c) 1993, 1994 Charles M. Hannum.
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * William Jolitz and Don Ahn.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)isa.c	7.2 (Berkeley) 5/12/91
 */

#define NO_ATOMIC_OP 1

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/pool.h>

#include <uvm/uvm_extern.h>

#include <machine/autoconf.h>
#include <machine/hcalls.h>
#include <machine/intr.h>
#include <machine/psl.h>

#define NIRQ 32

void intr_calculatemasks __P((void));
int fakeintr __P((void *));

static inline int cntlzw __P((int));
static inline int mapirq __P((int));
static void sortirq(void);

static void do_pending_int __P((void));
void ext_intr __P((void));

int imask[NIPL];

int intrtype[NIRQ], intrmask[NIRQ], intrlevel[NIRQ];
struct intrhand *intrhand[NIRQ];

static u_char hwirq[NIRQ], virq[ICU_LEN];

#define ipending	(curcpu()->ci_ipending)
#define cpl		(curcpu()->ci_cpl)
#define idepth		(curcpu()->ci_idepth)

struct pool extintr_pool;

/* atomic.h? */
static void atomic_bis(volatile int *, int);
static void atomic_bic(volatile int *, int);

static inline void
atomic_bis(addr, val)
	volatile int *addr;
	int val;
{
#if NO_ATOMIC_OP
	*addr |= val;
#else
	int scratch;

	asm volatile ("		\
				\n\
1:	lwarx	%0,0,%1		\n\
	or	%0,%0,%2	\n\
	stwcx.	%0,0,%1		\n\
	bne-	1b		\n\
				\n"
	: "=&r"(scratch) : "r"(addr), "r"(val) : "memory");
#endif
}

static inline void
atomic_bic(addr, val)
	volatile int *addr;
	int val;
{
#if NO_ATOMIC_OP
	*addr &= ~val;
#else
	int scratch;

	asm volatile ("		\
				\n\
1:	lwarx	%0,0,%1		\n\
	andc	%0,%0,%2	\n\
	stwcx.	%0,0,%1		\n\
	bne-	1b		\n\
				\n"
	: "=&r"(scratch) : "r"(addr), "r"(val) : "memory");
#endif
}

/*
 * Map 64 irqs into 32 (bits).
 */
int
mapirq(irq)
	int irq;
{
	int v;

	if (irq < 0 || irq >= ICU_LEN)
		panic("invalid irq %d", irq);
	if (virq[irq])
		return virq[irq];

	for (v = 1; v < NIRQ; v++) {
		if (intrhand[v] == NULL)
			goto found;
	}
	panic("virq overflow");

found:
	hwirq[v] = irq;
	virq[irq] = v;

	return v;
}

/*
 * Count leading zeros.
 */
static __inline int
cntlzw(x)
	int x;
{
	int a;

	asm volatile ("cntlzw %0,%1" : "=r"(a) : "r"(x));

	return a;
}

/*
 * Sort IRQs by ih_level.
 * XXX lock
 */
void
sortirq()
{
	struct intrhand *new_intrhand[NIRQ];
	u_char new_hwirq[NIRQ], new_virq[ICU_LEN];
	int new_intrtype[NIRQ];
	struct intrhand *ih;
	int irqs[NIPL], irq, level;
	int newirq = 1;

	bzero(irqs, sizeof irqs);
	for (irq = 0; irq < NIRQ; irq++) {
		level = IPL_NONE;
		ih = intrhand[irq];
		if (ih == NULL)
			continue;
		while (ih) {
			if (ih->ih_level > level)
				level = ih->ih_level;
			ih = ih->ih_next;
		}
		irqs[level] |= 1 << irq;
	}

	bzero(new_intrhand, sizeof new_intrhand);
	bzero(new_intrtype, sizeof new_intrtype);
	bzero(new_hwirq, sizeof new_hwirq);
	bzero(new_virq, sizeof new_virq);

	for (level = IPL_NONE; level < NIPL; level++) {
		while (irqs[level]) {
			irq = 31 - cntlzw(irqs[level]);
			irqs[level] &= ~(1 << irq);

			new_hwirq[newirq] = hwirq[irq];
			new_virq[new_hwirq[newirq]] = newirq;
			new_intrtype[newirq] = intrtype[irq];
			new_intrhand[newirq] = intrhand[irq];
			ih = new_intrhand[newirq];
			while (ih) {
				if (ih->ih_irq != irq) panic("sortirq");
				ih->ih_irq = newirq;
				ih = ih->ih_next;
			}
			newirq++;
		}
	}
	bcopy(new_intrhand, intrhand, sizeof intrhand);
	bcopy(new_intrtype, intrtype, sizeof intrtype);
	bcopy(new_hwirq, hwirq, sizeof hwirq);
	bcopy(new_virq, virq, sizeof virq);
}

/*
 * Recalculate the interrupt masks from scratch.
 * We could code special registry and deregistry versions of this function that
 * would be faster, but the code would be nastier, and we don't expect this to
 * happen very much anyway.
 */
void
intr_calculatemasks()
{
	int irq, level;
	struct intrhand *q;

	/* First, figure out which levels each IRQ uses. */
	for (irq = 0; irq < NIRQ; irq++) {
		register int levels = 0;
		for (q = intrhand[irq]; q; q = q->ih_next)
			levels |= 1 << q->ih_level;
		intrlevel[irq] = levels;
	}

	/* Then figure out which IRQs use each level. */
	for (level = 0; level < NIPL; level++) {
		register int irqs = 0;
		for (irq = 0; irq < NIRQ; irq++)
			if (intrlevel[irq] & (1 << level))
				irqs |= 1 << irq;
		imask[level] = irqs;
	}

	/*
	 * IPL_NONE is used for hardware interrupts that are never blocked,
	 * and do not block anything else.
	 */
	imask[IPL_NONE] = 0;

	/*
	 * Enforce a hierarchy that gives slow devices a better chance at not
	 * dropping data.
	 */
	imask[IPL_SOFTCLOCK] |= imask[IPL_NONE];
	imask[IPL_SOFTBIO] |= imask[IPL_SOFTCLOCK];
	imask[IPL_SOFTNET] |= imask[IPL_SOFTBIO];
	imask[IPL_BIO] |= imask[IPL_SOFTNET];
	imask[IPL_NET] |= imask[IPL_BIO];
	imask[IPL_SOFTSERIAL] |= imask[IPL_NET];
	imask[IPL_TTY] |= imask[IPL_SOFTSERIAL];

	/*
	 * There are tty, network and disk drivers that use free() at interrupt
	 * time, so vm > (tty | net | bio).
	 */
	imask[IPL_VM] |= imask[IPL_TTY];

	imask[IPL_AUDIO] |= imask[IPL_VM];

	/*
	 * Since run queues may be manipulated by both the statclock and tty,
	 * network, and disk drivers, clock > vm.
	 */
	imask[IPL_CLOCK] |= imask[IPL_AUDIO];

	/*
	 * IPL_HIGH must block everything that can manipulate a run queue.
	 */
	imask[IPL_HIGH] |= imask[IPL_CLOCK];

	/*
	 * We need serial drivers to run at the absolute highest priority to
	 * avoid overruns, so serial > high.
	 */
	imask[IPL_SERIAL] |= imask[IPL_HIGH];

	/* And eventually calculate the complete masks. */
	for (irq = 0; irq < NIRQ; irq++) {
		intrmask[irq] = IPL_NONE;
		for (q = intrhand[irq]; q; q = q->ih_next)
			if (q->ih_level > intrmask[irq])
				intrmask[irq] = q->ih_level;
	}

	/* Lastly, enable IRQs actually in use. */
	for (irq = 0; irq < ICU_LEN; irq++)
		ps3pic_disable_irq(irq);
	for (irq = 0; irq < NIRQ; irq++) {
		/* Skip software interrupts. */
		if (intrtype[irq] == IST_NONE)
			continue;
		if (intrhand[irq]) {
			int hirq = hwirq[irq];

			ps3pic_enable_irq(hirq);
		}
	}
}

int
fakeintr(arg)
	void *arg;
{

	return 0;
}

#define	LEGAL_IRQ(x)	((x) >= 0 && (x) < NIRQ)

char *
intr_typename(type)
	int type;
{

	switch (type) {
	case IST_NONE :
		return "none";
	case IST_PULSE:
		return "pulsed";
	case IST_EDGE:
		return "edge-triggered";
	case IST_LEVEL:
		return "level-triggered";
	default:
		panic("intr_typename: invalid type %d", type);
	}
}

/*
 * Register an interrupt handler.
 */
void *
intr_establish(irq, type, level, ih_fun, ih_arg)
	int irq;
	int type;
	int level;
	int (*ih_fun) __P((void *));
	void *ih_arg;
{
	struct intrhand **p, *q, *ih;
	static struct intrhand fakehand = {fakeintr};
	static int initted;

	if (initted == 0) {
		initted = 1;
		pool_init(&extintr_pool, sizeof(struct intrhand), 0, 0, 0,
		    "intrpool", NULL, IPL_NONE);
	}

	irq = mapirq(irq);

	/* no point in sleeping unless someone can free memory. */
	ih = pool_get(&extintr_pool, cold ? PR_NOWAIT : PR_WAITOK);
	if (ih == NULL)
		panic("intr_establish: can't malloc handler info");

	if (!LEGAL_IRQ(irq) || type == IST_NONE)
		panic("intr_establish: bogus irq (%d) or type (%d)", irq, type);

	switch (intrtype[irq]) {
	case IST_NONE:
		intrtype[irq] = type;
		break;
	case IST_EDGE:
	case IST_LEVEL:
		if (type == intrtype[irq])
			break;
	case IST_PULSE:
		if (type != IST_NONE)
			panic("intr_establish: can't share %s with %s",
			    intr_typename(intrtype[irq]),
			    intr_typename(type));
		break;
	}

	/*
	 * Figure out where to put the handler.
	 * This is O(N^2), but we want to preserve the order, and N is
	 * generally small.
	 */
	for (p = &intrhand[irq]; (q = *p) != NULL; p = &q->ih_next)
		;

	/*
	 * Actually install a fake handler momentarily, since we might be doing
	 * this with interrupts enabled and don't want the real routine called
	 * until masking is set up.
	 */
	fakehand.ih_level = level;
	*p = &fakehand;

	intr_calculatemasks();

	/*
	 * Poke the real handler in now.
	 */
	ih->ih_fun = ih_fun;
	ih->ih_arg = ih_arg;
	ih->ih_next = NULL;
	ih->ih_level = level;
	ih->ih_irq = irq;
	*p = ih;

	sortirq();
	intr_calculatemasks();

	return ih;
}

/*
 * Deregister an interrupt handler.
 */
void
intr_disestablish(arg)
	void *arg;
{
	struct intrhand *ih = arg;
	int irq = ih->ih_irq;
	struct intrhand **p, *q;

	if (!LEGAL_IRQ(irq))
		panic("intr_disestablish: bogus irq %d", irq);

	/*
	 * Remove the handler from the chain.
	 * This is O(n^2), too.
	 */
	for (p = &intrhand[irq]; (q = *p) != NULL && q != ih; p = &q->ih_next)
		;
	if (q)
		*p = q->ih_next;
	else
		panic("intr_disestablish: handler not registered");
	pool_put(&extintr_pool, ih);

	intr_calculatemasks();

	if (intrhand[irq] == NULL)
		intrtype[irq] = IST_NONE;
}

void *
softintr_establish(level, fun, arg)
	int level;
	void (*fun)(void *);
	void *arg;
{
	struct intrhand *ih;
	int irq, s;

	switch (level) {
	case IPL_SOFTCLOCK:
		irq = SIR_CLOCK;
		break;
	case IPL_SOFTBIO:
		irq = SIR_BIO;
		break;
	case IPL_SOFTNET:
		irq = SIR_NET;
		break;
	case IPL_SOFTSERIAL:
		irq = SIR_SERIAL;
		break;
	default:
		panic("softintr_establish");
	}
	irq = mapirq(irq);

	ih = pool_get(&extintr_pool, cold ? PR_NOWAIT : PR_WAITOK);
	if (ih == NULL)
		panic("softintr_establish: can't malloc handler info");
	ih->ih_fun = (void *)fun;
	ih->ih_arg = arg;
	ih->ih_level = level;
	ih->ih_irq = irq;
	s = splhigh();
	ih->ih_next = intrhand[irq];
	intrhand[irq] = ih;
	splx(s);

	sortirq();
	intr_calculatemasks();

	return ih;
}

void
softintr_disestablish(arg)
	void *arg;
{
	intr_disestablish(arg);
}

void
softintr_schedule(arg)
	void *arg;
{
	struct intrhand *ih = arg;
	int irq = ih->ih_irq;

	switch (hwirq[irq]) {
	case SIR_CLOCK:  setsoftclock();  break;
	case SIR_BIO:    setsoftbio();    break;
	case SIR_NET:    setsoftnet();    break;
	case SIR_SERIAL: setsoftserial(); break;
	default:
		panic("softintr_schedule");
	}
}

int cpuintr(void *);

/*
 * external interrupt handler
 */
void
ext_intr()
{
	int irq, realirq;
	int pcpl, msr;
	int r_imen;
	struct intrhand *ih;

#ifdef MULTIPROCESSOR
	/* Only cpu0 can handle interrupts. */
	if (cpu_number() != 0) {
		irq = ps3pic_read_irq(cpu_number());
		if (irq != 240)
			panic("secondary cpu");
		cpuintr(NULL);
		ps3pic_eoi(cpu_number());
		return;
	}
#endif

	pcpl = cpl;
	asm volatile ("mfmsr %0" : "=r"(msr));

	realirq = ps3pic_read_irq();

	if (realirq == -1) {
		/* printf("spurious interrupt\n"); */
		return;
	}
#ifdef MULTIPROCESSOR
	if (realirq == 240) {
		cpuintr(NULL);
		ps3pic_eoi(0);

		realirq = ps3pic_read_irq(0);
		if (realirq == 255)
			return;
	}
#endif

start:
	irq = virq[realirq];		/* XXX check range */
	r_imen = 1 << irq;

	if (imask[pcpl] & r_imen) {
		/* Masked!  Mark this as pending. */
		atomic_bis(&ipending, r_imen);
		ps3pic_disable_irq(realirq);
	} else {
		idepth++;
		splraise(intrmask[irq]);
		asm volatile ("mtmsr %0" :: "r"(msr | PSL_EE));
		ih = intrhand[irq];
		while (ih) {
			if (ih->ih_level == IPL_VM) {
				KERNEL_LOCK(1, NULL);
				(*ih->ih_fun)(ih->ih_arg);
				KERNEL_UNLOCK_ONE(NULL);
			} else
				(*ih->ih_fun)(ih->ih_arg);
			ih = ih->ih_next;
		}
		asm volatile ("mtmsr %0" :: "r"(msr));
		cpl = pcpl;

		curcpu()->ci_data.cpu_nintr++;
		intrcnt[realirq]++;
		idepth--;
	}

	ps3pic_eoi(realirq);

	realirq = ps3pic_read_irq();
	if (realirq != -1)
		goto start;

	asm volatile ("mtmsr %0" :: "r"(msr | PSL_EE));
	splx(pcpl);	/* Process pendings. */
	asm volatile ("mtmsr %0" :: "r"(msr));
}

static void
do_pending_int()
{
	struct intrhand *ih;
	int irq;
	int pcpl;
	int hwpend;
	int emsr, dmsr;
	const int cpu_id = cpu_number();
	static int processing[2];	/* XXX */

	if (processing[cpu_id])
		return;

	processing[cpu_id] = 1;
	asm volatile ("mfmsr %0" : "=r"(emsr));
	dmsr = emsr & ~PSL_EE;
	asm volatile ("mtmsr %0" :: "r"(dmsr));

	pcpl = cpl;

	/* Do now unmasked pendings */
	idepth++;
	while ((hwpend = ipending & ~imask[pcpl]) != 0) {
		irq = 31 - cntlzw(hwpend);
#ifdef MULTIPROCESSOR
		if (cpu_id != 0 && intrtype[irq] != IST_NONE)
			panic("hardware interrupt on secondary cpu");
#endif
		atomic_bic(&ipending, 1 << irq);
		splraise(intrmask[irq]);
		asm volatile ("mtmsr %0" :: "r"(emsr));
		ih = intrhand[irq];
		while(ih) {
			if (ih->ih_level == IPL_VM) {
				KERNEL_LOCK(1, NULL);
				(*ih->ih_fun)(ih->ih_arg);
				KERNEL_UNLOCK_ONE(NULL);
			} else
				(*ih->ih_fun)(ih->ih_arg);
			ih = ih->ih_next;
		}
		asm volatile ("mtmsr %0" :: "r"(dmsr));
		cpl = pcpl;

		intrcnt[hwirq[irq]]++;
		if (intrtype[irq] != IST_NONE)
			ps3pic_enable_irq(hwirq[irq]);
	}
	idepth--;

	cpl = pcpl;	/* Don't use splx... we are here already! */
	processing[cpu_id] = 0;
	asm volatile ("mtmsr %0" :: "r"(emsr));
}

int
splraise(ncpl)
	int ncpl;
{
	int ocpl;

	ocpl = cpl;
	if (ncpl > ocpl)
		cpl = ncpl;
	return ocpl;
}

void
splx(ncpl)
	int ncpl;
{

	cpl = ncpl;
	if (ipending & ~imask[ncpl])
		do_pending_int();
}

int
spllower(ncpl)
	int ncpl;
{
	int ocpl;

	ocpl = cpl;
	cpl = ncpl;
	if (ipending & ~imask[ncpl])
		do_pending_int();
	return ocpl;
}

/* Following code should be implemented with lwarx/stwcx to avoid
 * the disable/enable. i need to read the manual once more.... */
void
softintr(ipl)
	int ipl;
{
#if NO_ATOMIC_OP
	int msrsave;
#endif

	switch (ipl) {
	case SIR_CLOCK:
	case SIR_BIO:
	case SIR_NET:
	case SIR_SERIAL:
		break;
	default:
		panic("softintr");
	}
	ipl = virq[ipl];
#if NO_ATOMIC_OP
	asm volatile ("mfmsr %0" : "=r"(msrsave));
	asm volatile ("mtmsr %0" :: "r"(msrsave & ~PSL_EE));
#endif
	atomic_bis(&ipending, 1 << ipl);
#if NO_ATOMIC_OP
	asm volatile ("mtmsr %0" :: "r"(msrsave));
#endif
}

void
init_interrupt()
{
	ps3pic_init_irq();
	install_extint(ext_intr);
}

int
ps3_allocate_irq()
{
	/* XXX linux uses 128-255. */
	static int irq = 49;

	if (irq >= 60)
		panic("ps3_allocate_irq");

	return irq++;
}

int
ps3_connect_irq(bus_id, dev_id, intr_id)
	u_long bus_id, dev_id, intr_id;
{
	u_long outlet_id;
	int irq;

	irq = ps3_allocate_irq();
	if (HV_alloc_intrhand(&outlet_id))
		return -1;
	if (HV_set_device_intrhand(bus_id, dev_id, outlet_id, intr_id))
		return -1;
	if (HV_set_intrhand_irq(irq, outlet_id))
		return -1;

	return irq;
}

int
ps3_connect_io_irq(irq)
	u_long irq;
{
	u_long outlet_id;

	if (irq == 0)
		return -1;
	if (HV_alloc_intrhand_by_irq(irq, &outlet_id))
		return -1;
	if (HV_set_intrhand_irq(irq, outlet_id))
		return -1;

	return irq;
}
