/*	$Id: identifycpu.c,v 1.2 2008/05/07 17:30:02 tsubai Exp $	*/

/*-
 * Copyright (c) 2001, 2002, 2005 Tsubai Masanari.
 * Copyright (c) 1998, 1999, 2001 Internet Research Institute, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by
 *	Internet Research Institute, Inc.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/autoconf.h>
#include <machine/cpufunc.h>

#ifdef _
#error XXX
#endif

#ifdef CPUVERBOSE
#define _(name, mlen, vers, rev) { (vers) << 16 | (rev), mlen, #name "\0" }
struct cputab {
	u_int32_t version;
	char mlen;
	char name[7];
};
#else
#define _(name, mlen, vers, rev) { vers, #name "\0" }
struct cputab {
	u_int16_t version;
	char name[6];
};
#endif

static const struct cputab models[] = {
   _( 970,    16, PPC970, 0 ),
   _( 970FX,  16, PPC970FX, 0 ),
   _( 970MP,  16, PPC970MP, 0 ),
   _( Cell,   16, CELL, 0 ),

    { 0 }
};
#undef _

#define ppc750fx_probe() 0

void
identifycpu(cpu_model)
	char *cpu_model;
{
	u_int pvr, vers, rev;
	const struct cputab *cp = models;

	pvr = mfpvr();
	vers = pvr >> 16;
	rev = pvr & 0xffff;
	if (vers == MPC750) {
		/* XXX */
		if (ppc750fx_probe()) {
			vers = ppc750fx_probe();
			pvr = vers << 16 | rev;
		}
	}

	sprintf(cpu_model, "Version %x (Revision %x)", vers, rev);

	while (cp->version) {
#ifdef CPUVERBOSE
		if (cp->version >> (32 - cp->mlen) == pvr >> (32 - cp->mlen))
#else
		if (cp->version == vers)
#endif
			sprintf(cpu_model, "%s (Revision %x)", cp->name, rev);
		cp++;
	}
}
