/*	$Id: machdep.c,v 1.31 2011/10/20 07:12:49 tsubai Exp $	*/

/*
 * Copyright (C) 1995, 1996 Wolfgang Solfrank.
 * Copyright (C) 1995, 1996 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "opt_altivec.h"
#include "opt_compat_netbsd.h"
#include "opt_compat_netbsd32.h"
#include "opt_ddb.h"
#include "opt_ipkdb.h"
#include "opt_modular.h"
#include "opt_multiprocessor.h"
#include "ksyms.h"

#include <sys/param.h>
#include <sys/boot_flag.h>
#include <sys/buf.h>
#include <sys/cpu.h>
#include <sys/device.h>
#include <sys/kernel.h>
#include <sys/ksyms.h>
#include <sys/mount.h>
#include <sys/msgbuf.h>
#include <sys/reboot.h>
#include <sys/sysctl.h>
#include <sys/systm.h>

#include <machine/db_machdep.h>
#include <ddb/db_extern.h>

#include <machine/autoconf.h>
#include <machine/cpufunc.h>
#include <machine/pcb.h>
#include <machine/powerpc.h>
#include <machine/spr.h>
#include <machine/trap.h>
#include <machine/hcalls.h>
#include <machine/ps3.h>

#include <dev/cons.h>

#include <dev/wscons/wsksymvar.h>
#include <dev/wscons/wscons_callbacks.h>

#include <dev/usb/ukbdvar.h>

/*
 * Global variables used here and there
 */
struct vm_map *mb_map = NULL;
struct vm_map *phys_map = NULL;

struct pcb *proc0paddr;
char bootpath[256];
paddr_t msgbuf_paddr;

void initppc(u_int, u_int, long, char *, char *);
static void dumpsys(void);

static void get_memsize(void);

void
initppc(startkernel, endkernel, magic, ssym, esym)
	u_int startkernel, endkernel;
	long magic;
	char *ssym, *esym;
{
	extern int trapcode[], alitrap[], dsitrap[], isitrap[];
	extern int decrint[], perftrap[], syscalltrap[];
	extern int dsegtrap[], isegtrap[];
	extern int extint[];
#ifdef DDB
	extern int ddblow[];
#endif
#ifdef IPKDB
	extern int ipkdblow[];
#endif
	int exc, scratch;
	struct cpu_info *ci = &cpu_info[0];

	asm volatile ("mtsprg 0,%0" :: "r"(ci));

	/* Enter secondary thread into halt_secondary(). */
	*(volatile int *)0x20 = 1;
	asm volatile ("sync");
	while (*(volatile int *)0x20 != 0);

	/*
	 * i386 port says, that this shouldn't be here,
	 * but I really think the console should be initialized
	 * as early as possible.
	 */
	consinit();

	lwp0.l_cpu = ci;
	lwp0.l_addr = proc0paddr;
	bzero(lwp0.l_addr, sizeof(struct pcb));

	curlwp = &lwp0;
	curpcb = proc0paddr;

	curpm = curpcb->pcb_pmreal = curpcb->pcb_pm = pmap_kernel();

	/*
	 * Set up trap vectors
	 */
	for (exc = 0x100; exc <= EXC_LAST; exc += 0x100)
		bcopy(trapcode, (void *)(long)exc, 0x100);

	bcopy(alitrap, (void *)EXC_ALI, 0x100);
	bcopy(dsitrap, (void *)EXC_DSI, 0x100);
	bcopy(isitrap, (void *)EXC_ISI, 0x100);
	bcopy(decrint, (void *)EXC_DECR, 0x100);
	bcopy(perftrap, (void *)EXC_PERF, 0x100); /* EXC_PERF and EXC_VEC */
	bcopy(syscalltrap, (void *)EXC_SC, 0x100);
	bcopy(extint, (void *)EXC_EXI, 0x100);
#ifdef IPKDB
	bcopy(ipkdblow, (void *)EXC_PGM, 0x100);
	bcopy(ipkdblow, (void *)EXC_TRC, 0x100);
	bcopy(ipkdblow, (void *)EXC_BPT, 0x100);
#endif
#ifdef DDB
	bcopy(ddblow, (void *)EXC_PGM, 0x100);
	bcopy(ddblow, (void *)EXC_TRC, 0x100);
	bcopy(ddblow, (void *)EXC_BPT, 0x100);
#endif
	bcopy(dsegtrap, (void *)EXC_DSEG, 0x80);
	bcopy(isegtrap, (void *)EXC_ISEG, 0x80);

	/*
	 * Cause a PGM trap if we branch to 0.
	 */
	bzero(0, 0x100);
	memset(0, 0xff, 4);

	__syncicache((void *)EXC_RSVD, EXC_LAST - EXC_RSVD + 0x100);

	/*
	 * Parse arg string.
	 */

	/* XXX */

#if NKSYMS || defined(DDB) || defined(MODULAR)
	ksyms_addsyms_elf(esym - ssym, ssym, esym);
#endif

#ifdef DDB
	if (boothowto & RB_KDB)
		Debugger();
#endif

	/*
	 * Set the page size.
	 */
	uvm_setpagesize();

	get_memsize();

	/*
	 * Initialize pmap module.
	 */
	pmap_bootstrap(startkernel, endkernel);

	/*
	 * Now enable translation (and machine checks/recoverable interrupts).
	 */
	asm volatile ("mfmsr %0; ori %0,%0,%1; mtmsr %0; isync"
		      : "=r"(scratch) : "K"(PSL_IR|PSL_DR|PSL_ME|PSL_RI));
}

/*
 * This should probably be in autoconf!				XXX
 */
char cpu_model[80];
char machine[] = MACHINE;		/* from <machine/param.h> */
char machine_arch[] = MACHINE_ARCH;	/* from <machine/param.h> */

#ifdef COMPAT_NETBSD32
/* Provide a the name of the architecture we're emulating */
const char machine32[] = MACHINE;
const char machine_arch32[] = "powerpc";
#endif

#ifdef UCONS
void ucons_consinit(void);
int uplcom_cngetc(dev_t);
#endif

/*
 * Machine dependent startup code.
 */
void
cpu_startup()
{
	vaddr_t minaddr, maxaddr;
	char pbuf[9];

#ifdef UCONS
	ucons_consinit();
#endif
	initmsgbuf((caddr_t)msgbuf_paddr, round_page(MSGBUFSIZE));

	lwp0.l_addr = proc0paddr;

	printf("%s%s", copyright, version);
	identifycpu(cpu_model);

	format_bytes(pbuf, sizeof(pbuf), ctob((u_long)physmem));
	printf("total memory = %s\n", pbuf);

	minaddr = 0;

	/*
	 * Allocate a submap for physio
	 */
	phys_map = uvm_km_suballoc(kernel_map, &minaddr, &maxaddr,
	    VM_PHYS_SIZE, 0, FALSE, NULL);

	/*
	 * No need to allocate an mbuf cluster submap.  Mbuf clusters
	 * are allocated via the pool allocator, and we use direct-mapped
	 * pool pages.
	 */

	format_bytes(pbuf, sizeof(pbuf), ptoa(uvmexp.free));
	printf("avail memory = %s\n", pbuf);
}

/*
 * consinit
 * Initialize system console.
 */
void
consinit()
{
	static int initted;

	if (initted)
		return;
	initted = 1;
	cninit();
}

/*
 * Machine dependent system variables.
 */
#if 0
static int
sysctl_machdep_cacheinfo(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	static struct cache_info ci = {
		CACHELINESIZE, CACHELINESIZE,		/* XXX cache_size */
		CACHELINESIZE, CACHELINESIZE,
	};

	node.sysctl_data = &ci;
	node.sysctl_size = sizeof ci;
	return sysctl_lookup(SYSCTLFN_CALL(&node));
}
#endif

static int
sysctl_machdep_booted_kernel(SYSCTLFN_ARGS)
{
	struct sysctlnode node = *rnode;
	char *cp;

	cp = strrchr(bootpath, '/');
	if (cp == NULL)
		cp = bootpath;
	if (*cp == '/')
		cp++;
	node.sysctl_data = cp;
	node.sysctl_size = strlen(cp) + 1;
	return sysctl_lookup(SYSCTLFN_CALL(&node));
}

static u_int busfreq, busfreq_MHz;

static int
sysctl_machdep_cpu_clock(SYSCTLFN_ARGS)
{
	struct sysctlnode node;
	u_int corefreq;
	int error;

	node = *rnode;
	node.sysctl_data = &corefreq;

	corefreq = 0;	/* XXX */
	error = sysctl_lookup(SYSCTLFN_CALL(&node));
	if (error || newp == NULL)
		return error;

	/* XXX for now */
	error = ENODEV;
	return error;
}

SYSCTL_SETUP(sysctl_machdep_setup, "sysctl machdep subtree setup")
{
	static int altivec;

	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT,
	    CTLTYPE_NODE, "machdep", NULL,
	    NULL, 0, NULL, 0,
	    CTL_MACHDEP, CTL_EOL);

	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT|CTLFLAG_IMMEDIATE,
	    CTLTYPE_INT, "cachelinesize", NULL,
	    NULL, CACHELINESIZE, NULL, 0,
	    CTL_MACHDEP, CPU_CACHELINE, CTL_EOL);
#if 0
	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT,
	    CTLTYPE_STRUCT, "cacheinfo", NULL,
	    sysctl_machdep_cacheinfo, 0, NULL, 0,
	    CTL_MACHDEP, CPU_CACHEINFO, CTL_EOL);
#endif
	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT,
	    CTLTYPE_STRING, "booted_kernel", NULL,
	    sysctl_machdep_booted_kernel, 0, NULL, 0,
	    CTL_MACHDEP, CPU_BOOTED_KERNEL, CTL_EOL);

	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT,
	    CTLTYPE_INT, "altivec", NULL,
	    NULL, 0, &altivec, 0,
	    CTL_MACHDEP, CPU_ALTIVEC, CTL_EOL);

	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT,
	    CTLTYPE_INT, "bus_clock", NULL,
	    NULL, 0, &busfreq_MHz, 0,
	    CTL_MACHDEP, CTL_CREATE, CTL_EOL);

	sysctl_createv(clog, 0, NULL, NULL, CTLFLAG_PERMANENT|CTLFLAG_READWRITE,
	    CTLTYPE_INT, "cpu_clock", NULL,
	    sysctl_machdep_cpu_clock, 0, NULL, 0,
	    CTL_MACHDEP, CTL_CREATE, CTL_EOL);

#ifdef ALTIVEC
	switch (mfpvr() >> 16) {
	case PPC970:
	case PPC970FX:
	case CELL:
		altivec = 1;
		break;
	}
#endif

	/* Get bus clock frequency. */
	busfreq = 0;
	busfreq_MHz = (busfreq + 500000) / 1000000;
}

/*
 * Crash dump handling.
 */

void
dumpsys()
{
	printf("dumpsys: TBD\n");
}

#ifndef __HAVE_GENERIC_SOFT_INTERRUPTS
#error
#include "zsc.h"
#include "com.h"

int zssoft __P((void *));
void comsoft __P((void));

/*
 * Soft tty interrupts.
 */
void
softserial(v)
	void *v;
{
#if NZSC > 0
	zssoft(NULL);
#endif
#if NCOM > 0
	comsoft();
#endif
}
#endif

/*
 * Halt or reboot the machine after syncing/dumping according to howto.
 */
void
cpu_reboot(howto, what)
	int howto;
	char *what;
{
	static int syncing;

#if 0
	sysmgr_intr_enable(0);
#endif

	boothowto = howto;
	if (!cold && !(howto & RB_NOSYNC) && !syncing) {
		syncing = 1;
		vfs_shutdown();		/* sync */
		resettodr();		/* set wall clock */
	}

#ifdef MULTIPROCESSOR
	/* Halt other CPU.  XXX for now... */
	send_ipi(&cpu_info[1 - cpu_number()], IPI_HALT);
	delay(100000);	/* XXX */
#endif

	splhigh();

	if (!cold && (howto & RB_DUMP))
		dumpsys();

	doshutdownhooks();

	if ((howto & RB_POWERDOWN) == RB_POWERDOWN) {
		sysmgr_shutdown();

		printf("calling HV_shutdown_lpar(OFF)...\n");
		HV_shutdown_lpar(2);
		printf("WARNING: powerdown failed!\n");
	}

	if (howto & RB_HALT) {
		printf("halted\n\n");

		/* flush cache for msgbuf */
		__syncicache((void *)msgbuf_paddr, round_page(MSGBUFSIZE));

		ppc_exit();
	}

	printf("rebooting\n\n");

	/* flush cache for msgbuf */
	__syncicache((void *)msgbuf_paddr, round_page(MSGBUFSIZE));

#if 0	/* HV_shutdown_lpar() is much faster. */
	sysmgr_reboot();
#endif
	printf("calling HV_shutdown_lpar(RESTART)...\n");
	HV_shutdown_lpar(3);

	ppc_exit();
}

/*
 * Allocate vm space and mapin the I/O address
 */
void *
mapiodev(pa, len, ro)
	paddr_t pa;
	psize_t len;
	int ro;
{
	paddr_t faddr;
	vaddr_t taddr, va;
	int mode, off;

	if (ro)
		mode = VM_PROT_READ;
	else
		mode = VM_PROT_READ | VM_PROT_WRITE;

	faddr = trunc_page(pa);
	off = pa - faddr;
	len = round_page(off + len);
	va = taddr =
	    uvm_km_alloc(kernel_map, len, 0, UVM_KMF_VAONLY | UVM_KMF_NOWAIT);

	if (va == 0)
		return NULL;

	/*
	 * Only ptable is mapped with ro == 1, so use PTE_M too.
	 */
	if (ro)
		faddr |= PTE_M;

	for (; len > 0; len -= NBPG) {
		pmap_kenter_pa(taddr, faddr, mode, 0);
		faddr += NBPG;
		taddr += NBPG;
	}
	pmap_update(pmap_kernel());
	return (void *)(va + off);
}

#include "ukbd.h"
#include "rsxfb.h"

void
cninit()
{
#if NRSXFB > 0
	rsxfb_cnattach();
#endif
#if NUKBD > 0
	ukbd_cnattach();
#endif
}

static u_long ps3_rm_limit;
int ps3_ptab_cnt = 8192;

#if 0

puid    = 0x1
lpar_id = 0x2
rm_size      = 0x08000000 (0)
total        = 0x0f800000 (0)
2nd_mem_base = 0x6c01.4800.0000

vas_id  = 0x0c
actual_size  = 0x00100000
ptable_pa    = 0x5000.02a0.0000

#endif

void
get_memsize()
{
	u_long puid, lpar_id, rm_size;
	u_long min_htab_size, page_sizes, n_page_sizes, actual_size;
	u_long vas_id;
	u_long a1, a2, a3, a4;
	struct repository_value v;
	int err;

	HV_get_pu_id(&puid);
	HV_get_lpar_id(&lpar_id);

	/* Get real mode limit. */
	a1 = key1("bi");
	a2 = key2("pu");
	a3 = puid;
	a4 = key2("rm_size");
	err = HV_read_repository(lpar_id, a1, a2, a3, a4, &v);
	if (err)
		panic("get_memsize");
	rm_size = v.value1;

	ps3_rm_limit = rm_size;

#if 0
	/* Get total memory size. */
	a1 = key1("bi");
	a2 = key2("rgntotal");
	a3 = 0;
	a4 = 0;
	err = HV_read_repository(lpar_id, a1, a2, a3, a4, &v);
	totalmem = v.value1;
#endif

	/* Maximum htab size is 1MB because of hypervisor restriction. */
	min_htab_size = 20;	/* 2^20 (1MB) */

	/* Use two page sizes: 16MB(2^24) and 64KB(2^16). */
	page_sizes = 24L << 56 | 16L << 48;
	n_page_sizes = 2;

	err = HV_construct_virtual_address_space(min_htab_size,
	    n_page_sizes, page_sizes, &vas_id, &actual_size);

	if (err)
		panic("construct_virtual_address_space (%d)", err);

	err = HV_select_virtual_address_space(vas_id);
	if (err)
		panic("select_virtual_address_space (%d)", err);

	ps3_ptab_cnt = actual_size / 128;
}

#define	OFMEM_REGIONS	4
static struct mem_region OFmem[OFMEM_REGIONS + 1], OFavail[OFMEM_REGIONS + 3];

/*
 * This is called from pmap_bootstrap, before the system is really initialized.
 * It shall provide the total and the available regions of RAM.
 * Both lists must have a zero-size entry as terminator.
 * The available regions need not take the kernel into account, but needs
 * to provide space for two additional entry beyond the terminating one.
 */
void
mem_regions(memp, availp)
	struct mem_region **memp, **availp;
{
	OFmem[0].start = 0;
	OFmem[0].size  = ps3_rm_limit;
/*	OFmem[1].start = ps3_2nd_mem_base;	*/
/*	OFmem[1].size  = ps3_2nd_mem_size;	*/

	OFavail[0].start = 0x00100000;
	OFavail[0].size  = ps3_rm_limit - OFavail[0].start;

	*memp = OFmem;
	*availp = OFavail;
	return;
}

#define PS3_DMA_SIZE(size) roundup(size, 0x10000000)

u_long
ps3_allocate_dma_region(bus_id, dev_id, internal)
	u_long bus_id, dev_id;
	int internal;
{
	u_long len, dmabase, dmaaddr;
	u_long flags = 0;
	int err;
	struct mem_region *mp = OFmem;

	len = PS3_DMA_SIZE(ctob(physmem));
	if (internal)
		flags = 2;	/* 8bit mode */
	err = HV_alloc_device_dma_region(bus_id, dev_id, len, HV_PAGESIZE_16M,
	    flags, &dmabase);
	if (err)
		return 0;

	dmaaddr = dmabase;

	while (mp->size) {
		err = HV_map_device_dma_region(bus_id, dev_id, mp->start,
		    dmaaddr, mp->size, 0xf800000000000000);
		if (err)
			return 0;
		dmaaddr += mp->size;
		mp++;
	}

	return dmabase;
}

void
ppc_exit()
{
	mtdec(0x7fffffff);
	mtmsr(0);

	for (;;);
}

/*
 * Enter power-saving mode.
 */
void
cpu_idle()
{
	if (want_resched)
		return;
	asm volatile ("dssall; sync; isync");
	HV_pause(0);
}
