/*	$Id: ps3pic.c,v 1.4 2011/10/20 07:15:01 tsubai Exp $	*/

/*-
 * Copyright (c) 2007 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <uvm/uvm_extern.h>
#include <machine/autoconf.h>
#include <machine/cpufunc.h>
#include <machine/hcalls.h>

/* Interrupt state bitmap. */
static struct {
	u_long status[4];
	u_long enable[4];
} ps3pic[2];

static void isync(void);
static int cntlzd(long);
static u_long pu_id;

void
ps3pic_init_irq()
{
	int cpu, err;

	HV_get_pu_id(&pu_id);

	bzero(&ps3pic, sizeof ps3pic);
	for (cpu = 0; cpu < 2; cpu++) {
		err = HV_configure_irq_bitmap(pu_id, cpu,
		    vtophys((vaddr_t)&ps3pic[cpu]));
		if (err)
			panic("configure_irq_bitmap (%d)", err);
	}
}

inline void
isync()
{
	asm volatile ("isync");
}

void
ps3pic_enable_irq(irq)
	u_int irq;
{
	int idx = irq >> 6;	/* 0-3 */
	int bit = irq & 0x3f;	/* 0-63 */
	int cpu, err;
	u_long msr;

	if (irq > 255)
		panic("ps3pic_enable_irq: irq = %d", irq);

	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	isync();
	for (cpu = 0; cpu < 2; cpu++) {
		ps3pic[cpu].enable[idx] |= 0x8000000000000000UL >> bit;
		asm volatile ("sync");
		err  = HV_notify_imask_change(pu_id, cpu);
		if (err)
			printf("notify_imask_change failed\n");
	}
	mtmsr(msr);
}

void
ps3pic_disable_irq(irq)
	u_int irq;
{
	int idx = irq >> 6;	/* 0-3 */
	int bit = irq & 0x3f;	/* 0-63 */
	u_long msr;

	if (irq > 255)
		panic("ps3pic_disable_irq: irq = %d", irq);

	msr = mfmsr();
	mtmsr(msr & ~PSL_EE);
	isync();
	ps3pic[0].enable[idx] &= ~(0x8000000000000000UL >> bit);
	ps3pic[1].enable[idx] = ps3pic[0].enable[idx];
	mtmsr(msr);
}

static inline int
cntlzd(x)
	long x;
{
	int rv;

	asm ("cntlzd %0,%1" : "=r"(rv) : "r"(x));
	return rv;
}

int
ps3pic_read_irq()
{
	int base = 0;
	int irq, i;
	int cpu = cpu_number();
	u_long pending;

	for (i = 0; i < 4; i++) {
		pending = ps3pic[cpu].status[i] & ps3pic[cpu].enable[i];
		if (pending) {
			irq = base + cntlzd(pending);
			return irq;
		}
		base += 64;
	}
	return -1;
}

void
ps3pic_eoi(irq)
	u_int irq;
{
	HV_eoi(irq);
}
