/*	$Id: repo.c,v 1.4 2007/04/09 05:15:53 tsubai Exp $	*/

/*-
 * Copyright (c) 2006 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <machine/hcalls.h>
#include <machine/ps3.h>

int
ps3_read_timebase_freq(freq)
	u_long *freq;
{
	u_long node, a1, a2, a3, a4;
	struct repository_value v;
	int err;

	a1 = key1("be");
	a2 = key2("be");
	a3 = 0;
	a4 = 0;
	err = HV_read_repository(1, a1, a2, a3, a4, &v);
#if 0	/* XXX Ignore error. */
	if (err)
		return err;
#endif

	node = v.value1;

	a1 = key1("be");
	a2 = node;
	a3 = key2("clock");
	a4 = 0;
	err = HV_read_repository(1, a1, a2, a3, a4, &v);
	if (err)
		return err;

	*freq = v.value1;
	return 0;
}

int
ps3_read_bus_id(bus_index, bus_id)
	u_long bus_index;
	u_long *bus_id;
{
	u_long a1, a2;
	struct repository_value v;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("id");

	err = HV_read_repository(1, a1, a2, 0, 0, &v);
	if (err)
		return err;

	*bus_id = v.value1;
	return 0;
}

int
ps3_read_ndev(bus_index, ndev)
	u_long bus_index;
	u_long *ndev;
{
	u_long a1, a2;
	struct repository_value v;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("num_dev");

	err = HV_read_repository(1, a1, a2, 0, 0, &v);
	if (err)
		return err;

	*ndev = v.value1;
	return 0;
}

int
ps3_read_device_id(bus_index, device_index, device_id)
	u_long bus_index, device_index;
	u_long *device_id;
{
	u_long a1, a2, a3;
	struct repository_value v;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("dev") | device_index;
	a3 = key2("id");

	err = HV_read_repository(1, a1, a2, a3, 0, &v);
	if (err)
		return err;

	*device_id = v.value1;
	return 0;
}

int
ps3_read_device_type(bus_index, device_index, device_type)
	u_long bus_index, device_index;
	u_long *device_type;
{
	u_long a1, a2, a3;
	struct repository_value v;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("dev") | device_index;
	a3 = key2("type");

	err = HV_read_repository(1, a1, a2, a3, 0, &v);
	if (err)
		return err;

	*device_type = v.value1;
	return 0;
}

int
ps3_read_interrupt(bus_index, device_index, intr_index, intr_type, intr_data)
	u_long bus_index, device_index, intr_index;
	u_long *intr_type, *intr_data;
{
	u_long a1, a2, a3;
	struct repository_value v;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("dev") | device_index;
	a3 = key2("intr") | intr_index;

	err = HV_read_repository(1, a1, a2, a3, 0, &v);
	if (err)
		return err;

	*intr_type = v.value1;
	*intr_data = v.value2;
	return 0;
}

int
ps3_read_region_info(bus_index, device_index, region_index, addr, size, type)
	u_long bus_index, device_index, region_index;
	u_long *addr, *size, *type;
{
	u_long a1, a2, a3, a4;
	struct repository_value v1, v2;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("dev") | device_index;
	a3 = key2("reg") | region_index;
	a4 = key2("data");

	err = HV_read_repository(1, a1, a2, a3, a4, &v1);
	if (err)
		return err;

	a4 = key2("type");

	err = HV_read_repository(1, a1, a2, a3, a4, &v2);
	if (err)
		return err;

	*addr = v1.value1;
	*size = v1.value2;
	*type = v2.value1;
	return 0;
}

#if 0
int
ps3_read_platform_id(id)
	u_long *id;
{
	u_long a1, a2;
	struct repository_value v;
	int err;

	a1 = key1("plat");
	a2 = key2("id");

	err = HV_read_repository(1, a1, a2, 0, 0, &v);
	if (err)
		return err;

	*id = v.value1;
	return 0;
}

int
ps3_read_bus_type(bus_index, bus_type)
	u_long bus_index;
	u_long *bus_type;
{
	u_long a1, a2;
	struct repository_value v;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("type");

	err = HV_read_repository(1, a1, a2, 0, 0, &v);
	if (err)
		return err;

	*bus_type = v.value1;
	return 0;
}

int
ps3_read_pci_bus(bus_index, device_index, pci_bus)
	u_long bus_index, device_index;
	u_long *pci_bus;
{
	u_long a1, a2, a3;
	struct repository_value v;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("dev") | device_index;
	a3 = key2("pci_bus");

	err = HV_read_repository(1, a1, a2, a3, 0, &v);
	if (err)
		return err;

	*pci_bus = v.value1;
	return 0;
}

int
ps3_read_pci_device(bus_index, device_index, pci_device)
	u_long bus_index, device_index;
	u_long *pci_device;
{
	u_long a1, a2, a3;
	struct repository_value v;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("dev") | device_index;
	a3 = key2("pci_dev");

	err = HV_read_repository(1, a1, a2, a3, 0, &v);
	if (err)
		return err;

	*pci_device = v.value1;
	return 0;
}

int
ps3_read_pci_func(bus_index, device_index, pci_func)
	u_long bus_index, device_index;
	u_long *pci_func;
{
	u_long a1, a2, a3;
	struct repository_value v;
	int err;

	a1 = key1("bus") | bus_index;
	a2 = key2("dev") | device_index;
	a3 = key2("pci_fun");

	err = HV_read_repository(1, a1, a2, a3, 0, &v);
	if (err)
		return err;

	*pci_func = v.value1;
	return 0;
}
#endif
