/*	$Id: vuart.c,v 1.11 2007/04/10 08:06:05 tsubai Exp $	*/

/*-
 * Copyright (c) 2007 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <machine/autoconf.h>
#include <machine/hcalls.h>
#include <machine/ps3.h>

#include <uvm/uvm_extern.h>

int vuart_intr(void *);
static int cntlzd(u_int64_t);

#define NVUART_PORT 3

u_int64_t vuart_bitmap[4];
int (*vuart_intrhand[NVUART_PORT])(void);

void
vuart_init()
{
	u_long outlet_id;
	int irq, port, err;

	err = HV_configure_vuart(vtophys((vaddr_t)&vuart_bitmap),
	    &outlet_id);
	if (err) {
		printf("vuart_init: failed to configure vuart");
		return;
	}

	irq = ps3_allocate_irq();
	err = HV_set_intrhand_irq(irq, outlet_id);
	if (err) {
		printf("vuart_init: failed to configure interrupt");
		return;
	}

	/* Disable all interrupts. */
	for (port = 0; port < 256; port++)
		vuart_enable_irq(port, 0);

	intr_establish(irq, IST_LEVEL, IPL_HIGH, vuart_intr, NULL);
}

inline int
cntlzd(x)
	u_long x;
{
	int rv;

	asm ("cntlzd %0,%1" : "=r"(rv) : "r"(x));
	return rv;
}

int
vuart_intr(v)
	void *v;
{
	int port, base, i, n;
	int rv = 0;
	u_long bits;

	for (i = 0; i < 4; i++) {
		bits = vuart_bitmap[i];
		base = i * 64;

		while (bits) {
			n = cntlzd(bits);
			bits &= ~(0x8000000000000000 >> n);
			port = base + n;

			if (port >= NVUART_PORT)
				continue;
			if (vuart_intrhand[port] == NULL)
				continue;

			rv |= (*vuart_intrhand[port])();
		}
	}
	return rv;
}

int
vuart_intr_establish(port, func)
	u_int port;
	int (*func)(void);
{
	if (port >= NVUART_PORT) {
		printf("vuart_intr_establish: port = %d\n", port);
		return 0;
	}
	vuart_intrhand[port] = func;
	return 0;
}

void
vuart_enable_irq(port, mask)
	u_int port;
	u_long mask;
{
	int err;

	err = HV_set_vuart_param(port, VUART_PARAM_INTR_ENABLE, mask);
	switch (err) {
	case 0:
	case HV_ENOENT:
		return;
	}

	printf("vuart_enable_irq failed port = %d, err = %d\n", port, err);
}

int
vuart_set_recvsize(port, size)
	u_int port;
	u_long size;
{
	u_int64_t max;
	int err;

	err = HV_get_vuart_param(port, VUART_PARAM_RBUFSIZE, &max);
	if (err)
		panic("vuart_set_recvsize: failed to get rbuf size");

	/* ХåեλĤ꤬ n ХȤǳߤȯ롢Τ? */
	err = HV_set_vuart_param(port, VUART_PARAM_RECV_THRES, max - size);
	if (err)
		panic("vuart_set_recvsize: failed to set recv threshold");

	return 0;
}
