/*	$Id: boot.c,v 1.15 2007/04/20 12:57:48 tsubai Exp $	*/

/*-
 * Copyright (c) 2007 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <lib/libkern/libkern.h>
#include <lib/libsa/stand.h>
#include <lib/libsa/loadfile.h>

#include <machine/cpu.h>
#include <machine/cpufunc.h>
#include <machine/hcalls.h>
#include <machine/lv1call.h>

#include "ps3dev.h"

static void chain(void *, void *, void *);

#ifdef DEBUG
# define DPRINTF printf
#else
# define DPRINTF while (0) printf
#endif

static char *kernels[] = {
#if 0
	"net:/boot.ps3",
#endif
	"net:/netbsd", "net:/netbsd.gz", "net:/netbsd.ps3",
#if 0
	"disk:/netbsd", "disk:/netbsd.gz", "disk:/netbsd.ps3",
	"cd:/netbsd", "cd:/netbsd.gz", "cd:/netbsd.ps3",
#endif
	NULL
};

void
main()
{
	char *entry, *ssym, *esym;
	u_long marks[MARK_MAX];
	int i;

	gelic_consinit();

	printf("\n");
	printf(">> %s, Revision %s\n", bootprog_name, bootprog_rev);
	printf(">> (%s, %s)\n", bootprog_maker, bootprog_date);
	printf("\n");

	for (;;) {
		for (i = 0; kernels[i]; i++) {
			printf("Trying %s\n", kernels[i]);

			marks[MARK_START] = 0;
			if (loadfile(kernels[i], marks, LOAD_KERNEL) >= 0)
				goto loaded;
		}
		_rtt();
	}
loaded:
	entry = (char *)marks[MARK_ENTRY];
	ssym = (char *)marks[MARK_SYM];
	esym = (char *)marks[MARK_END];

	printf(" start=%p\n", entry);
	__syncicache(entry, ssym - entry);

	/* Start secondary thread. */
	*(volatile int *)0x20 = 0;
	asm volatile ("sync" ::: "memory");

	entry2[0] = entry;
	asm volatile ("sync" ::: "memory");
	delay(10000);

	chain(entry, ssym, esym);
	_rtt();
}

void
chain(entry, ssym, esym)
	void *entry, *ssym, *esym;
{
	/* XXX Close gelic. */
	lv1_net_stop_rx_dma(1, 0, 0);
#if 0
	HV_close(1, 0);
#endif

	/* Branch to entry point. */
	asm volatile ("		\n\
		mtctr	%0	\n\
		mr	3,%1	\n\
		mr	4,%2	\n\
		mr	5,%3	\n\
		bctr		\n"
	    :: "r"(entry), "r"(0x4e65744253440000), "r"(ssym), "r"(esym));

	panic("chain");
}

__dead void
_rtt()
{
	printf("rebooting...");
	HV_shutdown_lpar(3);
}

void
__syncicache(from, len)
	void *from;
	int len;
{
	int l, off;
	char *p;

	off = (u_long)from & (CACHELINESIZE - 1);
	len += off;

	p = from;
	l = len;
	do {
		asm volatile ("dcbst 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");

	p = from;
	l = len;
	do {
		asm volatile ("icbi 0,%0" :: "r"(p));
		p += CACHELINESIZE;
	} while ((l -= CACHELINESIZE) > 0);
	asm volatile ("sync");	/* required on 745x */
	asm volatile ("isync");
}

void
putchar(c)
	int c;
{
	gelic_cons(c);
	delay(200);
}

static u_int ns_per_tick = 12;

void
delay(us)
	unsigned int us;
{
	u_int64_t n = us;
	u_int64_t tb;

	tb = mftb();
	tb += (n * 1000 + ns_per_tick - 1) / ns_per_tick;

	while ((long)(mftb() - tb) < 0);
	asm volatile ("sync; isync");
}

void
dump(v)
	void *v;
{
	u_char *p = v;
	int i, j;

	for (i = 0; i < 16; i++) {
		printf("%p", p);
		for (j = 0; j < 16; j++) {
			printf(" %x", *p >> 4);
			printf("%x", *p & 0xf);
			p++;
		}
		printf("\n");
	}
}
