/*	$Id: devopen.c,v 1.13 2007/04/20 11:45:11 tsubai Exp $	*/

/*-
 * Copyright (C) 1999 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <lib/libkern/libkern.h>
#include <lib/libsa/stand.h>
#include <lib/libsa/ufs.h>
#include <lib/libsa/nfs.h>
#include <lib/libsa/cd9660.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <lib/libsa/net.h>

#include <machine/hcalls.h>
#include <machine/lv1call.h>

#include "ps3dev.h"

#ifdef DEBUG
# define DPRINTF printf
#else
# define DPRINTF while (0) printf
#endif

struct hvdev {
	int bus_id;
	int dev_id;
	int region;
	int blksize;
	paddr_t dma;
};

int dkopen(struct open_file *, ...);
int dkclose(struct open_file *);
int dkstrategy(void *, int, daddr_t, size_t, void *, size_t *);

struct devsw devsw[] = {
	{ "dk", dkstrategy, dkopen, dkclose, noioctl }
};
int ndevs = 1;

struct fs_ops file_system_ufs = FS_OPS(ufs);
struct fs_ops file_system_nfs = FS_OPS(nfs);
struct fs_ops file_system_cd9660 = FS_OPS(cd9660);

struct fs_ops file_system[2];
int nfsys;

int
devopen(f, fname, file)
	struct open_file *f;
	const char *fname;
	char **file;
{
	char *cp;
	int error;
	static struct hvdev hvdev;

	DPRINTF("devopen: %s\n", fname);

	cp = strchr(fname, ':');
	if (cp == NULL)
		return ENOENT;
	cp++;
	if (*cp == 0)
		return ENOENT;

	DPRINTF("devopen file = %s\n", cp);

	f->f_dev = devsw;
	f->f_devdata = &hvdev;
	*file = cp;

	if (strncmp(fname, "net:", 4) == 0) {
		file_system[0] = file_system_nfs;
		nfsys = 1;

		hvdev.bus_id = 1;
		hvdev.dev_id = 0;

		if ((error = net_open(&hvdev)) != 0) {
			printf("Can't open NFS network connection\n");
			return error;
		}
		return 0;
	}
	if (strncmp(fname, "disk:", 5) == 0) {		/* SATA */
		file_system[0] = file_system_ufs;
		file_system[1] = file_system_cd9660;
		nfsys = 2;
		hvdev.bus_id = 4;
		hvdev.dev_id = 2;
		hvdev.region = 3;
		hvdev.blksize = 512;

		HV_open(hvdev.bus_id, hvdev.dev_id, 0);
		dma_alloc(hvdev.bus_id, hvdev.dev_id, 0);
		return 0;
	}
	if (strncmp(fname, "cd:", 3) == 0) {		/* CDROM */
		file_system[0] = file_system_ufs;
		file_system[1] = file_system_cd9660;
		nfsys = 2;
		hvdev.bus_id = 4;
		hvdev.dev_id = 3;
		hvdev.region = 0;
		hvdev.blksize = 2048;

		HV_open(hvdev.bus_id, hvdev.dev_id, 0);
		dma_alloc(hvdev.bus_id, hvdev.dev_id, 0);
		return 0;
	}

	return ENXIO;
}

int
dkopen(struct open_file *f, ...)
{
	/* XXX not used */
	return 0;
}

int
dkclose(f)
	struct open_file *f;
{
#if 0
	struct hvdev *dev = f->f_devdata;

	HV_close(dev->bus_id, dev->dev_id);
#endif
	DPRINTF("dkclose\n");

	return 0;
}

int
dkstrategy(devdata, rw, blk, size, buf, rsize)
	void *devdata;
	int rw;
	daddr_t blk;
	size_t size;
	void *buf;
	size_t *rsize;
{
	struct hvdev *dev = devdata;
	u_int64_t tag, status;
	int devid = dev->dev_id;
	int region = dev->region;
	int i, rv;

	DPRINTF("strategy blk = %ld, size = %ld, buf = %p\n", blk, size, buf);

	/* XXX disklabel */

	blk = blk / (dev->blksize / 512);
	if (size % dev->blksize)
		return ENXIO;

	*rsize = size;
	size /= dev->blksize;

	rv = lv1_storage_read(devid, region, blk, size, 0, (paddr_t)buf, &tag);
	if (rv)
		return ENXIO;

	for (i = 0; i < 1000; i++) {
		rv = lv1_storage_get_async_status(devid, &tag, &status);
		if (rv == 0)
			break;
		delay(1000);
	}
	if (rv || status)
		return ENXIO;

	return 0;
}
