/*	$Id: gelic.c,v 1.14 2010/01/26 08:32:32 tsubai Exp $	*/

/*-
 * Copyright (c) 2006,2007 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <lib/libkern/libkern.h>
#include <lib/libsa/stand.h>

#include <machine/cpufunc.h>
#include <machine/hcalls.h>
#include <machine/lv1call.h>
#include <machine/psl.h>

#include "ps3dev.h"

struct gelic_dmadesc {
	u_int32_t address;
	u_int32_t bufsize;
	u_int32_t next;
	u_int32_t cmdstat;
	u_int32_t result_size;		/* N/U */
	u_int32_t rxlen;		/* Receive length */
	u_int32_t data_status;
	u_int32_t rxerror;		/* Receive errors */
} __attribute__((__aligned__(32)));

#define GELIC_CMDSTAT_NOCSUM	0x00080000
#define GELIC_CMDSTAT_TCPCSUM	0x000a0000
#define GELIC_CMDSTAT_UDPCSUM	0x000b0000
#define GELIC_CMDSTAT_EOF	0x00040000

#define GELIC_DESCR_COMPLETE	0x00000000
#define GELIC_DESCR_OWNED	0xa0000000

#define GELIC_TXINT		(1L << 56)	/* GDTDCEINT */
#define GELIC_RXINT0		(1L << 60)	/* GRFANMINT */
#define GELIC_RXINT1		(1L << 14)	/* GDADCEINT */
#define GELIC_RXINT		(GELIC_RXINT0 | GELIC_RXINT1)

#define GELIC_RXERRORS		0x7def8000	/* ? */

/* net_control commands */
#define GELIC_GET_MAC_ADDRESS	0x0001
#define GELIC_GET_LINK_STATUS	0x0002
#define GELIC_GET_VLAN_ID	0x0004

/* GET_LINK_STATUS values */
#define GELIC_LINK_UP		0x0001
#define GELIC_FULL_DUPLEX	0x0002
#define GELIC_AUTO_NEG		0x0004
#define GELIC_SPEED_10		0x0010
#define GELIC_SPEED_100		0x0020
#define GELIC_SPEED_1000	0x0040

#define GELIC_PORT 2

char laddr[6];
static u_int vlanid;
static u_long dma;
static int busid = 1;
static int devid = 0;

#define bus_addr(va) (dma + (vaddr_t)(va))

void
gelic_cons(data)
	int data;
{
	volatile static struct gelic_dmadesc desc;
	volatile struct gelic_dmadesc *dp = &desc;
	int len, rv, i;
	static char buf[64];
	char *p = buf + 4;

	static char ea1[] = { 0x00, 0x10, 0x18, 0x0b, 0x1d, 0xfd };
	char *ea2 = laddr;
	static char addr1[] = { 1, 1, 1, 1 };
	static char addr2[] = { 192, 168, 253, 2 };
	static u_int16_t id;

	bcopy(ea1, p + 0, sizeof ea1);		/* dhost */
	bcopy(ea2, p + 6, sizeof ea2);		/* shost */
	p[12] = 0x08;
	p[13] = 0x00;

	p[14] = 0x45;		/* IP header */
	p[15] = 0x00;

	p[16] = 0x00;
	p[17] = 0x1d;

	p[18] = id >> 8;	/* 18-19 id */
	p[19] = id++;

	p[20] = 0;
	p[21] = 0;

	p[22] = 0x40;		/* ttl */
	p[23] = 0x11;		/* UDP */

	p[24] = 0;		/* 24-25 IP cksum */
	p[25] = 0;

	bcopy(addr1, p + 26, sizeof addr1);	/* 26-29 src addr */
	bcopy(addr2, p + 30, sizeof addr2);	/* 30-33 dst addr */

	p[34] = 0x25;		/* src port */
	p[35] = 0x80;

	p[36] = 0x25;		/* dst port */
	p[37] = 0x80;

	p[38] = 0;		/* udp len */
	p[39] = 9;

	p[40] = 0;		/* 40-41 udp checksum */
	p[41] = 0;

	p[42] = data;		/* data */
	len = 43;

	u_int16_t *pp = (void *)(p + 14);
	u_int sum = 0;
	int xx;

	for (xx = 0; xx < 10; xx++)
		sum += pp[xx];

	sum = (sum >> 16) + (sum & 0xffff);
	if (sum > 0xffff)
		sum -= 0xffff;
	pp[5] = sum ^ 0xffff;

	/* XXX VLAN hack */
	if (vlanid != -1) {
		memmove(buf, p, 12);
		buf[12] = 0x81;
		buf[13] = 0x00;
		buf[14] = vlanid >> 8;
		buf[15] = vlanid;
		len += 4;
	}

	dp->address = bus_addr(buf);
	dp->bufsize = len;
	dp->cmdstat = GELIC_CMDSTAT_NOCSUM | GELIC_DESCR_OWNED;
	dp->cmdstat |= GELIC_CMDSTAT_EOF;
	dp->result_size = 0;
	dp->rxlen = 0;
	dp->data_status = 0;
	dp->rxerror = 0;
	dp->next = 0;

	asm volatile ("sync" ::: "memory");

	for (i = 0; i < 1000; i++) {
		rv = lv1_net_start_tx_dma(busid, devid, bus_addr(dp), 0);
		if (rv != HV_EBUSY)
			break;
	}
	if (rv)
		return;

	/* Wait for TX done. */
	for (i = 0; i < 10000; i++) {
		if ((dp->cmdstat & GELIC_DESCR_OWNED) == 0)
			break;
		delay(1);
	}
}

void
gelic_consinit()
{
	long v1, v2;
	int i, rv;

retry:
	rv = HV_open(busid, devid, 0);
	if (rv == HV_EBUSY)
		rv = 0;
	if (rv)
		goto retry;

	for (i = 0; i < 1000; i++) delay(1000);

	for (i = 0; i < 10; i++) {
		lv1_net_stop_tx_dma(busid, devid, 0);
		lv1_net_stop_rx_dma(busid, devid, 0);
	}

	for (i = 0; i < 1000; i++) {
		rv = lv1_net_control(busid, devid, GELIC_GET_LINK_STATUS,
		    GELIC_PORT, 0, 0, &v1, &v2);
		if (rv == 0 && (v1 & GELIC_LINK_UP))
			break;
		delay(10000);
	}

	lv1_net_control(busid, devid, GELIC_GET_MAC_ADDRESS, 0, 0, 0, &v1, &v2);
	bcopy((char *)&v1 + 2, laddr, 6);

	dma = dma_alloc(busid, devid, 0);
	// dma = 0xc0000000 (flash)
	// dma = 0xd0000000

	vlanid = -1;
	for (i = 1; i < 2; i++) {
		rv = lv1_net_control(busid, devid,
			GELIC_GET_VLAN_ID, i + 1, 0, 0, &v1, &v2);
		if (rv == 0)
			vlanid = v1;
	}

	rv = lv1_net_set_interrupt_mask(busid, devid, 0, 0);

#define GELIC_BROADCAST_ADDR 0xffffffffffff
	lv1_net_remove_multicast_address(busid, devid, 0, 1);
	lv1_net_add_multicast_address(busid, devid, GELIC_BROADCAST_ADDR, 0);
}

paddr_t
dma_alloc(bus_id, dev_id, internal)
	u_long bus_id, dev_id;
	int internal;
{
	u_long len, dmabase;
	u_long flags = 0;
	int err;

	len = 0x10000000;
	if (internal)
		flags = 2;	/* 8bit mode */
	err = lv1_allocate_device_dma_region(bus_id, dev_id, len,
	    PS3_IO_PAGESIZE, flags, &dmabase);
	if (err)
		return 0;

	err = lv1_map_device_dma_region(bus_id, dev_id, 0,
	    dmabase, 0x08000000, 0xf800000000000000);
	if (err)
		return 0;

	return dmabase;
}

volatile static struct gelic_dmadesc rxdesc;
volatile static struct gelic_dmadesc txdesc;
static char recvbuf[2048] __attribute__((aligned(128)));
static char sendbuf[2048];

int
gelic_send(pkt, len)
	void *pkt;
	int len;
{
	char *p = pkt;
	int rv, i;

	for (i = 0; i < 10000; i++) {
		if ((txdesc.cmdstat & GELIC_DESCR_OWNED) == 0)
			goto start;
		delay(10);
	}
	printf("GELIC BUSY\n");
	return -1;

start:
	bzero(sendbuf, sizeof sendbuf);

	bcopy(p, sendbuf, 6);
	bcopy(laddr, sendbuf + 6, 6);
	bcopy(p + 12, sendbuf + 16, len - 12);

	/* XXX VLAN hack */
	if (vlanid != -1) {
		sendbuf[12] = 0x81;
		sendbuf[13] = 0x00;
		sendbuf[14] = vlanid >> 8;
		sendbuf[15] = vlanid;
		len += 4;
	}

	bzero((void *)&txdesc, sizeof txdesc);
	txdesc.address = bus_addr(sendbuf);
	txdesc.bufsize = len;
	txdesc.cmdstat = GELIC_CMDSTAT_NOCSUM | GELIC_DESCR_OWNED;
	txdesc.cmdstat |= GELIC_CMDSTAT_EOF;

	asm volatile ("sync" ::: "memory");

	for (i = 0; i < 1000; i++) {
		rv = lv1_net_start_tx_dma(busid, devid, bus_addr(&txdesc), 0);
		if (rv != HV_EBUSY)
			break;
	}
	return 0;
}

int
gelic_recv(pkt, maxlen, timo)
	void *pkt;
	int maxlen, timo;
{
	int len;

	/* Wait for RX done. */
	timo *= 1000;
	while (timo-- > 0) {
		if ((rxdesc.cmdstat & GELIC_DESCR_OWNED) == 0)
			goto gotit;

		delay(1000);
	}
	printf("GELIC TIMEOUT\n");
	/* XXX reset receiver */
	return -1;

gotit:
	delay(200);
	if (rxdesc.rxerror & GELIC_RXERRORS) {
		printf("GELIC ERROR\n");
		return -1;
	}

	len = rxdesc.rxlen;
	len -= 2;
	if (len > maxlen)
		len = maxlen;
	bcopy(recvbuf + 2, pkt, len);

	bzero((void *)&rxdesc, sizeof rxdesc);
	rxdesc.address = bus_addr(recvbuf);
	rxdesc.bufsize = 2000;
	rxdesc.cmdstat = GELIC_DESCR_OWNED;

	asm volatile ("sync" ::: "memory");
	lv1_net_start_rx_dma(busid, devid, bus_addr(&rxdesc), 0);

	return len;
}

void
gelic_init()
{
	bzero((void *)&rxdesc, sizeof rxdesc);
	rxdesc.address = bus_addr(recvbuf);
	rxdesc.bufsize = 2000;
	rxdesc.cmdstat = GELIC_DESCR_OWNED;

	asm volatile ("sync" ::: "memory");
	lv1_net_start_rx_dma(busid, devid, bus_addr(&rxdesc), 0);
}
