/*	$Id: ucons.c,v 1.4 2011/10/20 06:16:18 tsubai Exp $	*/

/*-
 * Copyright (c) 2006 Tsubai Masanari.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/device.h>

#include <machine/bus.h>
#include <machine/hcalls.h>
#include <machine/ps3.h>

#include <dev/cons.h>

#include <arch/ps3/ucons/usb.h>
#include <arch/ps3/ucons/usbdi.h>

#include <dev/usb/ohcireg.h>
#include <arch/ps3/ucons/ohcivar.h>

static void ohci_obio_attach(struct device *, struct device *, void *);
static void usb_attach(usbd_bus_handle);

void ucons_consinit(void);
void uhub_attach(void *);

extern struct consdev consdev_uplcom;

void
ucons_consinit()
{
	static struct ohci_softc sc;
	struct confargs ca;
	static int initted;

	if (initted == 0) {
		ca.ca_bus = 1;
		ca.ca_dev = 1;
		ca.ca_func = 0;
		ca.ca_intr = 16;
		ohci_obio_attach(NULL, (void *)&sc, &ca);
		(*consdev_uplcom.cn_init)(0);
		initted = 1;
	}

	cn_tab = &consdev_uplcom;
}

void
ohci_obio_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct ohci_softc *sc = (void *)self;
	struct confargs *ca = aux;
	u_long bus_id = ca->ca_bus;
	u_long dev_id = ca->ca_dev;
	u_long func = ca->ca_func;
	u_long bus_index = bus_id;	/* XXX */
	u_long device_index = dev_id;	/* XXX */
	u_long region_addr, region_size, region_type, addr, dma_base;
	int irq, err;

	err = HV_open(bus_id, dev_id, 0);
	if (err == HV_EBUSY)
		err = 0;
	if (err) {
		printf("open failed (%d)\n", err);
		return;
	}

	if (ps3_read_region_info(bus_index, device_index, func, 
	    &region_addr, &region_size, &region_type)) {
		printf("read_region_data failed\n");
		return;
	}

	err = HV_map_device_mmio_region(bus_id, dev_id, region_addr,
	    region_size, HV_PAGESIZE_4K, &addr);
	if (err) {
		printf("map_device_mmio_region failed (%d)\n", err);
		return;
	}

	dma_base = ps3_allocate_dma_region(bus_id, dev_id, 1);
	if (dma_base == 0) {
		printf("couldn't map dma region\n");
		return;
	}

	sc->sc_size = region_size;
	sc->iot = 0;
	bus_space_map(0, addr, region_size, 0, &sc->ioh);

	/* Disable interrupts, so we don't get any spurious ones. */
	bus_space_write_4(sc->iot, sc->ioh, OHCI_INTERRUPT_DISABLE,
	    OHCI_ALL_INTRS);

	irq = ps3_connect_io_irq(ca->ca_intr);
	if (irq < 0) {
		printf("connect_io_irq failed\n");
		return;
	}

	/* Figure out vendor for root hub descriptor. */
	sc->sc_id_vendor = 0;
	strcpy(sc->sc_vendor, "SONY PS3");

	err = ohci_init(sc);
	if (err != USBD_NORMAL_COMPLETION) {
		printf("ohci_init failed (%d)\n", err);
		return;
	}

// 2KB
//printf("sizeof ohci_softc = %ld\n", sizeof *sc);

	usb_attach(&sc->sc_bus);
}

static struct usbd_device sc_device;	/* dummy device for root hub */

void
usb_attach(bus)
	usbd_bus_handle bus;
{
	usbd_device_handle dev;
	int err, speed;

	switch (bus->usbrev) {
	case USBREV_1_0:
	case USBREV_1_1:
		speed = USB_SPEED_FULL;
		break;
	case USBREV_2_0:
		speed = USB_SPEED_HIGH;
		break;
	default:
		return;
	}

	dev = &sc_device;
	err = usbd_new_device(bus, speed, 0, dev);
	if (err)
		return;

	if (dev->ddesc.bDeviceClass != UDCLASS_HUB)
		panic("root hub is not a hub");

	/* Attach root hub. */
	uhub_attach(dev);

	if (dev->hub == NULL)
		return;

	dev->hub->explore(dev);
}
