/*	$NetBSD: usbdi.h,v 1.72 2006/12/05 17:35:35 christos Exp $	*/

/*
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Lennart Augustsson (lennart@augustsson.net) at
 * Carlstedt Research & Technology.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* Rename symbols to avoid conflicts. */
#define uhub_attach ucons_uhub_attach
#define uplcom_attach ucons_uplcom_attach
#define uplcom_match ucons_uplcom_match
#define ohci_init ucons_ohci_init

#define usb_delay_ms ucons_usb_delay_ms
#define usbd_fill_iface_data ucons_usbd_fill_iface_data
#define usbd_new_device ucons_usbd_new_device

#define usbd_set_config_index ucons_usbd_set_config_index
#define usbd_setup_pipe ucons_usbd_setup_pipe
#define usb_match_device ucons_usb_match_device
#define usb_transfer_complete ucons_usb_transfer_complete
#define usbd_clear_endpoint_stall_async ucons_usbd_clear_endpoint_stall_async

#define usbd_do_request ucons_usbd_do_request

#define usbd_open_pipe ucons_usbd_open_pipe
#define usbd_open_pipe_intr ucons_usbd_open_pipe_intr

#define usbd_setup_xfer ucons_usbd_setup_xfer
#define usbd_transfer ucons_usbd_transfer

#define M_USB M_UCONS
#define M_USBDEV M_UCONSDEV

typedef struct usbd_bus		*usbd_bus_handle;
typedef struct usbd_device	*usbd_device_handle;
typedef struct usbd_interface	*usbd_interface_handle;
typedef struct usbd_pipe	*usbd_pipe_handle;
typedef struct usbd_xfer	*usbd_xfer_handle;

enum {		/* keep in sync with usbd_status_msgs */
	USBD_NORMAL_COMPLETION = 0, /* must be 0 */
	USBD_IN_PROGRESS,	/* 1 */
	/* errors */
	USBD_PENDING_REQUESTS,	/* 2 */
	USBD_NOT_STARTED,	/* 3 */
	USBD_INVAL,		/* 4 */
	USBD_NOMEM,		/* 5 */
	USBD_CANCELLED,		/* 6 */
	USBD_BAD_ADDRESS,	/* 7 */
	USBD_IN_USE,		/* 8 */
	USBD_NO_ADDR,		/* 9 */
	USBD_SET_ADDR_FAILED,	/* 10 */
	USBD_NO_POWER,		/* 11 */
	USBD_TOO_DEEP,		/* 12 */
	USBD_IOERROR,		/* 13 */
	USBD_NOT_CONFIGURED,	/* 14 */
	USBD_TIMEOUT,		/* 15 */
	USBD_SHORT_XFER,	/* 16 */
	USBD_STALLED,		/* 17 */
	USBD_INTERRUPTED,	/* 18 */

	USBD_ERROR_MAX		/* must be last */
};

/* Open flags */
#define USBD_EXCLUSIVE_USE	0x01

/* Use default (specified by ep. desc.) interval on interrupt pipe */
#define USBD_DEFAULT_INTERVAL	(-1)

/* Request flags */
#define USBD_NO_COPY		0x01	/* do not copy data to DMA buffer */
#define USBD_SYNCHRONOUS	0x02	/* wait for completion */
#define USBD_SHORT_XFER_OK	0x04	/* allow short reads */
#define USBD_FORCE_SHORT_XFER	0x08	/* force last short packet on write */

#define USBD_NO_TIMEOUT 0
#define USBD_DEFAULT_TIMEOUT 5000 /* ms = 5 s */

int usbd_open_pipe(usbd_interface_handle, u_int, u_int, usbd_pipe_handle *);
int usbd_transfer(usbd_xfer_handle);

void usbd_setup_xfer(usbd_xfer_handle, usbd_pipe_handle, void *, u_int32_t,
    u_int16_t);
int usbd_clear_endpoint_stall_async(usbd_pipe_handle);

int usbd_open_pipe_intr(usbd_interface_handle, u_int, usbd_pipe_handle *,
    void *, u_int32_t);
int usbd_do_request(usbd_device_handle, usb_device_request_t *, void *);

void usbd_delay_ms(usbd_device_handle, u_int);

int usbd_set_config_index(usbd_device_handle, int, int);

struct usb_devno {
	u_int16_t ud_vendor;
	u_int16_t ud_product;
};
const struct usb_devno *usb_match_device(const struct usb_devno *,
	u_int, u_int, u_int16_t, u_int16_t);
#define usb_lookup(tbl, vendor, product) \
	usb_match_device((const struct usb_devno *)(tbl), sizeof (tbl) / sizeof ((tbl)[0]), sizeof ((tbl)[0]), (vendor), (product))
#define	USB_PRODUCT_ANY		0xffff

/* NetBSD attachment information */

/* Attach data */
struct usb_attach_arg {
	int			vendor;
	int			product;
	usbd_device_handle	device;	/* current device */
};

/* XXX Perhaps USB should have its own levels? */
#define splusb splbio
#define splhardusb splbio

struct usbd_xfer;
struct usbd_pipe;

struct usbd_endpoint {
	usb_endpoint_descriptor_t *edesc;
	int	refcnt;
};

struct usbd_bus_methods {
	int	(*open_pipe)(struct usbd_pipe *pipe);
};

struct usbd_pipe_methods {
	int	(*transfer)(usbd_xfer_handle xfer);
	void	(*close)(usbd_pipe_handle pipe);
	void	(*done)(usbd_xfer_handle xfer);
};

struct usbd_port {
	usb_port_status_t	status;
	u_int8_t		portno;
	u_int8_t		restartcnt;
#define USBD_RESTART_MAX 5
	struct usbd_device     *device;	/* Connected device */
};

struct usbd_hub {
	int		      (*explore)(usbd_device_handle hub);
	usb_hub_descriptor_t	hubdesc;
	struct usbd_port        ports[1];
};

/*****/

struct usbd_bus {
	/* Filled by HC driver */
	const struct usbd_bus_methods *methods;
	u_int32_t		pipe_size; /* size of a pipe struct */

	/* Filled by usb driver */
	usbd_device_handle	devices[USB_MAX_DEVICES];
	char			usbrev;	/* USB revision */
#define USBREV_UNKNOWN	0
#define USBREV_PRE_1_0	1
#define USBREV_1_0	2
#define USBREV_1_1	3
#define USBREV_2_0	4
};

struct usbd_device {
	struct usbd_bus	       *bus;           /* our controller */
	struct usbd_pipe       *default_pipe;  /* pipe 0 */
	u_int8_t		address;       /* device addess */
	u_int8_t		config;	       /* current configuration # */
	u_int8_t		speed;         /* low/full/high speed */

	struct usbd_endpoint	def_ep;	       /* for pipe 0 */
	usb_endpoint_descriptor_t def_ep_desc; /* for pipe 0 */
	struct usbd_interface  *ifaces;        /* array of all interfaces */
	usb_device_descriptor_t ddesc;         /* device descriptor */
	usb_config_descriptor_t *cdesc;	       /* full config descr */
	struct usbd_hub	       *hub;           /* only if this is a hub */
};

struct usbd_interface {
	struct usbd_device     *device;
	usb_interface_descriptor_t *idesc;
	int			index;
	struct usbd_endpoint   *endpoints;
};

struct usbd_pipe {
	struct usbd_interface  *iface;
	struct usbd_device     *device;
	struct usbd_endpoint   *endpoint;
	int			refcnt;

	usbd_xfer_handle	intrxfer; /* used for repeating requests */
	char			repeat;

	/* Filled by HC driver. */
	const struct usbd_pipe_methods *methods;
};

struct usbd_xfer {
	struct usbd_pipe *pipe;
	void		*buffer;
	u_int32_t	length;
	u_int32_t	actlen;
	u_int16_t	flags;
//	u_int32_t	timeout;
	int		status;
	volatile u_int8_t done;

	/* For control pipe */
	usb_device_request_t request;

	/* For memory allocation */
	char		*dmabuf;

	u_int8_t	rqflags;
#define URQ_REQUEST	0x01
#define URQ_AUTO_DMABUF	0x10
#define URQ_DEV_DMABUF	0x20
};

/* Routines from usb_subr.c */
void	usb_delay_ms(u_int);

usbd_pipe_handle usbd_setup_pipe(usbd_device_handle, usbd_interface_handle,
    struct usbd_endpoint *);
int	usbd_new_device(usbd_bus_handle, int, int, struct usbd_device *);
int	usbd_fill_iface_data(usbd_device_handle, int, int);
void	usb_transfer_complete(usbd_xfer_handle);
